#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2014 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
convert cbf files to CrystFEL hdf5 files.
the program searches all .cbf files of a directory and converts them to
CrystFEL compatible hdf5.
'''

import sys,os,time,glob,stat
import numpy as np
import traceback
import logging
import subprocess
import h5py
import utilities as ut
import multiprocessing as mproc
from libDetXR import LibDetXR

_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

from CbfFile import *
from DatParser import *
from FileNameGen import *

class Cbf2CrystFEL():
  _dateFmt='%Y-%m-%d_%H:%M:%S'
  def __init__(self,args):
    self.args=args

  @staticmethod
  def AddCbfAttr(ds,attr):
    for k in CbfParser._keywords.keys():
      try:
        v=attr[k]
      except KeyError as e:
        continue
      if type(v)==tuple:
        if len(v)==2:
          ds.attrs[k]=v[0]
          ds.attrs['_u_'+k]=v[1]
        else:
          ds.attrs[k]=' '.join(map(str,v))
      else:
        ds.attrs[k]=v

  @staticmethod
  def AddMetaCbf(gidParent,attr):
    lcpl = h5py.h5p.create(h5py.h5p.LINK_CREATE)
    lcpl.set_create_intermediate_group(True)
    gid = h5py.h5g.create(gidParent, 'SLS', lcpl=lcpl)

    grp=h5py.Group(gid)

    for k in CbfParser._keywords.keys():
      try:
        v=attr[k]
      except KeyError as e:
        continue
      if type(v)==tuple:
        if len(v)==2:
          ds=grp.create_dataset(k, data=v[0])
          ds.attrs['unit']=v[1]
          #ds.attrs[k]=v[0]
        else:
          grp.create_dataset(k, data=' '.join(map(str,v)))
      else:
        grp.create_dataset(k, data=v)



  def ConvertLst(self,fnCBFLst,outDir):
    for fnCbf in fnCBFLst:
      fnHdf5=os.path.join(outDir,os.path.basename(fnCbf.rsplit('.',1)[0])+'.h5')
      self.Convert(fnCbf,fnHdf5)

  def Convert(self,fnCbf,fnHdf5):
    cmpr=self.args.cmpr
    CmprChunkBuild=self.CmprChunkBuild
    cbfParser=self.cbfParser
    _log.info(str((fnCbf,fnHdf5)))
    try:
      fid=ut.OpenFile(fnHdf5,'w')
    except IOError as e:
      _log.error(fnHdf5+'\n'+str(e))
      return
    obj=h5py.File(fid)
    obj.attrs['myValue'] = 'myAttribute'
    res,v=LibDetXR.GetVersion()
    attr=(('creator','Converted data with '+os.path.basename(__file__)+' ver. %d.%d.%d.%d'%tuple(v[::-1])+' from file '+fnCbf),
          ('target',"converted for CrystFEL"))
    ut.AddAttributes(fid,attr)

    lcpl = h5py.h5p.create(h5py.h5p.LINK_CREATE)
    lcpl.set_create_intermediate_group(True)
    gid = h5py.h5g.create(fid, 'data', lcpl=lcpl)

    try:
      cbfParser.Open(fnCbf)
    except (OSError,ValueError) as e:
      traceback.print_exc(1)
    else:
      cbfParser.ParseHeader(7)
      cbfArr,res=cbfParser.GetBinary()

    if self.args.uint16:
      cbfArr.clip(0,255) #todo: convert 32 to 16 bit
      cbfArr=np.asarray(cbfArr, dtype=np.uint16)
    #  my_array.clip(0,255)

    attr=cbfParser.header['header_contents']
    dtype=cbfArr.dtype
    #shape=(1,)+cbfArr.shape
    #maxshape=(h5py.h5s.UNLIMITED,)+cbfArr.shape
    #chunk=(1,)+cbfArr.shape
    shape=cbfArr.shape
    maxshape=cbfArr.shape
    chunk=cbfArr.shape
    dsid=ut.CreateDatasetChunked(gid,'rawdata',dtype,shape,maxshape,chunk,cmpr)

    ds=h5py.Dataset(dsid)
    Cbf2CrystFEL.AddCbfAttr(ds,attr)

    if cmpr:
      (cmprBuf,filter)=CmprChunkBuild(cbfArr)
      #dsid.write_chunk((0,0,0),cmprBuf,filters=filter)
      dsid.write_chunk((0,0),cmprBuf,filters=filter)
    else:
      #dsid.write_chunk((0,0,0),cbfArr)
      dsid.write_chunk((0,0),cbfArr)


    #grp=h5py.Group(gid)
    #grp["data"] =h5py.SoftLink('/data/rawdata')
    gid.links.create_soft('data', '/data/rawdata', lcpl=lcpl, lapl=None)

    Cbf2CrystFEL.AddMetaCbf(fid,attr)
    if self.args.pixMaskLink:
      self.AddPixelMaskLink(fid)
    ut.CloseFile(fid)

  def Run(self):
    inGlob=self.args.inGlob
    outDir=self.args.outDir
    verbose=self.args.verbose
    #if self.args.pixMaskLink:
    #  self.AddPixelMaskLink()
    #  return
    if not os.path.exists(outDir):
      outDir=os.path.expanduser(outDir)

    fnCBFLst=[]
    for glb in inGlob:
      glb=os.path.expanduser(glb)
      try:
        s=os.stat(glb)
        if stat.S_ISDIR(s.st_mode):
          glb=os.path.join(glb,'*.cbf')
      except OSError:
        pass
      else:
        pass
      fnCBFLst.extend(glob.iglob(glb))

    cmpr=self.args.cmpr
    if cmpr:
      self.CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)
    else:
      self.CmprChunkBuild=None
    self.cbfParser=CbfParser()

    ut.StopWatch.Start()
    fnCBFLst.sort()
    _log.info('converting %d cbf-files to hdf5-files...',len(fnCBFLst))
    #fnCBFLst=fnCBFLst[:5]
    #print fnCBFLst
    if self.args.multiproc:
      jobs = []
      maxJobs=mproc.cpu_count()
      for idx in range(maxJobs):
        #print 'files',fnCBFLst[idx::maxJobs]
        pipeParent,pipeChild = mproc.Pipe()
        _log.info('ConvertLst %d'%idx)
        p = mproc.Process(target=self.ConvertLst, args=(fnCBFLst[idx::maxJobs],outDir))
        jobs.append((p,pipeParent))
        p.start()
      for proc,pipe in jobs:
        proc.join()
      print 'done'

    else:
      self.ConvertLst(fnCBFLst,outDir)


    if verbose&2:
      ut.StopWatch.Log('CBF2HDF5', True)

  def AddPixelMaskLink(self,fid):
    '-v --inGlob=/tmp/cbf2CrystFEL/out_*.h5 --pixMaskLink /scratch/tomizaki/lysozyme1_master.h5 entry/instrument/detector/detectorSpecific/pixel_mask'
    verbose=self.args.verbose

    #mask = /processing/hitfinder/masks
    #mask_good = 0x27
    #mask_bad = 0x00
    #gid.links.create_soft('data', '/data/rawdata', lcpl=lcpl, lapl=None)



    fs=h5py.File(fid)
    #dset = fs.create_dataset("mask_good", data=np.array((0),dtype=np.uint16))
    #dset = fs.create_dataset("mask_bad", data=np.array((1),dtype=np.uint16))

    if fid.name.endswith('000001.h5'):
      fidMsk=ut.OpenFile(self.args.pixMaskLink[0],'r')
      pixMskSrc=h5py.h5o.open(fidMsk, self.args.pixMaskLink[1])#, lapl=self._lapl)
      pixMsk=np.ndarray(pixMskSrc.shape,np.uint16)
      ds=h5py.Dataset(pixMskSrc)
      pixMsk[:]=ds[:]
      pixMsk[pixMsk>0]=1
      dset = fs.create_dataset("mask", data=pixMsk)
    else:
      #maskFile=fid.name.rsplit('_')[0]+'_000001.h5'
      maskFile='out_000001.h5'
      fid.links.create_external('mask',maskFile, 'mask')

    pass


if __name__ == '__main__':
  import argparse #since python 2.7
  def RunWithArgs():

    if len(sys.argv)>1 and sys.argv[1][:4]=='Test':
      print eval(sys.argv[1]+'()')
      return


    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('-v --inGlob=/scratch/detectorData/liso/*.cbf --outDir=/tmp/cbf2CrystFEL',
                '-v --inGlob=/scratch/detectorData/liso/ --outDir=/tmp/cbf2CrystFEL  --cmpr s-zlib4',
                '-v --inGlob=/scratch/detectorData/liso/liso2_1_0000*.cbf --outDir=/tmp/cbf2CrystFEL  --cmpr s-zlib4 --multiproc',
                '-v --inGlob=/scratch/detectorData/liso/liso2_1_0000*.cbf --outDir=/tmp/cbf2CrystFEL  --cmpr s-zlib4 --multiproc --pixMaskLink /scratch/tomizaki/lysozyme1_master.h5 entry/instrument/detector/detectorSpecific/pixel_mask'
                )

    epilog='''\n
the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress
   bit2(0x04): multiproc
   bit3(0x08): cbffiles
   bit4(0x10): logparse
   bit5(0x20): datparse
   bit6(0x40): mcsparse
   default is 1+2=3 (no -v)
   -v is 1+2+4+16+32+64=119
   -v0 is turn all off\n\nExamples:'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

#Trial test to copy additional data:
# /home/zamofing_t/Documents/prj/WorkBench/hdf5/bin/h5copy -i /scratch/detectorData/CrystFEL/run0349/LCLS_2011_Feb20_r0349_223043_14307_cspad.h5 -s LCLS -o /tmp/cbf2CrystFEL/liso2_1_00001.h5 -d LCLS
# /home/zamofing_t/Documents/prj/WorkBench/hdf5/bin/h5copy -i /scratch/detectorData/CrystFEL/run0349/LCLS_2011_Feb20_r0349_223043_14307_cspad.h5 -s processing -o /tmp/cbf2CrystFEL/liso2_1_00001.h5 -d processing

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__,
                                     epilog=epilog)

    parser.add_argument('--inGlob', '-i',  nargs='+', required=True, help='input directory, glob or cbf file(s)')
    parser.add_argument('--outDir', '-o',             required=True, help='output directory of the file(s) to generate')
    parser.add_argument('-v',         dest='verbose',  type=int,  nargs='?', const=119,default=3,  help='verbose level (see below)')
    parser.add_argument('--multiproc',                 action='store_true', help='use multiprocessing method')
    parser.add_argument('--cmpr',                                           help='compression method (zlib[lvl],lz4,lzf,cbf)')
    parser.add_argument('--uint16',                   action='store_true', help='downscale 32 bit to 16 bit image')
    parser.add_argument('--pixMaskLink', nargs=2,                          help='add a pixel mask of a hdf5 file')
    args = parser.parse_args()
    Cbf2CrystFEL(args).Run()
    pass

#------------------ Main Code ----------------------------------

  hdlr=logging.StreamHandler()
  fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
  hdlr.setFormatter(fmt)
  del logging.root.handlers[:]
  logging.root.addHandler(hdlr)
  logging.root.setLevel(logging.DEBUG)

  logLvls={'CbfFile'    : logging.WARN,
           'DatParser'  : logging.WARN,
           'utilities'  : logging.WARN,
           'libDetXR'   : logging.WARN,
           __name__     : logging.INFO,
          }
  for k,v in logLvls.iteritems():
    logging.getLogger(k).setLevel(v)
  RunWithArgs()

