#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2015 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
add Nexus data to an existing hdf5 file that was converted with SPEC2hdf5
This class is directly called from SPEC2hdf5.
To modify the Nexus structure look adapt following files:
 - def AddDetector(self,gid,detKey,detDict):
 - def AddStructure(self):
'''

import numpy as np
import h5py
import utilities as ut
import logging, os, re

_log=logging.getLogger(__name__)


class HdfStruct:
  def __init__(self,fid,path,cls):
    self.obj=ut.CreateGrpPath(fid,path,cls)
  def __enter__(self):
    #print 'enter',self.obj #,self
    return self.obj
  def __exit__(self, exc_type, exc_value, traceback):
    #print 'exit',self.obj #self
    pass

class CSAXS_Nexus():
  def __init__(self,fid,basePath='/entry/collection/data', specPath='spec'):
    self.fid=fid
    self.basePath=basePath
    self.specPath=specPath
    specLst=list()
    base=h5py.Group(fid)[ os.path.join(basePath)]
    for k,v in base.iteritems():
      #print k,v
      if type(v)==h5py.Group and re.match('S[0-9]+$',k):
        try: specLst.append(v[specPath])
        except KeyError: _log.warning('no '+specPath+' in hdf5 obj '+str(v))
      elif k==specPath:
        try: specLst.append(v)
        except KeyError: _log.warning('no '+specPath+' in hdf5 obj '+str(v))

    self.specLst=specLst

  def NXObj(self,gid,name,data=None,spec=None,cls='?',cast=None,**kwargs):
    g=h5py.Group(gid)
    if spec:
      data=self.GetSpecVal(spec)
      if type(data)!=np.ndarray and data==np.NaN:
        _log.warning('no '+spec+' in SPEC')
        return

    if cast:
      if data.find('[')>=0:
        data=np.array(eval(data),cast)
      else:
        data=cast(data)

    if cls:
      if cls=='?':
        t=type(data)
        if t==np.ndarray:
          t=data.dtype
          if t in(np.float32,np.float64):
            cls='NX_FLOAT'
          elif t in(np.uint8,np.uint16,np.uint32):
            cls='NX_INT'

          else:
            _log.warning('unknown type: '+str(t))
        elif t==str:
          cls='NX_CHAR'
        elif t in(float,np.float32,np.float64):
          cls='NX_FLOAT'
        elif t==int:
          cls='NX_INT'
        elif t==bool:
          cls='NX_BOOLEAN'
        else:
          _log.warning('unknown type: '+str(t))
    try:
      o=g.create_dataset(name,data=data)
      for (k,v) in kwargs.iteritems(): o.attrs[k] = v
    except RuntimeError:
      #_log.warning(name+' already in HDF5 file')
      o=g[name]
    o.attrs['NX_class'] = cls
    for (k,v) in kwargs.iteritems(): o.attrs[k] = v
    pass

  def GetSpecVal(self,k):
    specLst=self.specLst
    vLst=list()
    for s in specLst:
      try: vLst.append(s.attrs[k])
      except KeyError:
        try: vLst.append(s[k]) #seek in dataset
        except ValueError: pass
    if not vLst:
      _log.warning('no ' +k+ ' in SPEC')
      v=np.NaN
    elif len(vLst)==1:
      v=vLst[0]
    else:
      v=np.array(vLst)
    return v

  def GetSpecDict(self,*kw):
    specDict=dict() #some common spec values used in calculation, if not found, set to np.NaN
    for k in kw:
      v=self.GetSpecVal(k)
      specDict[k]=v
    return specDict

  def AddStructure(self):
    fid=self.fid
    specDict=self.GetSpecDict('samz','mokev','ebtrz','ftrans','date','scan')
    if np.isnan(specDict['samz']):
      specDict['samz']=0.
    d=specDict['date']
    if type(d)==np.ndarray:
      d=d[0]
    d=d.replace('_','T')
    specDict['date']=d

    with HdfStruct(fid,'entry','NXentry') as gid:
      self.NXObj(gid,'title',specDict['scan'])
      self.NXObj(gid,'start_time',specDict['date'],cls='NX_DATE_TIME')
      self.NXObj(gid,'end_time','???',cls='NX_DATE_TIME')
      self.NXObj(gid,'definition','NXsas')
      with HdfStruct(gid,'instrument','NXinstrument') as gid0:
        self.NXObj(gid0,'name','cSAXS beamline')
        with HdfStruct(gid0,'source','NXsource') as gid1:
          d=dict(sigma_x=0.202,sigma_y=0.018,divergence_x=0.000135,divergence_y=0.000025)
          self.NXObj(gid1,'type','Synchrotron X-ray Source')
          self.NXObj(gid1,'name','Swiss Light Source')
          self.NXObj(gid1,'probe','x-ray')
          self.NXObj(gid1,'distance', -33800.-specDict['samz'],NX_LENGTH='mm')
          self.NXObj(gid1,'sigma_x',d['sigma_x'],NX_LENGTH='mm')
          self.NXObj(gid1,'sigma_y',d['sigma_y'],NX_LENGTH='mm')
          self.NXObj(gid1,'divergence_x',d['divergence_x'],NX_ANGLE='radians')
          self.NXObj(gid1,'divergence_y',d['divergence_y'],NX_ANGLE='radians')
          #self.NXObj(gid1,'emittance_x',d['sigma_x']*d['divergence_x']*1e6,NX_EMITTANCE='nm-rad')
          #self.NXObj(gid1,'emittance_y',d['sigma_x']*d['divergence_x']*1e6,NX_EMITTANCE='nm-rad')
          self.NXObj(gid1,'current',spec='curr',NX_CURRENT='mA')
          #self.NXObj(gid1,'frequency',np.NaN)
          #self.NXObj(gid1,'number_of_bunches',np.NaN)
          #self.NXObj(gid1,'bunch_length',np.NaN)
          #self.NXObj(gid1,'bunch_distance',np.NaN)
          #self.NXObj(gid1,'pulse_width',np.NaN)
          #self.NXObj(gid1,'mode','???')
          #self.NXObj(gid1,'top_up',True) #???
          #self.NXObj(gid1,'notes','???')
        with HdfStruct(gid0,'insertion_device','NXinsertion_device') as gid1:
          self.NXObj(gid1,'type','undulator')
          self.NXObj(gid1,'gap',spec='idgap',NX_LENGTH='mm')
          #self.NXObj(gid1,'taper',np.NaN,NX_ANGLE='radians')
          #self.NXObj(gid1,'phase',np.NaN,NX_ANGLE='radians')
          #self.NXObj(gid1,'poles',np.NaN,NX_UNITLESS=True)
          #self.NXObj(gid1,'magnetic_wavelength',np.NaN,cls='NX_WAVELENGTH')
          self.NXObj(gid1,'k',2.46,NX_DIMENSIONLESS=True)
          self.NXObj(gid1,'length',1820.,NX_LENGTH='mm')
          self.NXObj(gid1,'harmonic',spec='harmonic',cls='NX_INT',NX_UNITLESS=True)
        with HdfStruct(gid0,'slit_0','NXslit') as gid1:
          self.NXObj(gid1,'material','OFHC Cu')
          self.NXObj(gid1,'description','Horizontal secondary source slit')
          #self.NXObj(gid1,'depends_on','???')
          self.NXObj(gid1,'x_gap',spec='sl0wh',NX_LENGTH='mm')
          #self.NXObj(gid1,'y_gap',spec='sl0wv',NX_LENGTH='mm')
          self.NXObj(gid1,'distance', -33800.+12100.-specDict['samz'],NX_LENGTH='mm')
          self.NXObj(gid1,'x_translation',spec='sl0ch',NX_LENGTH='mm')
        with HdfStruct(gid0,'slit_1','NXslit') as gid1:
          self.NXObj(gid1,'material','OFHC Cu')
          self.NXObj(gid1,'description','Slit 1, optics hutch')
          #self.NXObj(gid1,'depends_on','???')
          self.NXObj(gid1,'x_gap',spec='sl1wh',NX_LENGTH='mm')
          self.NXObj(gid1,'y_gap',spec='sl1wv',NX_LENGTH='mm')
          self.NXObj(gid1,'distance', -33800.+26000.-specDict['samz'],NX_LENGTH='mm')
          self.NXObj(gid1,'x_translation',spec='sl1ch',NX_LENGTH='mm')
          self.NXObj(gid1,'height',spec='sl1cv',NX_LENGTH='mm')
        with HdfStruct(gid0,'monochromator','NXmonochromator') as gid1:
          self.NXObj(gid1,'wavelength',12.3984193/specDict['mokev'],NX_WAVELENGTH='Angstrom')
          self.NXObj(gid1,'wavelength_spread',0.0001)
          self.NXObj(gid1,'energy',spec='mokev',NX_ENERGY='keV')
          self.NXObj(gid1,'type','Double crystal fixed exit monochromator',cls='NXnote')
          self.NXObj(gid1,'distance', -33800.+ 28580.-specDict['samz'],NX_LENGTH='mm')
          # crystals
          with HdfStruct(gid1,'crystal_1', 'NXcrystal') as gid2:
            self.NXObj(gid2,'usage','Bragg')
            self.NXObj(gid2,'type','Si')
            self.NXObj(gid2,'order_no',1)
            self.NXObj(gid2,'reflection',np.array([1,1,1]),cls='NX_INT',NX_UNITLESS = '')
            self.NXObj(gid2,'bragg_angle',spec='moth1',NX_ANGLE = 'degrees')
            self.NXObj(gid2,'temperature',spec='temp_mono_cryst_1')
          with HdfStruct(gid1,'crystal_2', 'NXcrystal') as gid2:
            self.NXObj(gid2,'usage','Bragg')
            self.NXObj(gid2,'type','Si')
            self.NXObj(gid2,'order_no',2)
            self.NXObj(gid2,'reflection',np.array([1,1,1]),cls='NX_INT',NX_UNITLESS = '')
            self.NXObj(gid2,'bragg_angle',spec='moth2',NX_ANGLE = 'degrees')
            self.NXObj(gid2,'temperature',spec='temp_mono_cryst_2')
            self.NXObj(gid2,'bend_x',spec='mobd',NX_ANY='')
        with HdfStruct(gid0,'XBPM4','NXdetector_group') as gid1:
          self.NXObj(gid1,'group_names','XBPM4, XBPM4_sum, XBPM4_x, XBPM4_y, XBPM4_skew')
          self.NXObj(gid1,'group_index',2)
          self.NXObj(gid1,'group_parent',np.array([-1,1,1,1,1],dtype=np.int8))
          with HdfStruct(gid1,'XBPM4','NX_detector') as gid2:
            self.NXObj(gid2,'distance',np.NaN)
            self.NXObj(gid2,'description','Four quadrant x-ray beam position monitor')
            #self.NXObj(gid2,'dead_time',np.NaN)
            self.NXObj(gid2,'type','???')
            self.NXObj(gid2,'calibration_date','???')
            self.NXObj(gid2,'count_time',spec='sec',cls='NX_NUMBER', NX_TIME = 's')
            #self.NXObj(gid2,'acquisition_mode','???')
            #self.NXObj(gid2,'detector_readout_time',np.NaN)
            #self.NXObj(gid2,'trigger_delay_time',np.NaN)
            #self.NXObj(gid2,'trigger_dead_time',np.NaN)
            #self.NXObj(gid2,'frame_time',np.NaN)
            self.NXObj(gid2,'gain_setting',spec='bpm4_gain_setting')
            #self.NXObj(gid2,'calibration_method','???',cls='NXnote')
          with HdfStruct(gid1,'XBPM4_sum','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm4s')
            self.NXObj(gid2,'description','Sum of counts for the four quadrants')
            self.NXObj(gid2,'saturation_value',spec='bpm4_saturation_value')
          with HdfStruct(gid1,'XBPM4_x','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm4x')
            self.NXObj(gid2,'description','Normalized difference of counts between left and right quadrants')
          with HdfStruct(gid1,'XBPM4_y','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm4y')
            self.NXObj(gid2,'description','Normalized difference of counts between high and low quadrants')
          with HdfStruct(gid1,'XBPM4_skew','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm4z')
            self.NXObj(gid2,'description','Normalized difference of counts between diagonal quadrants')
        with HdfStruct(gid0,'mirror','NXmirror') as gid1:
          self.NXObj(gid1,'type','single')
          self.NXObj(gid1,'description','Grazing incidence mirror to reject high-harmonic wavelengths from the monochromator. There are three coating options available that are used depending on the X-ray energy, no coating (SiO2), rhodium (Rh) or platinum (Pt).')
          self.NXObj(gid1,'incident_angle',spec='mith',NX_ANGLE='degrees')
          #self.NXObj(gid1,'external_material','???')
          self.NXObj(gid1,'substrate_material','SiO2')
          self.NXObj(gid1,'coating_material',spec='mirror_coating')
          #self.NXObj(gid1,'reflectivity','???')
          self.NXObj(gid1,'bend_y',spec='mibd',NX_ANY='')
          self.NXObj(gid1,'distance', -33800.+ 29430.-specDict['samz'],NX_LENGTH='mm')
        with HdfStruct(gid0,'XBPM5','NXdetector_group') as gid1:
          self.NXObj(gid1,'group_names','XBPM5, XBPM5_sum, XBPM5_x, XBPM5_y, XBPM5_skew')
          self.NXObj(gid1,'group_index',3)
          self.NXObj(gid1,'group_parent',np.array([-1,1,1,1,1],dtype=np.int8))
          with HdfStruct(gid1,'XBPM5','NXdetector') as gid2:
            self.NXObj(gid2,'distance',np.NaN)
            self.NXObj(gid2,'description','Four quadrant x-ray beam position monitor')
            #self.NXObj(gid2,'dead_time',np.NaN)
            self.NXObj(gid2,'type','???')
            self.NXObj(gid2,'calibration_date','???')
            self.NXObj(gid2,'count_time',spec='sec',cls='NX_NUMBER', NX_TIME = 's')
            #self.NXObj(gid2,'acquisition_mode','???')
            #self.NXObj(gid2,'detector_readout_time',np.NaN)
            #self.NXObj(gid2,'trigger_delay_time',np.NaN)
            #self.NXObj(gid2,'trigger_dead_time',np.NaN)
            #self.NXObj(gid2,'frame_time',np.NaN)
            self.NXObj(gid2,'gain_setting',spec='bpm5_gain_setting')
            #self.NXObj(gid2,'calibration_method','???',cls='NXnote')
          with HdfStruct(gid1,'XBPM5_sum','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm5s')
            self.NXObj(gid2,'description','Sum of counts for the four quadrants')
            self.NXObj(gid2,'saturation_value',spec='bpm5_saturation_value')
          with HdfStruct(gid1,'XBPM5_x','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm5x')
            self.NXObj(gid2,'description','Normalized difference of counts between left and right quadrants')
          with HdfStruct(gid1,'XBPM5_y','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm5y')
            self.NXObj(gid2,'description','Normalized difference of counts between high and low quadrants')
          with HdfStruct(gid1,'XBPM5_skew','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm5z')
            self.NXObj(gid2,'description','Normalized difference of counts between diagonal quadrants')
        with HdfStruct(gid0,'slit_2','NXslit') as gid1:
          self.NXObj(gid1,'material','Ag')
          self.NXObj(gid1,'description','Slit 2, optics hutch')
          #self.NXObj(gid1,'depends_on','???')
          self.NXObj(gid1,'x_gap',spec='sl2wh',NX_LENGTH='mm')
          self.NXObj(gid1,'y_gap',spec='sl2wv',NX_LENGTH='mm')
          self.NXObj(gid1,'distance', -34130.+30660.-specDict['samz'],NX_LENGTH='mm')
          self.NXObj(gid1,'x_translation',spec='sl2ch',NX_LENGTH='mm')
          self.NXObj(gid1,'height',spec='sl2cv',NX_LENGTH='mm')
        with HdfStruct(gid0,'XBPM6','NXdetector_group') as gid1:
          self.NXObj(gid1,'group_names','XBPM6, XBPM6_sum, XBPM6_x, XBPM6_y, XBPM6_skew')
          self.NXObj(gid1,'group_index',4)
          self.NXObj(gid1,'group_parent',np.array([-1,1,1,1,1],dtype=np.int8))
          with HdfStruct(gid1,'XBPM6','NXdetector') as gid2:
            self.NXObj(gid2,'distance',np.NaN)
            self.NXObj(gid2,'description','Four quadrant x-ray beam position monitor')
            #self.NXObj(gid2,'dead_time',np.NaN)
            self.NXObj(gid2,'type','???')
            self.NXObj(gid2,'calibration_date','???')
            self.NXObj(gid2,'count_time',spec='sec',cls='NX_NUMBER', NX_TIME = 's')
            #self.NXObj(gid2,'acquisition_mode','???')
            #self.NXObj(gid2,'detector_readout_time',np.NaN)
            #self.NXObj(gid2,'trigger_delay_time',np.NaN)
            #self.NXObj(gid2,'trigger_dead_time',np.NaN)
            #self.NXObj(gid2,'frame_time',np.NaN)
            self.NXObj(gid2,'gain_setting',spec='bpm6_gain_setting')
            #self.NXObj(gid2,'calibration_method','???',cls='NXnote')
          with HdfStruct(gid1,'XBPM6_sum','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm6s')
            self.NXObj(gid2,'description','Sum of counts for the four quadrants')
            self.NXObj(gid2,'saturation_value',spec='bpm6_saturation_value')
          with HdfStruct(gid1,'XBPM6_x','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm6x')
            self.NXObj(gid2,'description','Normalized difference of counts between left and right quadrants')
          with HdfStruct(gid1,'XBPM6_y','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm6y')
            self.NXObj(gid2,'description','Normalized difference of counts between high and low quadrants')
          with HdfStruct(gid1,'XBPM6_skew','NX_detector') as gid2:
            self.NXObj(gid2,'data',spec='bpm6z')
            self.NXObj(gid2,'description','Normalized difference of counts between diagonal quadrants')
        with HdfStruct(gid0,'slit_3','NXslit') as gid1:
          self.NXObj(gid1,'material','Si')
          self.NXObj(gid1,'description','Slit 3, experimental hutch, exposure box')
          #self.NXObj(gid1,'depends_on','???')
          self.NXObj(gid1,'x_gap',spec='sl3wh',NX_LENGTH='mm')
          self.NXObj(gid1,'y_gap',spec='sl3wv',NX_LENGTH='mm')
          self.NXObj(gid1,'distance', 0.0-specDict['samz'],NX_LENGTH='mm') #???+specDict['ebtrz']
          self.NXObj(gid1,'x_translation',spec='sl3ch',NX_LENGTH='mm')
          self.NXObj(gid1,'height',spec='sl3cv',NX_LENGTH='mm')
        with HdfStruct(gid0,'filter_set','NXattenuator') as gid1:
          self.NXObj(gid1,'distance', 0.0-specDict['samz'],NX_LENGTH='mm') #???
          self.NXObj(gid1,'description','The filter set consists of 4 linear stages, each with five filter positions. Additionally each one allows for an out position to allow no filtering.')
          self.NXObj(gid1,'type',spec='fil_comb_description')
          #self.NXObj(gid1,'thickness',np.NaN)
          self.NXObj(gid1,'attenuator_transmission',10.**specDict['ftrans'])
        with HdfStruct(gid0,'slit_4','NXslit') as gid1:
          self.NXObj(gid1,'material','Si')
          self.NXObj(gid1,'description','Slit 4, experimental hutch, exposure box')
          self.NXObj(gid1,'x_gap',spec='sl4wh',NX_LENGTH='mm')
          self.NXObj(gid1,'y_gap',spec='sl4wv',NX_LENGTH='mm')
          self.NXObj(gid1,'distance', 0.0-specDict['samz'],NX_LENGTH='mm') #???
          self.NXObj(gid1,'x_translation',spec='sl4ch',NX_LENGTH='mm')
          self.NXObj(gid1,'height',spec='sl4cv',NX_LENGTH='mm')
        with HdfStruct(gid0,'beam_stop_1','NX_beamstop') as gid1:
          self.NXObj(gid1,'description','circular')
          self.NXObj(gid1,'size',3.,NX_LENGTH='mm')
          self.NXObj(gid1,'x',spec='bs1x',NX_LENGTH='mm')
          self.NXObj(gid1,'y',spec='bs1y',NX_LENGTH='mm')
          self.NXObj(gid1,'distance_to_detector',spec='bs1_det_dist',NX_LENGTH='mm')
          self.NXObj(gid1,'status',spec='bs1_status')
        with HdfStruct(gid0,'beam_stop_2','NX_beamstop') as gid1:
          self.NXObj(gid1,'description','rectangular')
          self.NXObj(gid1,'size_x',5.00,NX_LENGTH='mm')
          self.NXObj(gid1,'size_y',2.25,NX_LENGTH='mm')
          self.NXObj(gid1,'x',spec='bs2x')
          self.NXObj(gid1,'y',spec='bs2y')
          self.NXObj(gid1,'distance_to_detector',spec='bs2_det_dist',NX_LENGTH='mm')
          self.NXObj(gid1,'status',spec='bs2_status')
          self.NXObj(gid1,'data',spec='diode')
        self.AddPilatus2M(gid0)
        self.AddPilatus300K(gid0)
        self.AddEiger(gid0)
        #self.AddAllDetectors(gid0)
      with HdfStruct(gid,'sample','NXsample') as gid0:
        self.NXObj(gid0,'name',spec='sample_name')
        self.NXObj(gid0,'aequatorial_angle',0.,NX_ANGLE='deg')
        self.NXObj(gid0,'description',spec='sample_description')
        #self.NXObj(gid0,'myfloat',spec='sample_float',cast=float)
        self.NXObj(gid0,'x_translation',spec='samx',NX_LENGTH='mm')
        self.NXObj(gid0,'y_translation',spec='samy',NX_LENGTH='mm')
        #self.NXObj(gid0,'distance',spec='samz',NX_LENGTH='mm')
        self.NXObj(gid0,'temperature_log',spec='temp')
      with HdfStruct(gid,'collection','NXcollection') as gid0:
        #All spec motor values and counters are stored here. For the counters there can be links to the data below.
        pass
      with HdfStruct(gid,'control','NXmonitor') as gid0:
        self.NXObj(gid0,'mode','monitor')
        #self.NXObj(gid0,'preset',np.NaN)
        self.NXObj(gid0,'integral',spec='bpm4s')
        self.NXObj(gid0,'note','Value from bpm4s',cls='NXnote')
    pass

  def AddPilatus2M(self,gid):
    fid=self.fid
    detKey='pilatus_1'
    name='Pilatus_2M'
    try:
      ds=h5py.Group(gid)[os.path.join(self.basePath,detKey)]
    except KeyError:
      _log.warning('no ' +name+ ' found in hdf5 file')
      return
    with HdfStruct(gid,name,'NXdetector') as gid0:
      try:
        h5py.Group(gid0)['data']=ds
      except RuntimeError:
        _log.warning(name+' already in HDF5 file')
      else:
        _log.info(name+' adding hard link in HDF5 file')

      #self.NXObj(gid0,'data',ds) # hard link in hdf5 file
      #self.NXObj(gid0,'data','data')
      # link to '/entry/data/'+detKey
      self.NXObj(gid0,'distance',np.NaN,NX_LENGTH='mm') 	# From SPEC detector geometry
      self.NXObj(gid0,'x_pixel_size',172.,NX_LENGTH='um')
      self.NXObj(gid0,'y_pixel_size',172.,NX_LENGTH='um')
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')
      self.NXObj(gid0,'azimuthal_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'rotation_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')		# From SPEC detector geometry
      self.NXObj(gid0,'aequatorial_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'beam_center_x',np.NaN,NX_LENGTH='um')	# From SPEC detector center
      self.NXObj(gid0,'beam_center_y',np.NaN,NX_LENGTH='um')	# From SPEC detector center
      self.NXObj(gid0,'description','Pilatus detector, in house developed, Paul Scherrer Institut. Cite [http://dx.doi.org/10.1016/j.nima.2009.03.200]')
      self.NXObj(gid0,'type','Single-photon counting detector, 320 micron-thick Si chip')
      #self.NXObj(gid0,'calibration_date','???',cls='NX_DATE_TIME')
      self.NXObj(gid0,'count_time',spec='sec',cls='NX_TIME',NX_TIME='s')
      #self.NXObj(gid0,'acquisition_mode','gated triggered')
      #self.NXObj(gid0,'pixel_mask','???') # don't know were to get it from
      self.NXObj(gid0,'countrate_correction_applied',True)
      #self.NXObj(gid0,'bit_depth_readout',32,cls='NX_INT') #TODO: read bit depth from data type
      #self.NXObj(gid0,'detector_readout_time',np.NaN)
      #self.NXObj(gid0,'trigger_delay_time',np.NaN)
      #self.NXObj(gid0,'trigger_dead_time',np.NaN)
      #self.NXObj(gid0,'frame_time',np.NaN)
      self.NXObj(gid0,'gain_setting',ds.attrs.get('Gain_setting',np.NaN))	# From the attributes of the dataset
      self.NXObj(gid0,'saturation_value',ds.attrs.get('Count_cutoff',np.NaN),cls='NX_INT')
      self.NXObj(gid0,'threshold_energy',ds.attrs.get('Threshold_setting',np.NaN),NX_ENERGY=ds.attrs.get('_u_Threshold_setting',np.NaN))
      #self.NXObj(gid0,'efficiency',np.NaN,cls='NXdata')
      self.NXObj(gid0,'geometry_note','How was the position, angle and beam determined',cls='NXnote')	# From SPEC geometry note
    pass

  def AddPilatus300K(self,gid):
    fid=self.fid
    detKey='pilatus_2'
    name='Pilatus_300K'
    try:
      ds=h5py.Group(gid)[os.path.join(self.basePath,detKey)]
    except KeyError:
      _log.warning('no ' +name+ ' found in hdf5 file')
      return
    with HdfStruct(gid,name,'NXdetector') as gid0:
      try:
        h5py.Group(gid0)['data']=ds
      except RuntimeError:
        _log.warning(name+' already in HDF5 file')
      else:
        _log.info(name+' adding hard link in HDF5 file')

      #self.NXObj(gid0,'data',ds) # hard link in hdf5 file
      #self.NXObj(gid0,'data','data')
      # link to '/entry/data/'+detKey
      self.NXObj(gid0,'distance',np.NaN,NX_LENGTH='mm') 	# From SPEC detector geometry
      self.NXObj(gid0,'x_pixel_size',172.,NX_LENGTH='um')
      self.NXObj(gid0,'y_pixel_size',172.,NX_LENGTH='um')
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')		# From SPEC detector geometry
      self.NXObj(gid0,'azimuthal_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'rotation_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')		# From SPEC detector geometry
      self.NXObj(gid0,'aequatorial_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'beam_center_x',np.NaN,NX_LENGTH='um')	# From SPEC detector center
      self.NXObj(gid0,'beam_center_y',np.NaN,NX_LENGTH='um')	# From SPEC detector center
      self.NXObj(gid0,'description','Dectris Pilatus detector')
      self.NXObj(gid0,'type','Single-photon counting detector, 320 micron-thick Si chip')
      #self.NXObj(gid0,'calibration_date','???',cls='NX_DATE_TIME')
      self.NXObj(gid0,'count_time',spec='sec',cls='NX_NUMBER',NX_TIME='s')
      #self.NXObj(gid0,'acquisition_mode','gated triggered')
      #self.NXObj(gid0,'pixel_mask','???') # don't know were to get it from
      self.NXObj(gid0,'countrate_correction_applied',True)
      #self.NXObj(gid0,'bit_depth_readout',32,cls='NX_INT') #TODO: read bit depth from data type
      #self.NXObj(gid0,'detector_readout_time',np.NaN)
      #self.NXObj(gid0,'trigger_delay_time',np.NaN)
      #self.NXObj(gid0,'trigger_dead_time',np.NaN)
      #self.NXObj(gid0,'frame_time',np.NaN)
      self.NXObj(gid0,'gain_setting',ds.attrs.get('Gain_setting',np.NaN))	# From the attributes of the dataset
      self.NXObj(gid0,'saturation_value',ds.attrs.get('Count_cutoff',np.NaN),cls='NX_INT')
      self.NXObj(gid0,'threshold_energy',ds.attrs.get('Threshold_setting',np.NaN),NX_ENERGY=ds.attrs.get('_u_Threshold_setting',np.NaN))
      #self.NXObj(gid0,'efficiency',np.NaN,cls='NXdata')
      self.NXObj(gid0,'geometry_note','How was the position, angle and beam determined',cls='NXnote')	# Different detector support needed
    pass

  def AddEiger(self,gid):
    fid=self.fid
    detKey='eiger'
    name='Eiger_500k'
    try:
      ds=h5py.Group(gid)[os.path.join(self.basePath,detKey)]
    except KeyError:
      _log.warning('no ' +name+ ' found in hdf5 file')
      return
    with HdfStruct(gid,name,'NXdetector') as gid0:
      try:
        h5py.Group(gid0)['data']=ds
      except RuntimeError:
        _log.warning(name+' already in HDF5 file')
      else:
        _log.info(name+' adding hard link in HDF5 file')

      #self.NXObj(gid0,'data',ds) # hard link in hdf5 file
      #self.NXObj(gid0,'data','data')
      # link to '/entry/data/'+detKey
      self.NXObj(gid0,'distance',np.NaN,NX_LENGTH='mm') 	# From SPEC detector geometry
      self.NXObj(gid0,'x_pixel_size',75.,NX_LENGTH='um')
      self.NXObj(gid0,'y_pixel_size',75.,NX_LENGTH='um')
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')		# From SPEC detector geometry
      self.NXObj(gid0,'azimuthal_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'rotation_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')		# From SPEC detector geometry
      self.NXObj(gid0,'aequatorial_angle',0.,NX_ANGLE='deg')	# From SPEC detector geometry
      self.NXObj(gid0,'beam_center_x',np.NaN,NX_LENGTH='um')	# From SPEC detector center
      self.NXObj(gid0,'beam_center_y',np.NaN,NX_LENGTH='um')	# From SPEC detector center
      self.NXObj(gid0,'description','Eiger detector, in house developed, Paul Scherrer Institut. Cite [doi:10.1016/j.nima.2010.12.005]')
      self.NXObj(gid0,'type','Single-photon counting detector, 300 micron-thick Si chip')
      #self.NXObj(gid0,'calibration_date','???',cls='NX_DATE_TIME')
      self.NXObj(gid0,'count_time',spec='sec',cls='NX_NUMBER',NX_TIME='s')
      #self.NXObj(gid0,'acquisition_mode','gated triggered')
      #self.NXObj(gid0,'pixel_mask','???') # don't know were to get it from
      self.NXObj(gid0,'countrate_correction_applied',True)
      #self.NXObj(gid0,'bit_depth_readout',32,cls='NX_INT') #TODO: read bit depth from data type
      #self.NXObj(gid0,'detector_readout_time',np.NaN)
      #self.NXObj(gid0,'trigger_delay_time',np.NaN)
      #self.NXObj(gid0,'trigger_dead_time',np.NaN)
      #self.NXObj(gid0,'frame_time',np.NaN)
      self.NXObj(gid0,'gain_setting','???')			# Different detector support needed
      self.NXObj(gid0,'saturation_value',np.NaN,cls='NX_INT')
      self.NXObj(gid0,'threshold_energy',np.NaN)		# Different detector support needed pilatus_1='Threshold_setting'
      #self.NXObj(gid0,'efficiency',np.NaN,cls='NXdata')
      self.NXObj(gid0,'geometry_note','How was the position, angle and beam determined',cls='NXnote')	# Different detector support needed
    pass


  def AddAllDetectors(self,gid):
    detDict={
      'pilatus_1':dict(name='Pilatus_2M',
                       descr='Pilatus detector, in house developed, Paul Scherrer Institut. Cite [http://dx.doi.org/10.1016/j.nima.2009.03.200]',
                       pixSz=(172.,172.),
                       ),
      'pilatus_2':dict(name='Pilatus_300K',
                       descr='Dectris Pilatus detector',
                       pixSz=(172.,172.),
                       ),
      'eiger':    dict(name='Eiger_500k',
                       descr='Eiger detector, in house developed, Paul Scherrer Institut. Cite [doi:10.1016/j.nima.2010.12.005]',
                       pixSz=(75.,75.),
                      ),
    }
    for k,v in detDict.iteritems():
      self.AddDetector(gid,k,v)

  def AddDetector(self,gid,detKey,detDict):
    fid=self.fid
    #pilatus_1 PILATUS 2M
    #pilatus_2 PILATUS 300K
    #eiger     EIGER 500K
    try:
      ds=h5py.Group(gid)[os.path.join(self.basePath,detKey)]
    except KeyError:
      _log.warning('no ' +detDict['name']+ ' found in hdf5 file')
      return


    with HdfStruct(gid,detDict['name'],'NXdetector') as gid0:
      try:
        h5py.Group(gid0)['data']=ds
      except RuntimeError:
        _log.warning(detDict['name']+' already in HDF5 file')
      else:
        _log.info(detDict['name']+' adding hard link in HDF5 file')

      #self.NXObj(gid0,'data',ds) # hard link in hdf5 file
      #self.NXObj(gid0,'data','data')
      # link to '/entry/data/'+detKey
      self.NXObj(gid0,'distance',np.NaN,NX_LENGTH='mm') 	# Different detector support needed
      #self.NXObj(gid0,'distance',spec = 'pilatus_1_dist',NX_LENGTH='mm') # Some value from spec, this worked but is the same for all
      self.NXObj(gid0,'x_pixel_size',detDict['pixSz'][0],NX_LENGTH='um')
      self.NXObj(gid0,'y_pixel_size',detDict['pixSz'][1],NX_LENGTH='um')
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')
      self.NXObj(gid0,'azimuthal_angle',0.,NX_ANGLE='deg')
      self.NXObj(gid0,'rotation_angle',0.,NX_ANGLE='deg')	# Different detector support needed
      self.NXObj(gid0,'polar_angle',0.,NX_ANGLE='deg')
      self.NXObj(gid0,'beam_center_x',np.NaN,NX_LENGTH='um')	# Different detector support needed
      self.NXObj(gid0,'beam_center_y',np.NaN,NX_LENGTH='um')	# Different detector support needed
      self.NXObj(gid0,'description',detDict['descr'])
      self.NXObj(gid0,'type','Single-photon counting detector, 300 micron-thick Si chip')
      #self.NXObj(gid0,'calibration_date','???',cls='NX_DATE_TIME')
      self.NXObj(gid0,'count_time',spec='sec',cls='NX_NUMBER',NX_TIME='s')
      #self.NXObj(gid0,'acquisition_mode','gated triggered')
      #self.NXObj(gid0,'pixel_mask','???') # don't know were to get it from
      self.NXObj(gid0,'countrate_correction_applied',True)
      #self.NXObj(gid0,'bit_depth_readout',32,cls='NX_INT') #TODO: read bit depth from data type
      #self.NXObj(gid0,'detector_readout_time',np.NaN)
      #self.NXObj(gid0,'trigger_delay_time',np.NaN)
      #self.NXObj(gid0,'trigger_dead_time',np.NaN)
      #self.NXObj(gid0,'frame_time',np.NaN)
      self.NXObj(gid0,'gain_setting','???')			# Different detector support needed
      self.NXObj(gid0,'saturation_value',np.NaN,cls='NX_INT')
      self.NXObj(gid0,'threshold_energy',np.NaN)		# Different detector support needed pilatus_1='Threshold_setting'
      #self.NXObj(gid0,'efficiency',np.NaN,cls='NXdata')
      self.NXObj(gid0,'geometry_note','How was the position, angle and beam determined',cls='NXnote')	# Different detector support needed
    pass
