#!/usr/bin/python

# *-----------------------------------------------------------------------*
# |                                                                       |
# |  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
# |                                                                       |
# |              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
# *-----------------------------------------------------------------------*
'''
libDetXR wrapper class to various C/C++ implemented functions

here a not complete list of the most important functions:

  - Ptr2Arr(data,shape,type_)
  - CompressCBF(inBuf, inNumBytes, outBuf, outNumBytes,elemSz)
  - DecompressCBF(inBuf, inNumBytes, outBuf, outNumBytes,elemSz)
  - CmpBufSzZLIB(inNumBytes)
  - CompressZLIB(inBuf, inNumBytes, outBuf, outNumBytes, level)
  - DecompressZLIB(inBuf, inNumBytes, outBuf, outNumBytes)
  - CmpBufSzLZF(inNumBytes)
  - CompressLZF(inBuf, inNumBytes, outBuf, outNumBytes)
  - DecompressLZF(inBuf, inNumBytes, outBuf, outNumBytes)
  - CmpBufSzLZ4(inNumBytes)
  - CompressLZ4(inBuf, inNumBytes, outBuf, outNumBytes)
  - DecompressLZ4(inBuf, inNumBytes, outBuf, outNumBytes)
  - CompressZRLE(inBuf, inNumBytes, outBuf, outNumBytes, mode)
  - ByteShuffle(inBuf, outBuf, numBytes, elemSize)
  - ProcessRoiStatistics(data, roiIdxArr, wgtArr, roiIdxSz, roiLenArr, roiLenSz, resArr, mode)
  - ProcessMoments(mode, data, width, height, hOfs, resArr)
  - ProcessPorodGuinier(pCnt, pFirst, gCnt, gFirst, q, Iq, IqErr, qCnt, res)
  - Shuffle32Bit(src, dst, numByte)
  - Convert12b16b(src, dst, numPix)
  - Convert12b16bOfs(src, dst, width, height, lnPitch)
  - Convert4b8b(src, dst, numPix)
'''
import ctypes as ct, numpy as np
#import warnings as wrn
import sys,os
import logging

_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')

class LibDetXR(object):
  #-----------------------------------
  #static class variables, definitions
  #-----------------------------------
  @staticmethod
  def Ptr2Arr(data,shape,type_):
    """converts pointer data to a numpyarray with shape 'shape' and of type 'type_'"""
    buf = np.core.multiarray.int_asbuffer(data, np.prod(shape)*ct.sizeof(type_))
    data = np.frombuffer(buf, type_)
    data=data.reshape(shape,order='C')
    return data

  @classmethod
  def LoadLib(cls):
    pf=sys.platform
    if pf=="win32":
      pathBase=os.path.dirname(__file__)
      path=['..','../../../Debug/bin','../../../Release/bin']
      for p in path:
        fn=os.path.abspath(pathBase+'/'+p+r'/LibDetXR.dll')
        try:
          cls.lib=ct.cdll.LoadLibrary(fn) #for Windows
        except WindowsError as err:
          _log.debug(fn+' failed')
        else:
          _log.debug(fn+' loaded')
          break
      if not hasattr(cls,'lib'):
        raise WindowsError('failed to load LibDetXR.dll')
    elif pf=='linux2':
        pathBase=os.path.dirname(__file__)
        path=['.','../../src/release','../../src/debug']
        libName='libDetXR.so.0'
        for p in path:
          fn=os.path.abspath(os.path.join(pathBase,p,libName))
          try:
            cls.lib=ct.cdll.LoadLibrary(fn) #for Linux
          except OSError as err:
            _log.debug(fn+' failed')
          else:
            _log.debug(fn+' loaded')
            break
        if not hasattr(cls,'lib'):
          raise OSError('failed to load '+libName)
    elif pf=='darwin': #MacOS
        try:
            fn='/usr/lib/libDetXR.dylib'
            #fn='../../MacOS/build/libhelic.dylib'
            cls.lib=ct.cdll.LoadLibrary(fn) #for Linux
        except IOError as err:
            _log.critical(fn+' failed')
        else:
            _log.debug(fn+' loaded')
    else:
      raise IOError('unsupported platform '+str(pf))

    cls.lib.DT_SetCallback.restype=cls.funcCBType
    cls.lib.DT_GetDefaultCallback.restype=cls.funcCBType
    cls.lib.DT_CompressCBF.argtypes   = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32, ct.c_uint32)
    cls.lib.DT_DecompressCBF.argtypes = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32, ct.c_uint32)
    cls.lib.DT_CompressZLIB.argtypes  = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32, ct.c_uint32)
    cls.lib.DT_CompressZRLE.argtypes  = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32, ct.c_uint32)
    cls.lib.DT_DecompressZLIB.argtypes= (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_CompressLZF.argtypes   = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_DecompressLZF.argtypes = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_CompressLZ4.argtypes   = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_DecompressLZ4.argtypes = (ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_ByteShuffle.argtypes   = (ct.c_void_p, ct.c_void_p, ct.c_uint32, ct.c_uint32)

    cls.lib.DT_ProcessRoiStatistics.argtypes = (ct.c_void_p, ct.c_void_p, ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_ProcessMoments.argtypes = (ct.c_uint32, ct.c_void_p, ct.c_uint32, ct.c_uint32, ct.c_uint32, ct.c_void_p)
    cls.lib.DT_ProcessPorodGuinier.argtypes = (ct.c_int32, ct.c_int32, ct.c_int32, ct.c_int32,ct.c_void_p,ct.c_void_p,ct.c_void_p, ct.c_int32, ct.c_void_p)

    cls.lib.DT_Shuffle32Bit.argtypes     = (ct.c_void_p, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_Convert12b16b.argtypes    = (ct.c_void_p, ct.c_void_p, ct.c_uint32)
    cls.lib.DT_Convert12b16bOfs.argtypes = (ct.c_void_p, ct.c_void_p, ct.c_uint32, ct.c_uint32, ct.c_uint32)
    cls.lib.DT_Convert4b8b.argtypes      = (ct.c_void_p, ct.c_void_p, ct.c_uint32)

  #------------------------------
  #member functions and variables
  #------------------------------
  def __init__ (self):
    if not hasattr(LibDetXR,'lib'):
      LibDetXR.LoadLib()

    self.handle = ct.c_long(0)

  funcCBType = ct.CFUNCTYPE(ct.c_int, ct.c_long, ct.c_int, ct.c_int,ct.c_void_p)

# most of the following code is generated with: c2py.py
# if the libHeLIC.h header changes, please execute c2py.py and make a diff of the output
# to update this file
#-------------------------------- GENERATED c2py.py ----------------------------
  SRMsgKind={
    'MK_DEBUG_STRING':0x0,
    'MK_BOX_INFO':0x1,
    'MK_BOX_WARN':0x2,
    'MK_BOX_ERR':0x3,
  }
  SRMsgCategory={
    'MC_GENERAL':0x0,
    'MC_CAM':0x100,
    'MC_MOTION':0x200,
    'MC_USB':0x300,
    'MC_ETH':0x400,
    'MC_SYSTEM':0x500,
    'MC_FIRMWARE':0x600,
    'MC_FILEIO':0x700,
    'MC_XML':0x800,
    'MC_DLG':0x900,
  }
  @staticmethod
  def CompressCBF(inBuf, inNumBytes, outBuf, outNumBytes,elemSz):
    res=LibDetXR.lib.DT_CompressCBF(inBuf, inNumBytes, outBuf, outNumBytes,elemSz)
    return res
  @staticmethod
  def DecompressCBF(inBuf, inNumBytes, outBuf, outNumBytes,elemSz):
    res=LibDetXR.lib.DT_DecompressCBF(inBuf, inNumBytes, outBuf, outNumBytes,elemSz)
    return res
  @staticmethod
  def CmpBufSzZLIB(inNumBytes):
    res=LibDetXR.lib.DT_CmpBufSzZLIB(inNumBytes)
    return res
  @staticmethod
  def CompressZLIB(inBuf, inNumBytes, outBuf, outNumBytes, level):
    res=LibDetXR.lib.DT_CompressZLIB(inBuf, inNumBytes, outBuf, outNumBytes, level)
    return res
  @staticmethod
  def DecompressZLIB(inBuf, inNumBytes, outBuf, outNumBytes):
    res=LibDetXR.lib.DT_DecompressZLIB(inBuf, inNumBytes, outBuf, outNumBytes)
    return res
  @staticmethod
  def CmpBufSzLZF(inNumBytes):
    res=LibDetXR.lib.DT_CmpBufSzLZF(inNumBytes)
    return res
  @staticmethod
  def CompressLZF(inBuf, inNumBytes, outBuf, outNumBytes):
    res=LibDetXR.lib.DT_CompressLZF(inBuf, inNumBytes, outBuf, outNumBytes)
    return res
  @staticmethod
  def DecompressLZF(inBuf, inNumBytes, outBuf, outNumBytes):
    res=LibDetXR.lib.DT_DecompressLZF(inBuf, inNumBytes, outBuf, outNumBytes)
    return res
  @staticmethod
  def CmpBufSzLZ4(inNumBytes):
    res=LibDetXR.lib.DT_CmpBufSzLZ4(inNumBytes)
    return res
  @staticmethod
  def CompressLZ4(inBuf, inNumBytes, outBuf, outNumBytes):
    res=LibDetXR.lib.DT_CompressLZ4(inBuf, inNumBytes, outBuf, outNumBytes)
    return res
  @staticmethod
  def DecompressLZ4(inBuf, inNumBytes, outBuf, outNumBytes):
    res=LibDetXR.lib.DT_DecompressLZ4(inBuf, inNumBytes, outBuf, outNumBytes)
    return res
  @staticmethod
  def CompressZRLE(inBuf, inNumBytes, outBuf, outNumBytes, mode):
    res=LibDetXR.lib.DT_CompressZRLE(inBuf, inNumBytes, outBuf, outNumBytes, mode)
    return res
  @staticmethod
  def ByteShuffle(inBuf, outBuf, numBytes, elemSize):
    res=LibDetXR.lib.DT_ByteShuffle(inBuf, outBuf, numBytes, elemSize)
    return res
  @staticmethod
  def ProcessRoiStatistics(data, roiIdxArr, wgtArr, roiIdxSz, roiLenArr, roiLenSz, resArr, mode):
    res=LibDetXR.lib.DT_ProcessRoiStatistics(data, roiIdxArr, wgtArr, roiIdxSz, roiLenArr, roiLenSz, resArr, mode)
    return res
  @staticmethod
  def ProcessMoments(mode, data, width, height, hOfs, resArr):
    res=LibDetXR.lib.DT_ProcessMoments(mode, data, width, height, hOfs, resArr)
    return res
  @staticmethod
  def ProcessPorodGuinier(prdStart, prdNum, gnrStart, gnrNum, q, Iq, IqErr, qCnt, res):
    res=LibDetXR.lib.DT_ProcessPorodGuinier(prdStart, prdNum, gnrStart, gnrNum, q, Iq, IqErr, qCnt, res)
    return res
  @staticmethod
  def Shuffle32Bit(src, dst, numByte):
    res=LibDetXR.lib.DT_Shuffle32Bit(src, dst, numByte)
    return res
  @staticmethod
  def Convert12b16b(src, dst, numPix):
    res=LibDetXR.lib.DT_Convert12b16b(src, dst, numPix)
    return res
  @staticmethod
  def Convert12b16bOfs(src, dst, width, height, lnPitch):
    res=LibDetXR.lib.DT_Convert12b16bOfs(src, dst, width, height, lnPitch)
    return res
  @staticmethod
  def Convert4b8b(src, dst, numPix):
    res=LibDetXR.lib.DT_Convert4b8b(src, dst, numPix)
    return res
  @staticmethod
  def SetCallback(cb):
    res=LibDetXR.lib.DT_SetCallback(cb)
    return res
  @staticmethod
  def GetDefaultCallback():
    res=LibDetXR.lib.DT_GetDefaultCallback()
    return res
  @staticmethod
  def CheckForNewDllVersion(mode):
    res=LibDetXR.lib.DT_CheckForNewDllVersion(mode)
    return res
#-------------------------------- GENERATED END --------------------------------
  LZF_FILTER=32000 #HDF5 compression filter
  LZ4_FILTER=32004 #HDF5 compression filter
  def GetCamArr(self,idx):
    cdm=LibDetXR.CamDataMeta()
    cd=self.GetCamData(idx,0,ct.byref(cdm))
    data=cd.contents.data
    if cd.contents.prop&LibDetXR.CamDataProperty['DP_ENCODER']:
      data+=1024*2 #skip 2k header

    shape=tuple(cdm.dimSz[cdm.numDim-1::-1])
    arr=LibDetXR.Ptr2Arr(data,shape,ct.c_ushort)
    return arr

  def Open(self, mode, *args, **kwargs): #kwargs=keyword arguments
    if mode==0:
      s=''
      for k,v in kwargs.iteritems():
        if s:
          s+=' '
        s=s+k+'='+v
      res=self.lib.DT_Open(ct.byref(self.handle),mode,s)
    elif mode==1:
      res=self.lib.DT_Open(ct.byref(self.handle),mode,args[0])
    else:
      res=-1
    return res

  @staticmethod
  def InternalFunc(mode, *args):
    res=LibDetXR.lib.DT_InternalFunc(mode, *args)
    return res

  @staticmethod
  def GetVersion():
    #ver = (ct.c_ushort * 4)()
    #v=ct.cast(ver, ct.POINTER(ct.c_ushort))
    nv=np.ndarray(4,dtype=np.uint16)
    v=ct.cast(nv.ctypes, ct.POINTER(ct.c_ushort))
    res=LibDetXR.lib.DT_GetVersion(v)
    return res,nv

LibDetXR()#force to load the library
if __name__ == '__main__':
  def test(s,res):
    pass
    print s,':',res

  heSys=LibDetXR()
  test('GetVersion SrCam',LibDetXR.GetVersion())

  pass

