#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
Preprocess data for ptychography.
Inputs are:
- data source
- roi (region of interest)

The output will be fft-shifted ROI
'''
import matplotlib as mpl
#if __name__ == '__main__':
#  mpl.use('WXAgg')

import logging, time, json, os, sys
import numpy as np
import h5py, zmq
import multiprocessing as mproc
import utilities as ut
from Clamp_mod import ClampIn,ClampOut,ClampLst
from libDetXR import LibDetXR
from DatParser import OMNIDatParser
from operator import mul
LibDetXR()

_log=logging.getLogger(__name__)

def LogOut(s):
  log=logging.root.handlers[0].stream
  log.write(s)
  log.flush()

class PtychoMeta():
  def Read(self,fnMeta):
    if not os.access(fnMeta, os.F_OK):
      _log.error('generate missing metaFile: %s'%fnMeta)

    fn=open(fnMeta,'r')
    try:
      d=json.load(fn, object_hook=ut.ConvUtf8)
      #self._m=meta=Struct(json.load(fn, object_hook=ut.ConvUtf8))
      self._m=meta=ut.dotdict(d)

    except ValueError as e:
      _log.error(e)
      fn.close()
      return
    fn.close()
    glob=meta.glob
    #~ glob.lambda_ = 1.2398e-9/glob.energy  # wavelength

    for meas in meta.measurement.itervalues():
      roiSz=np.array(meas.roi[2:])
      #~ meas.dxSpec = glob.lambda_*glob.z/(roiSz*glob.ds)# resolution in the specimen plane              resolution in the specimen plane
    pass

  def WriteHdfBaseStructure(self,args,fid):
    meta=self._m
    detRoi=dict() #dictionary for detector rois
    grpRoot=h5py.Group(fid).require_group('measurement')
    for k,v in meta['measurement'].iteritems():
      grpMeas=grpRoot.require_group('n'+k)
      #~ grpMeas.attrs['detector']=v['detector']
      #~ grpMeas.attrs['probe']=v['probe']
      #~ grpMeas.attrs['object']=v['object']

      try:
        roi=v['roi']
      except KeyError: pass
      else:
        cx,cy,w,h=roi
        sx=cx-w/2; sy=cy-h/2
        roi=(slice(sx,sx+w),slice(sy,sy+h))
        try:
          roiOld=detRoi[k]
        except KeyError:
          detRoi[k]=roi
        else:
          assert(roiOld==roi) #if same detector is used, the roi must be the same

      #~ pos=self.GetPos(v,np.uint32)
      #~ grpMeas.create_dataset('positions', data=pos,compression='gzip',shuffle=True)
      #~ prbSz=np.array(v.roi[2:],dtype=np.uint64)


    #~ grpRoot=h5py.Group(fid).require_group('detector')
    #~ for k,v in meta['detector'].iteritems():
      #~ grpDet=grpRoot.create_group('n'+k)
      #~ (module,badPix)=self.GetDetector(v,detRoi[k],np.uint32,args.verbose)
      #~ grpDet.attrs['rows']   =(module[:,2]+module[:,0]).max()-module[:,2].min()
      #~ grpDet.attrs['columns']=(module[:,3]+module[:,1]).max()-module[:,3].min()
      #~ grpDet.create_dataset('modules', data=module,compression='gzip',shuffle=True)
      #~ grpDet.create_dataset('bad_pixels', data=badPix,compression='gzip',shuffle=True)

  def GetPos(self,meas,dtype):
    try:
      posParam=meas['posParam']
    except KeyError:
      try:
        pos=np.array(meas['pos'],dtype=dtype)
      except KeyError:
        _log.error('missing position information in meta file')
        return
    else:
      #print scanParam
      mode=posParam['mode']
      if mode=='omni_opos':
        parser=OMNIDatParser()
        parser.Open(posParam['file'])
        parser.Parse()
        parser.Close()
        d=parser.dataOMNI
        # - sign: Change from object to probe positions
        posReal=np.array([ d['Average_x']*1e-6, d['Average_y']*1e-6]).T


        try:
          trfPos=np.array(posParam['trfPos'])
        except KeyError:
          pass
        else:
          posReal=np.dot(trfPos,posReal.T).T
      #elif mode=='omni_opos_angle':
        #  deltax = (16.06-p.spec_all{ii}.samy)*out_orch.Average_angle*1e-6/13;
        #  deltay = 18.5*out_orch.Average_angle*1e-6/13;
        #  posx = out_orch.Average_x*1e-6 - deltax;
        #  posy = out_orch.Average_y*1e-6 - deltay;
        #  positions_real = [posy posx];
      #% Affine transform on probe positions (this will modify the object
      #% according to the transformation matrix)
      #if isfield(p,'affine_matrix')&&(~isempty(p.affine_matrix))
      #    p.positions_real = p.affine_matrix*p.positions_real.';
      #    p.positions_real = p.positions_real.';
      #end
      dxSpec=meas.dxSpec
      pos=dtype(np.round((posReal-posReal.min(axis=0))/dxSpec))
      meas['pos']=pos
    return pos

  def GetDetector(self,detector,roi,dtype,verbose):
    try:
      fnValidMsk=detector['validMsk']
    except KeyError:
      module=np.array(detector['module'],dtype=dtype)[:,::-1] #convert from x,y,w,h to h,w,y,x=nrow,ncol,row,col (in hdf5 file)
      badPix=np.array(detector['badPix'],dtype=dtype)[:,::-1] #convert from x,y to y,x=row,col (in hdf5 file)
    else:
      import scipy.io
      mat=scipy.io.loadmat(fnValidMsk)
      validMsk=mat['mask']
      try:
        rot90,fliplr=detector['trfMsk']
      except KeyError: pass
      else:
        if rot90:
          validMsk=np.rot90(validMsk,rot90)
        if fliplr:
          validMsk=np.fliplr(validMsk)
      #validMsk[934:954,744:764]=4
      #validMsk[944,754]=5
      if verbose&16:
        import pylab as plt #used for the colormaps
        plt.figure()
        plt.imshow(validMsk,interpolation='nearest')
        plt.show()
      sx,sy=roi
      validMsk=validMsk[sy,sx]
      validMsk=np.fft.fftshift(validMsk)
      if (validMsk.sum(axis=0)==0).any() or (validMsk.sum(axis=1)==0).any():
        x=validMsk.sum(axis=0)
        y=validMsk.sum(axis=1)
        xModule=[]
        xx=np.where(x==0)[0]
        p0=-1
        for p1 in xx:
          if p1-p0>1:
            xModule.append((p0+1,p1))
            p0=p1
          else:
            p0=p1
        if x.shape[0]-p0>1:
         xModule.append((p0+1,x.shape[0]))

        yModule=[]
        yy=np.where(y==0)[0]
        p0=-1
        for p1 in yy:
          if p1-p0>1:
            yModule.append((p0+1,p1))
            p0=p1
          else:
            p0=p1
        if y.shape[0]-p0>1:
         yModule.append((p0+1,x.shape[0]))

        module=np.zeros(shape=(len(xModule)*len(yModule),4),dtype=dtype)
        idx=0
        for ym in yModule:
          for xm in xModule:
            module[idx,:]=(ym[1]-ym[0],xm[1]-xm[0],ym[0],xm[0])
            idx+=1
        validMsk[:,xx]=2
        validMsk[yy,:]=3
      else:
        module=np.zeros(shape=(1,4),dtype=dtype)
        module[0,:2]=validMsk.shape
      #detector.module=module #put into meta structure

      badPix=np.where(validMsk.T==0)
      badPix=np.rot90(np.array(badPix,np.uint32),-1)
      #detector.badPix=badPix #put into meta structure

      if verbose&16:
        import pylab as plt #used for the colormaps
        plt.figure()
        plt.imshow(validMsk,interpolation='nearest')
        plt.show()
    return (module,badPix)


class InitialCondition:
  'shape initial conditions'
  @staticmethod
  def Build(meta,param,shape):
    dtype=np.complex128
    if param[0]=='zeros':
      arr=np.zeros(shape,dtype)
    elif param[0]=='random':
      #arr=np.ndarray(buffer=np.random.bytes(reduce(mul,shape)*np.dtype(dtype).itemsize),shape=shape, dtype=dtype)
      arr=np.ones(shape,dtype)+1j*1e-6*np.random.random(shape)
    elif param[0]=='hdf':
      fid = h5py.h5f.open(param[1],flags=h5py.h5f.ACC_RDONLY)
      dsid=h5py.h5d.open(fid,param[2])
      ds=h5py.Dataset(dsid)
      arr=ds.value
      fid.close()
    elif param[0]=='mat':
      fn=param[1];elem=param[2]
      import scipy.io
      mat=scipy.io.loadmat(fn)
      arr=mat[elem]
    elif param[0]=='focusMdl': #focused probe model
      p=param[1]#dictionary
      #zone_plate_diameter
      #outer_zone_width
      #focal_length
      #propagation_dist
      #diameter
      #upsample
      lmbda = meta.glob.lambda_ # wavelength
      z = meta.glob.z   # Distance from object to detector
      ds = meta.glob.ds # Camera pixel size
      dx_spec = lmbda*z/(shape*ds) #resolution in the specimen plane
      try:
        zp_f = p.zone_plate_diameter * p.outer_zone_width / lmbda
      except KeyError:
        zp_f = p.focal_length

      upsample = p.upsample;

      defocus = p.propagation_dist;
      Nprobe = upsample*shape[0];                       # Array dimension for the simulation
      dx = (zp_f+defocus)*lmbda/(Nprobe*dx_spec[0]);   # pixel size in the pupil plane
      r1_pix = .5 * p.diameter / dx;              # size in pixels of first pinhole

      # Pupil
      x=np.arange(0,Nprobe)-Nprobe/2
      y=np.arange(0,Nprobe)-Nprobe/2
      x,y=np.meshgrid(x,y)
      r2 = x**2 + y**2
      w = (r2 < (r1_pix)**2)

      # Propagation
      probe_hr = InitialCondition.PropagationFF(w * np.exp(-1j * np.pi * r2 * dx**2 / (lmbda * zp_f)), lmbda, zp_f + defocus, dx)

      # Cropping back to field of view
      probe = probe_hr[Nprobe/2+(-shape[0]/2):Nprobe/2+np.floor((shape[0]-1)/2), Nprobe/2+(-shape[1]/2):Nprobe/2+(np.floor((shape[1]-1)/2))]
      #arr=np.double(w)
      arr=probe
      pass
    elif param[0]=='pinholeMdl': #pinhole probe model
      p=param[1]#dictionary
      #dx_spec
      #diameter
      #propagation_dist
      lmbda = meta.glob.lambda_ # wavelength
      z = meta.glob.z   # Distance from object to detector
      ds = meta.glob.ds # Camera pixel size
      dx_spec = lmbda*z/(shape*ds) #resolution in the specimen plane
      x=np.arange(0,shape[0])-shape[0]/2
      y=np.arange(0,shape[0])-shape[0]/2
      x,y=np.meshgrid(x,y)
      arr = ( (x * dx_spec[0])**2 + (y * dx_spec[1])**2 < (p.diameter/2)**2);
      arr = InitialCondition.PropagationNF(np.float64(arr), lmbda, p.propagation_dist, dx_spec)
      arr = arr * np.sqrt(np.prod(shape)/(np.abs(arr)**2).sum())
      pass
    else:
      _log.warning('unsupported object description')
    return arr

  @staticmethod
  def PropagationNF(win, lmbda, z, dx=(1.,1.)):
    # ported from MATLAB function PROP_FREE_NF
    #PROP_FREE_NF(WIN, LAMBDA, Z, PIXSIZE) returns the propagated wavefield
    #WIN by a distance Z, using wavelength LAMBDA. PIXSIZE is the dimension
    #of one pixel.
    #PROP_FREE_NF(WIN, LAMBDA, Z) is the same as above, assuming PIXSIZE=1
    #(that is, Z and LAMBDA are expressed in pixel units).
    sz = win.shape
    # Evaluate if aliasing could be a problem

    if np.sqrt(sum(1./dx**2) * sum(1./(sz*dx)**2)) * abs(z) * lmbda > 1:
      _log.warning('there could be some aliasing issues...\nyou could enlarge your array, or try a far field method');

    x=np.arange(0,sz[0],dtype=np.float64)-sz[0]/2;x/=sz[0]*dx[0]
    y=np.arange(0,sz[0],dtype=np.float64)-sz[1]/2;y/=sz[1]*dx[1]
    x,y=np.meshgrid(x,y)
    q2 = np.fft.fftshift(x**2 + y**2);
    return np.fft.ifft2(np.fft.fft2(win) * np.exp(2j * np.pi * (z / lmbda) * (np.sqrt(1 - q2*lmbda**2) - 1) ) )

  @staticmethod
  def PropagationFF(win, lmbda, z, dx=(1.,1.)):
    # ported from MATLAB function PROP_FREE_NF
    #PROP_FREE_FF    Far field propagation
    #
    #   PROP_FREE_FF(WIN, LAMBDA, Z, PIXSIZE) returns the propagated wavefield
    #   WIN by a distance Z, using wavelength LAMBDA. PIXSIZE is the dimension
    #   of one pixel.
    #
    #   PROP_FREE_FF(WIN, LAMBDA, Z) is the same as above, assuming PIXSIZE=1
    #   (that is, Z and LAMBDA are expressed in pixel units).
    #
    # In this implementation, the output wave pixel size becomes
    # Z*LAMBDA/(N*PIXSIZE) (where N is the linear dimension of the array).
    sz = win.shape
    N = sz[0]
    #Evaluate if aliasing could be a problem
    if N*np.sqrt(2.) > abs(z)*lmbda:
      _log.warning('Warning: there could be some aliasing issues...\nyou could try a near field method)')

    x=np.arange(0,N,dtype=np.float64)-N/2
    y=np.arange(0,N,dtype=np.float64)-N/2
    x,y=np.meshgrid(x,y)
    r2 = x**2 + y**2

    wout = -1j * np.exp(1j * np.pi * lmbda * z * r2 /N**2) * np.fft.fftshift(np.fft.fft2(np.fft.fftshift(win * np.exp(1j * np.pi * r2 / (lmbda*z)))))
    return wout

  @staticmethod
  def Show(arr):
    import matplotlib.pyplot as plt
    from matplotlib.colors import hsv_to_rgb

    phase = (np.angle(arr) + np.pi) / (2*np.pi)
    amp = np.absolute(arr)
    amp /= amp.max()

    pnr, pnc = arr.shape
    hsv = np.ndarray((pnr, pnc, 3))
    hsv[:,:,0] = phase
    hsv[:,:,1] = 1.0
    hsv[:,:,2] = amp

    fig=plt.figure(0)
    fig.canvas.mpl_connect('motion_notify_event', InitialCondition.OnMotion)
    hi=plt.imshow(hsv_to_rgb(hsv),interpolation='nearest')
    fig.colorbar(hi,orientation='vertical')
    hi.raw=arr

    fig=plt.figure(1)
    fig.canvas.mpl_connect('motion_notify_event', InitialCondition.OnMotion)
    hi=plt.imshow(amp,interpolation='nearest')
    fig.colorbar(hi,orientation='vertical')
    fig=plt.figure(2)
    fig.canvas.mpl_connect('motion_notify_event', InitialCondition.OnMotion)
    hi=plt.imshow(phase,interpolation='nearest')
    fig.colorbar(hi,orientation='vertical')
    plt.show()

  @staticmethod
  def OnMotion(event):
    ax=event.inaxes
    if ax!=None:
      try:
        hi=ax.get_images()[0]
      except IndexError:
        return
      else:
        try:
          arr=hi.raw
        except AttributeError:
          arr=hi.get_array()
        x=round(event.xdata)
        y=round(event.ydata)
        try:
          s= 'x=%d, y=%d, val=%s'%(x,y, str(arr[x,y]))
        except IndexError:
          return
        tb=event.canvas.toolbar
        if mpl.backends.backend=='WXAgg':
          event.canvas.Parent.SetStatusText(s,1) #0=left, 1=right status
        elif  mpl.backends.backend=='TkAgg':
          event.canvas.toolbar.message.set(s)

class PtychoProc():
  def SetProcess(self,args,meas):
    '''setup process'''
    roi=meas.roi
    try:
      trfData=meas.trfData
    except AttributeError:
      trfData=None
    if roi:
      cx,cy,w,h=roi
      sx=cx-w/2; sy=cy-h/2
      self.roi=(slice(sx,sx+w),slice(sy,sy+h))
    self.Process=self._ProcessNew
    self.trfData=trfData

  def _ProcessNew(self,data):
    '''fft shifted data
       calculate maxPower
       random badPixels
    '''
    if self.trfData: #THIS whole BLOCK SHOULD BE REMOVED !! ONLY FOR TESTING IMAGE TRANSFORMATION
      (rot90,fliplr)=self.trfData
      if rot90:
        data=np.rot90(data,rot90)
      if fliplr:
        data=np.fliplr(data)
    try:
      sx,sy=self.roi
    except AttributeError: pass
    else:
      #~ print data.shape, sy, sx
      data=data[sy,sx]
    data=np.fft.fftshift(data)
    self.maxPower=data.sum()
    self.resArr=data
    #n=np.random.randint(0,10)
    n=0
    if n:
      shape=(n,2)
      dtype=np.uint32
      sz=np.prod(shape)*np.dtype(dtype).itemsize
      self.badPix=np.ndarray(buffer=np.random.bytes(sz),dtype=np.uint32,shape=shape)
      self.badPix.flags.writeable=True
      self.badPix&=0xff
    else:
      self.badPix=None

class PtychoOutHDF():

  def Open(self,args,meta):
    fnHdf5=meta._m.glob.dst[1]
    self.fid=fid=ut.OpenFile(fnHdf5,'w')
    res,v=LibDetXR.GetVersion()
    attr=[('creator','Converted data with '+os.path.basename(__file__)),
          ('version', 'LibDetXR  %d.%d.%d.%d'%tuple(v[::-1])),
          #('source',json.dumps(meas.src))
          ]
    attr.append(('format',2))
    ut.AddAttributes(fid,attr)
    meta.WriteHdfBaseStructure(args,fid)

  def Close(self,meta):
    ut.CloseFile(self.fid)

  def AddMeasurement(self,args,clpIn,proc,meta,measKey):
    fid=self.fid
    meas=meta._m.measurement[measKey]

    try:
      num=1
      #~ num=clpIn.param[1].shape[0]
      _log.info('number of Images,'+str(num))
    except BaseException:
      _log.info('could not determine number of frames')
      num=1

    grpName='measurement/n'+measKey
    dsName='data'
    gid=ut.CreateGrpPath(fid,grpName)

    if args.multiproc:
      #-> read  MultiProc doc
      if type(clpIn.param[0]) ==h5py.h5f.FileID:
        try:
          (ffid,ds)=clpIn.param
          ffid.close()
          del clpIn.param
          del ds
        except (TypeError,ValueError): pass
      self.MultiProc(args,meas.src,proc,gid,num)
      #dsObj.FitShape()
      #ut.CloseFile(fid)
      return
    verb=(args.verbose&2) and not (args.verbose&0x200)
    if args.cmpr:
      CmprChunkBuild=ut.GetChunkBuildFunc(args.cmpr)
    t0=time.time(); tProc=0.
    badPixMeas=np.ndarray(num,dtype=object)
    numBadPixMeas=0
    maxPower=0
    pipe=args.get('pipe'); frmCnt=0
    while True:
      if pipe and pipe.poll():
        cmd=pipe.recv()
        if cmd=='stop':
          break
        elif cmd=='frmCnt':
          pipe.send(str(frmCnt))
      data=clpIn.GetData()
      #~ print np.shape(data)
      #~ import matplotlib.pyplot as plt    
      #~ fig=plt.figure(1)
      #~ hi=plt.imshow(np.abs(data))
      #~ plt.show()
      if data is None:
        break
      t=time.time()
      proc.Process(data);frmCnt+=1
      tProc+=time.time()-t
      idx=clpIn.idx
      mp=proc.maxPower
      if mp>maxPower:maxPower=mp
      outBuf=proc.resArr
      print proc.resArr.shape
      if idx+1>badPixMeas.size:
        badPixMeas.resize(idx+64,refcheck=False)
      badPixMeas[idx]=proc.badPix
      if proc.badPix!=None: numBadPixMeas+=proc.badPix.shape[0]

      if idx==0:
        dtype=outBuf.dtype
        shape=(num,)+outBuf.shape
        maxshape=(h5py.h5s.UNLIMITED,)+outBuf.shape
        chunk=(1,)+outBuf.shape
        dsObj=ut.DsObj(ut.CreateDatasetChunked(gid,dsName,dtype,shape,maxshape,chunk,args.cmpr),shape,chunk)
      dsObj.MaxReshape(idx)
      if args.cmpr:
        (cmprBuf,filter)=CmprChunkBuild(outBuf)
        dsObj.ds.write_chunk((idx,0,0),cmprBuf,filters=filter)
        if args.verbose&0x200:
          LogOut('%d bytes filter %d\n'%(cmprBuf.nbytes,filter))
      else:
        outBuf=proc.resArr #needed because prs.resArr overwritten
        dsObj.ds.write_chunk((idx,0,0),outBuf)
        if args.verbose&0x200:
          LogOut('%d bytes\n'%outBuf.nbytes)
      if verb:
        LogOut('.')
    if pipe: pipe.send(str(frmCnt)) #send the frame numbers to the pipe

    if verb:
      LogOut('\n')
    #ut.StopWatch.Log('Total time')
    _log.info('Process %f Total %f'%(tProc,time.time()-t0))

    dsObj.FitShape()
    ut.SetAttribute(gid,'max_sum',maxPower)

    n=dsObj.maxIdx+1
    if numBadPixMeas>0:
      badPix=np.ndarray((numBadPixMeas,2),dtype=np.uint32)
      badPixIdx=np.ndarray(n,dtype=np.uint32)
      p=0
      for idx in range(n):
        bp=badPixMeas[idx]
        if bp!=None:
          q=p+bp.shape[0]
          badPix[p:q,:]=bp
          p=q
        badPixIdx[idx]=p
      grp=h5py.Group(gid)
      grp.create_dataset('bad_pixels_index', data=badPixIdx,compression='gzip',shuffle=True)
      grp.create_dataset('bad_pixels',    data=badPix,compression='gzip',shuffle=True)

  def MultiProc(self,args,src,proc,gid,num):
    jobs = []
    maxJobs=mproc.cpu_count()
    for idx in range(maxJobs):
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=PtychoOutHDF.MultiProcChildProc, args=(idx,maxJobs,src,args.cmpr,proc,args.verbose,pipeChild))
      #raw=clpIn.GetData()
      jobs.append((p,pipeParent))
      p.start()

    dsName='data'

    maxPower=0
    while True:
      for (p,pipeParent) in jobs:
        (idx,outBuf,shape,dtype,filter,mp)=pipeParent.recv()
        if mp>maxPower:maxPower=mp
        if idx==0:
          cmpr=args.cmpr
          dsShape=(num,)+shape
          maxshape=(h5py.h5s.UNLIMITED,)+shape
          chunk=(1,)+shape
          dsObj=ut.DsObj(ut.CreateDatasetChunked(gid,dsName,dtype,dsShape,maxshape,chunk,cmpr),dsShape,chunk)
          dsObj.ds.write_chunk((idx,0,0),outBuf,filters=filter)
          ds=dsObj.ds
        elif idx!=None:
          dsObj.MaxReshape(idx)
          dsObj.ds.write_chunk((idx,0,0),outBuf,filters=filter)
      if idx==None:
        break
    ut.SetAttribute(gid,'max_sum',maxPower)
    dsObj.FitShape()

  @staticmethod
  def MultiProcChildProc(idx,step,src,cmpr,ppty,verbose,pipeChild):
    if cmpr:
      CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)
    else:
      filter=0
    args=ut.dotdict({'src':src})
    clpIn=ClampIn(args,verbose=verbose)
    clpIn.SetIdx(idx)

    raw=clpIn.GetData()
    _log.info('running MultiProcChildProc PID %d'%os.getpid())
    while raw is not None:
      try:
        raw.ravel()[ppty.mskIdx]=0
      except AttributeError: pass
      ppty.Process(raw)
      outBuf=ppty.resArr
      try:
        maxPower=ppty.maxPower
      except AttributeError:
        maxPower=0
      if cmpr:
        (cmprBuf,filter)=CmprChunkBuild(outBuf)
      else:
        cmprBuf=outBuf
      pipeChild.send((clpIn.idx,cmprBuf,outBuf.shape,outBuf.dtype,filter,maxPower))
      if verbose&0x200:
        LogOut('%d bytes filter %d\n'%(cmprBuf.nbytes,filter))
      elif verbose&2:
        sys.stdout.write('.');sys.stdout.flush()
      clpIn.idx+=step-1
      raw=clpIn.GetData()

    pipeChild.send((None,)*6)
    _log.info('end MultiProcChildProc PID %d'%os.getpid())


class PtychoOutZMQ():
  def Open(self,args,meta):
    pass

  def AddMeasurement(self,args,clpIn,proc,meta,measKey):
    ifDict={'PUSH':zmq.PUSH,'PUB':zmq.PUB}

    dst= meta._m.glob.dst[1]
    print dst
    context = zmq.Context()
    try:
      ifType=ifDict[dst['ifType']]
    except KeyError:
      ifType=zmq.PUSH
    sock = context.socket(ifType)#default is zmq.PULL
    #if ifType==zmq.PUB:
    #  sock.setsockopt(zmq.SUBSCRIBE, '')
    try:
      queueSz=dst['queueSz']
    except KeyError:
      queueSz=4
    sock.set_hwm(queueSz)#limit to queueSz messages MUST BE DONE ON SENDER AND RECEIVER?
    addr=dst['addr']
    if addr.find('*')>=0:
      sock.bind(addr)
    else:
      sock.connect(addr)
    try:
      timeout=dst['timeout']
    except KeyError:pass
    else:
      sock.RCVTIMEO=timeout # set 1000 ms timeout

    cmpr=args.cmpr
    if cmpr:
      CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)

    verb=(args.verbose&2) and not (args.verbose&0x200)
    while True:
      data=clpIn.GetData()
      if data==None:
        break
      proc.Process(data)
      idx=clpIn.idx
      outBuf=proc.resArr
      if idx==0:
        header ={}
        header['htype']='chunk-1.0'
        header['shape']=tuple(reversed(outBuf.shape))
        header['type'] =str(outBuf.dtype)
      try:
        tag=clpIn._tag
      except AttributeError: pass
      else:
        header['tag']=tag
      header['frame']=idx
      if cmpr:
        (cmpArr,filter)=CmprChunkBuild(outBuf)
        if filter:
          header.pop('cmpr')
        else:
          header['cmpr']=cmpr
          outBuf=cmpArr
      try:
        #Send multipart message
        if args.verbose&0x200:
          LogOut(str(header)+'\n')
        sock.send_json(header, zmq.SNDMORE) #@UndefinedVariable
        sock.send(outBuf)
      except zmq.error.Again as e:
        _log.error(e)
      if verb:
        LogOut('.')
    if verb:
      LogOut('\n')
    sock.send_json({'cmd':'stop'})

  def Close(self,meta):
    pass

def RunWithArgs(args):
  if type(args)==dict:
    args=ut.dotdict(args)
    print args
    default={
            'numFrm':sys.maxint,
            'multiproc':False,
            'cmpr':None,
            'verbose':7
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
    #if args.multiproc==True: args.multiproc=[3]
    _log.setLevel(logging.INFO if args.verbose&1 else logging.WARNING)
  else:
    args=ut.dotdict(args.__dict__)

  meta=PtychoMeta()
  meta.Read(args.meta)
  if args.verbose&8:
    _log.info('PrettyPrint')
    sys.stdout.flush()
    meta._m.PrettyPrint()

  if meta._m.glob.dst[0]=='hdf':
    ptOut=PtychoOutHDF()
  elif meta._m.glob.dst[0]=='zmq':
    ptOut=PtychoOutZMQ()
  else:
    _log.error('unsupported destination:'+str(meta._m.glob.dst))

  ptOut.Open(args,meta)

  ptProc=PtychoProc()

  for k,meas in meta._m.measurement.iteritems():
    print meas.src[1][0]
    print args
    for j in range(np.shape(meas.src[1])[0]):
        submeas = meas.src[1][j]
        clpIn=ClampIn(meas,verbose=args.verbose)
        ptProc.SetProcess(args,meas)
        ptOut.AddMeasurement(args,clpIn,ptProc,meta,j)
  ptOut.Close(meta)

if __name__ == '__main__':
  import os,sys,argparse #since python 2.7

  def ParseArgs():
    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('--meta ptycho.json',
                '--meta ptycho.json --cmpr s-zlib4 --multiproc'
                )
    epilog='''\n
the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress
   bit2(0x04): multiproc
   bit2(0x08): show meta
   bit1(0x10): show valid masks
   bit3(0x200): inStream,outStream Clamps (each frame)
   default is 1+2+4=7 (no -v)
   -v is 511
   -v0 is turn all off

Data Source description:'''+ClampIn.__types__+'''
Data Destination description:
  HDF file from a given object:
    --dst hdf FILE ELEM
    --dst hdf myfile.hdf5 entry/mygrp/dataset
  ZMQ stream
    --dst zmq JSON-config (with addr and optional queueSz, ifType and timeout)
    --dst zmq '{"addr":"tcp://localhost:8080","queueSz":4,"ifType":"PULL"}',

Examples:
  sample data are in
  cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne/scan_00106-00161.hdf5 /entry/data/pilatus_1 roi 778 758 280 220
  AnaESRF/forward_ptycho/forward_ptycho_edf3004_3004.hdf entry/data/dataset roi 155 347 300 300
  cSAXS_sxdm_2014_07_omny_commissioning/eiger/S00000-00999/S00033/e15285_1_00033_00000_00000.h5 roi 688 256 400 400
  cSAXS_sxdm_2014_07_omny_commissioning/eiger/S00000-00999/S00033/e15285_1_00033_00000_00000.h5 eh5/images roi 688 256 400 400

  IN_DIR=/scratch/detectorData'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__,
                                     epilog=epilog)

    #parser.add_argument('--src', nargs='+',      required=True,                  help='data source (see below for examples)')
    #parser.add_argument('--dst', nargs='+',         metavar=('mode','param'),    help='data destination: hdf or zmq sink')
    #parser.add_argument('--roi', nargs=4, type=int, metavar=('CX','CY','W','H'), help='region of interest used for data extraction)')
    parser.add_argument('-v', dest='verbose', type=int, nargs='?', const=511, default=7, help='verbose (optional level bitmask)')
    parser.add_argument('--multiproc',                action='store_true', help='use multiprocessing method (if possible)')
    parser.add_argument('--cmpr',                                          help='compression method (zlib[lvl],lz4,lzf,cbf)')
    parser.add_argument('--meta',                      required=True,      help='JSON file with meta information')
    #parser.add_argument('--validMask',                                     help='file with the valid pixel mask (!=0 for valid)')

    args = parser.parse_args()
    ut.LogConfig(args, {__name__     : -1, 'utilities'  : -1,})
    RunWithArgs(args)

  def Test0():
    ic0=["hdf", "/scratch/detectorData/cSAXS_sxdm_2014_07_omny_commissioning/analysis/S00033/S00033_initial_conditions_400x400.h5","probes/probe_0"]
    ic1=["mat","/afs/psi.ch/user/z/zamofing_t/tmp/probe.mat","probe"]
    ic2=["mat","/scratch/detectorData/cSAXS_sxdm_2014_07_omny_commissioning/matlab/ptycho/probe_eiger_S00033_400x400_test_1_c.mat","probe"]
    ic3=["focusMdl",
        {
          "zone_plate_diameter":150e-6,
          "outer_zone_width":60e-9,
          "focal_length":0,
          "propagation_dist":13.2e-3,
          "diameter":10e-6,
          "upsample":10
        }]
    ic4=["pinholeMdl",
       {
         "propagation_dist":13.2e-3,
         "diameter":10e-6
       }]
    for v in (ic3,ic4):
      v[1]=ut.dotdict(v[1])


    pm=PtychoMeta()
    pm.Read('/tmp/ptycho.json')
    sz=np.array((400,400))
    arr=InitialCondition.Build(pm._m,ic4,sz)
    InitialCondition.Show(arr)
    #arr=InitialCondition.Build(pm._m,ic3,sz)
    #InitialCondition.Show(arr)

  def Test1(fn1,fn2):
    fid1=h5py.File(fn1)
    fid2=h5py.File(fn2)
    #fid1['objects'].items()
    arr1=fid1['objects']['object_0'].value
    arr2=fid2['objects']['object_0'].value
    arr=np.absolute(arr1-arr2)
    import matplotlib.pyplot as plt
    from matplotlib.colors import hsv_to_rgb

    fig=plt.figure(1)
    fig.canvas.mpl_connect('motion_notify_event', InitialCondition.OnMotion)
    hi=plt.imshow(arr,interpolation='nearest')
    fig.colorbar(hi,orientation='vertical')
    plt.show()
  #------------------ Main Code ----------------------------------
  ut.LogConfig(ut.dotdict({'verbose':7}), {__name__     : -1, 'utilities'  : -1,})
  ParseArgs()

  #Test0()
  #Test1('/scratch/tmp/libDetXRTest/ptychoOut0.hdf5','/scratch/tmp/libDetXRTest/ptychoOut1.hdf5')
  #Test1('/scratch/tmp/libDetXRTest/ptychoOut1.hdf5','/scratch/tmp/libDetXRTest/ptychoOut2.hdf5')

