#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
Process statistics of multiple ROI.
Inputs are:
- a indexable data
- multiple list of indexes to calculate statistics on
- (optional) weights if the indexes
- outVectors: sum, n, avg, min, max, square
The output will be the statistics for each ROI
'''
import logging,os
import numpy as np
import json, zmq, h5py, time, sys
import utilities as ut
from Clamp import Clamp,ClampIn,ClampOut,ClampLst,MultiProc,SingleProc
from libDetXR import LibDetXR
import pylab as plt
import matplotlib as mpl

LibDetXR()

_log=logging.getLogger(__name__)

def LogOut(s):
  log=logging.root.handlers[0].stream
  log.write(s)
  log.flush()

class ProcMoment():
  mode=('python','pyFast','opencv','c')

  def __init__(self,args=None):
    if args==None: return # only for independent tests
    self.args=args
    if args.dst==None:
      assert(args.src[0]=='hdf')
      pt,fn=os.path.split(args.src[1])
      assert(fn.startswith('scan_'))
      args.dst=('hdf',os.path.join(pt,'moments_'+fn[5:]),'entry/data/moments')
    if args.matMsk==None and args.src[0]=='hdf':
      pt=os.path.dirname(args.src[1])
      args.matMsk=os.path.join(pt,'analysis/data/pilatus_valid_mask.mat')

    if args.matMsk!=None:
      self.SetMskMat(args.matMsk,args.matMskFlt,args.verbose&8)
    if args.roi!=None:
      self.roi=(slice(args.roi[1],args.roi[1]+args.roi[3]),slice(args.roi[0],args.roi[0]+args.roi[2]))
      self.shape=(args.roi[3],args.roi[2])
    self.SetProcess(args.procmode)

  def Close(self):
    if self.args.verbose&0x04:
      _log.info('processed data %d, time masking %f, time processing %f'%(self.tCnt,self.tMsk,self.tProc))
    pass

  def SetMskMat(self, fn, fltFullColRow=False, show=False):
    '''Sets the valid mask with a .mat-file. the shape is (y,x) of the data image.
       The Content of the .mat-file is as it was at the cSAXS beamline.'''
    import scipy.io
    mat=scipy.io.loadmat(fn)

    mskIndices=mat['valid_mask'][0,0]['indices'].reshape(-1)
    shape     =mat['valid_mask'][0,0]['framesize'].reshape(-1)

    mask=np.ndarray(shape,dtype=np.bool8)
    mask[:]=True
    idx=(mskIndices.reshape(-1)-1)
    y,x=np.unravel_index(idx,shape,order='F')
    x=shape[1]-1-x; y=shape[0]-1-y
    #try:
    #  idx=np.ravel_multi_index((y,x),shape)
    #except ValueError:
    #  print x,y

    mask[y,x]=False
    #remove fully masked cols and rows from mask
    if fltFullColRow:
      xZero=mask.all(axis=0)
      yZero=mask.all(axis=1)
      mask[:,xZero]=False
      mask[yZero,:]=False


    if show:
      import pylab as plt #used for the colormaps
      plt.figure()
      plt.imshow(mask,interpolation='nearest')
      plt.show()

    self.shape=shape
    #self.mskIdx=np.where(mask.reshape(-1)==True)[0]
    #self.mskIdx=np.where(mask.ravel()==True)[0]
    self.mskIdx=np.flatnonzero(mask)

  def Process(self,idx,data):
    self.tCnt+=1
    t=time.time()
    data=self.Mask(data)
    self.tMsk+=time.time()-t
    t=time.time()
    res=self._Process(idx,data)
    self.tProc+=time.time()-t
    #_log.info('process %d'%idx)
    #time.sleep(.3)
    return res

  def Mask(self,data):
    try:
      data.ravel()[self.mskIdx]=0
    except AttributeError: pass
    try:
      data=data[self.roi]
    except AttributeError: pass
    return data

  def SetProcess(self,proc='python',dtype=np.float32):
    '''set the roi process and allocates result memory
       !!!self.shape must exist. e.g. by calling SetMskMat or by setting it from  args.roi!!!'''
    self.tMsk=self.tProc=0.;self.tCnt=0
    shape=self.shape
    w=6 #m00 m01 m10 m11 m02 m20
    self.args.comment='the moments m00 m01 m10 m11 m02 m20'
    if proc==ProcMoment.mode[0]:
      self._Process=self._ProcessMoment
    elif proc==ProcMoment.mode[1]:
      self._Process=self._ProcessMomentFast
      self.w01=np.arange(shape[0],dtype=np.uint32)
      self.w10=np.arange(shape[1],dtype=np.uint32)
      self.w02=self.w01*self.w01
      self.w20=self.w10*self.w10
      self.w11=np.dot(self.w01.reshape((-1,1)),self.w10.reshape((1,-1)))
      self.w11=np.array(self.w11,dtype=np.uint64)# need to avoid overflow in some images!!!
      #time for 3696 2M-32bit images in 32 bit:251 sec
      #time for 3696 2M-32bit images in 64 bit:257 sec
      #if the moment is calculates around the center x=-800..800 instead x=0..1600
      #32 bit could be perhaps used again and the calculation would be numerically more stable.
    elif proc==ProcMoment.mode[2]:
      self.args.comment='the moments m00 m01 m10 m11 m02 m20 m12 m21 m03 m30'
      self._Process=self._ProcessMomentCV
      w=10 #m00 m01 m10 m11 m02 m20 m12 m21 m03 m30
    elif proc==ProcMoment.mode[3]:
      res=LibDetXR.ProcessMoments(1, 0, shape[1], shape[0],0,0) #AllocMomentsI32F32
      self._Process=self._ProcessMoment_C
    else:
      raise 'unknown processing type'
    self.resArr=np.ndarray((w,),dtype=dtype)
    self.resArr[:]=0

  def _ProcessMoment(self,idx,data):
    data=self.Mask(data)
    def raw_moment(data, iord, jord):
        nrows, ncols = data.shape
        y, x = np.mgrid[:nrows, :ncols]
        data = data * x**iord * y**jord
        return data.sum()

    m00 = data.sum()
    m10 = raw_moment(data, 1, 0)
    m01 = raw_moment(data, 0, 1)
    m11 = raw_moment(data, 1, 1)
    m20 = raw_moment(data, 2, 0)
    m02 = raw_moment(data, 0, 2)
    m=(m00,m01,m10,m11,m02,m20)
    self.resArr[0:len(m)]=m
    return self.resArr

  def _ProcessMomentFast(self,idx,data):
    s0=data.sum(axis=0)
    s1=data.sum(axis=1)
    m00=s0.sum()
    m10=(s0*self.w10).sum()
    m20=(s0*self.w20).sum()
    m01=(s1*self.w01).sum()
    m02=(s1*self.w02).sum()
    m11=(data*self.w11).sum()
    m=(m00,m01,m10,m11,m02,m20)
    self.resArr[0:len(m)]=m
    return self.resArr

  def _ProcessMomentCV(self,idx,data):
    import cv2
    m=cv2.moments(data)
    m00=m['m00'];m01=m['m01'];m10=m['m10'];m11=m['m11'];m02=m['m02'];m20=m['m20'];
    xbar=m['m10']/m00
    ybar=m['m01']/m00
    cov=np.array([[m['mu20']/m00, m['mu11']/m00], [m['mu11']/m00, m['mu02']/m00]])
    #print 'openCV\n',xbar, ybar, cov
    u11 = (m11 - xbar * m01)
    print 'openCV',m['mu11']/m00,u11/m00

    m=(m00,m01,m10,m11,m02,m20)
    self.resArr[0:len(m)]=m
    return self.resArr

  def _ProcessMoment_C(self,idx,data):
    #m00,m10,m01,m11,m20,m02
    res=LibDetXR.ProcessMoments(0,data.ctypes.data, data.shape[1], data.shape[0],data.strides[0]/data.strides[1]-data.shape[1],self.resArr.ctypes.data) #ProcMomentsUI32F32
    return self.resArr

class ClampOutMoment(ClampOut):
  @staticmethod
  def GetCaps(param):
    if param.dst[0]=='hdf':
      return 0
    return 2

  def __init__(self,args,**kwargs):
    ClampOut.__init__(self,**kwargs)
    self.args=args
    #_log.info('initialize ClampOutMoment'+str(args.dst))
    if args.dst[0]=='hdf':
      self.mode=0
      self.InitHDF()#clpIn,proc)
    elif args.dst[0]=='zmq':
      self.mode=1
      self.InitZMQ()
    else:
      _log.error('unsupported Out Clamp')

  def InitHDF(self):#,clpIn,proc):
    args=self.args
    fnHdf5=args.dst[1]
    self.fid=ut.OpenFile(fnHdf5,'w')
    self.idxMax=0

  def WriteHDF(self):
    args=self.args
    cmpr=args.cmpr
    fid=self.fid
    outBuf=self.resArr
    grpName,dsName=args.dst[2].rsplit('/',1)
    gid=ut.CreateGrpPath(fid,grpName)
    res,v=LibDetXR.GetVersion()
    attr=[('creator','Converted data with '+os.path.basename(__file__)),
          ('version', 'LibDetXR  %d.%d.%d.%d'%tuple(v[::-1])),
          #('source',json.dumps(args.src)),
          ('procmode',args.procmode),
          #('contains',args.comment)
          ]
    if args.matMsk!=None:
      attr.append(('matMsk',args.matMsk))
    if args.roi!=None:
      attr.append(('roi',args.roi))

    sh=outBuf.shape;sh=(self.idxMax+1,)+sh[1:]
    outBuf.resize(sh,refcheck=False)

    dtype=outBuf.dtype
    shape=outBuf.shape
    maxshape=outBuf.shape
    chunk=outBuf.shape
    dsObj=ut.DsObj(ut.CreateDatasetChunked(gid,dsName,dtype,shape,maxshape,chunk,cmpr),shape,chunk)
    ut.AddAttributes(dsObj.ds,attr)
    if cmpr:
      CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)
      (cmprBuf,filter)=CmprChunkBuild(outBuf)
      dsObj.ds.write_chunk((0,0),cmprBuf,filters=filter)
      if args.verbose&64:
        LogOut('shape %s, %d bytes filter %d\n'%(str(shape),cmprBuf.nbytes,filter))
    else:
      dsObj.ds.write_chunk((0,0),outBuf)
      if args.verbose&64:
        LogOut('shape %s, %d bytes\n'%(str(shape),outBuf.nbytes))
    ut.CloseFile(fid)
    print maxshape

  def InitZMQ(self):
    args=self.args.dst[1]
    print args
    _log.info('open socket'+str(args))
    self.sock=ut.ZMQ.Socket(args,zmq.PUSH,4)

  def SetData(self,idx,res):
    if self.mode==0:
      try: resArr=self.resArr
      except AttributeError:
        self.resArr=resArr=np.ndarray(shape=(512,)+res.shape,dtype=res.dtype)
      try: resArr[idx,:]=res
      except IndexError:
        sh=resArr.shape;sh=(sh[0]+512,)+sh[1:]
        resArr.resize(sh,refcheck=False)
        resArr[idx,:]=res
      if idx>self.idxMax:
        self.idxMax=idx
    elif self.mode==1:
      sock=self.sock
      try:
        header=self.header
      except AttributeError:
        header ={'htype':'chunk-1.0','shape':tuple(reversed(res.shape)),'type':str(res.dtype)}
        #try:
        #  tag=clpIn._tag
        #except AttributeError: pass
        #else:
        #  header['tag']=tag
      header['frame']=idx
      sock.send_json(header, zmq.SNDMORE) #@UndefinedVariable
      sock.send(res)
      pass

    if self.args.verbose&0x200:
      _log.info('SetData '+str(idx)+str(res))
    elif self.args.verbose&2:
      LogOut('.')

  def Close(self):
    if self.args.verbose&2:
      LogOut('\n')
    if self.mode==0:
      self.WriteHDF()
    if self.mode==1:
      self.sock.close()
      _log.info('close socket'+str(self.args.dst))
    #print 'Close PID %d'%os.getpid()

class OriEccInt: #orientation eccentricity intensity
  def __init__(self,data):
    m00=data[...,0];m01=data[...,1];m10=data[...,2];m11=data[...,3];m02=data[...,4];m20=data[...,5]

    xm = m10 / m00
    ym = m01 / m00
    u11 = (m11 - xm * m01) / m00
    u20 = (m20 - xm * m10) / m00
    u02 = (m02 - ym * m01) / m00

    a=(u20+u02)/2
    b=np.sqrt(4*u11**2+(u20-u02)**2)/2
    l0=a+b
    l1=a-b

    shape=data.shape[:2] #y,x (orientation, roundness, sum) h s v

    #oei=orientation eccentricity intensity
    o=np.ndarray(shape, dtype=np.float32)
    e=np.ndarray(shape, dtype=np.float32)
    i=np.ndarray(shape, dtype=np.float32)
    #eccentricity: http://en.wikipedia.org/wiki/Eccentricity_%28mathematics%29

    o=0.5*np.arctan2(2*u11,(u20-u02)) #orientation value -pi..pi
    e=np.sqrt(1-l1/l0) #eccentricity value: 0 (circle) to 1 (line)
    i=m00 #sum

    #map orientation to range to [0..1]
    o[o<0]+=np.pi # -120deg identical to 60deg
    o/=np.pi        # set range to [0..1]

    norm=np.ndarray(shape+(3,), dtype=np.float32)
    norm[...,0]=o
    self.norm=norm
    self.e=e
    self.i=i
    self.eMax=e.max()
    self.eMin=e.min()
    self.iMax=i.max()
    self.iMin=i.min()

  def Normalize(self,eGamma,i0,iScale,iGamma):
    norm=self.norm
    e=self.e
    i=self.i
    eMax=self.eMax
    eMin=self.eMin
    iMax=self.iMax
    iMin=self.iMin
    #1: eccentricity
    norm[...,1]=(e-eMin)/(eMax-eMin)
    if eGamma!=1: norm[...,1]=norm[...,1]**eGamma #gamma correction

    #2: intensity
    norm[...,2]=(i-iMin)/(iMax-iMin)
    if iGamma!=1: norm[...,2]=norm[...,2]**iGamma #gamma correction
    norm[...,2]=i0+norm[...,2]*(1.-i0)*iScale
    #m[...,2]=0.0+m[...,2]/(1.0*m2Max)
    #m[:,:,2]=1. #turns off transmission mapping.
    #m[:,:,2]=np.log(m[:,:,2]+1) #log
    norm[norm[...,2]>1,2]=1. #limit to maximum value 1.
    return norm

class ShowMoment():

  @staticmethod
  def GetIntertialAxis(moments):
    """Calculate the x-mean, y-mean, and cov matrix of an image."""
    m=moments
    #m=(2345769, 1777974540, 1813911665, 1313734064881, 1420153616578, 1430108808649)
    m=np.array(m,dtype=np.float128)
    m00=m[0];m01=m[1];m10=m[2];m11=m[3];m02=m[4];m20=m[5]
    xm = m10 / m00
    ym = m01 / m00
    u11 = (m11 - xm * m01) / m00
    u20 = (m20 - xm * m10) / m00
    u02 = (m02 - ym * m01) / m00
    cov = np.array([[u20, u11], [u11, u02]],dtype=np.float64)
    print cov
    a=(u20+u02)/2
    b=np.sqrt(4*u11**2+(u20-u02)**2)/2
    l0=a+b
    l1=a-b
    ang=0.5*np.arctan2(2*u11,(u20-u02))/(2*np.pi)*360. #orientation value 0..1
    exc=np.sqrt(1-l1/l0) #eccentricity :circle=0: http://en.wikipedia.org/wiki/Eccentricity_%28mathematics%29

    print 'xb:%g yb:%g cov:%g %g %g %g  ang:%g exc:%g'%((xm, ym)+tuple(cov.ravel())+(ang,exc))
    return xm, ym, cov

  @staticmethod
  def PlotMoments(ax, xbar, ybar, cov):
    def make_lines(eigvals, eigvecs, mean, i):
        """Make lines a length of 2 stddev."""
        std = np.sqrt(eigvals[i])
        vec = 2 * std * eigvecs[:,i] / np.hypot(*eigvecs[:,i])
        x, y = np.vstack((mean-vec, mean, mean+vec)).T
        return x, y
    mean = np.array([xbar, ybar])
    eigvals, eigvecs = np.linalg.eigh(cov)
    go=ax.plot(*make_lines(eigvals, eigvecs, mean, 0), marker='o', color='white')
    go+=ax.plot(*make_lines(eigvals, eigvecs, mean, -1), marker='o', color='red')
    return go

  def SetHDFSrc(self,hdfFile,hdfEntry,width,hdfRawFile=None,hdfRawEntry=None):
    #base='/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne/'
    #fid = h5py.h5f.open(os.path.join(base,'moments_00106-00161.hdf5'))
    #hid = h5py.h5o.open(fid,'entry/data/moments/c')
    fid = h5py.h5f.open(hdfFile,flags=h5py.h5f.ACC_RDONLY)
    hid = h5py.h5o.open(fid,hdfEntry)
    ds=h5py.Dataset(hid)
    m=ds.value
    m2=m.reshape((-1,width,m.shape[-1]))
    self.data=m2
    try:
      self.roi=ds.attrs['roi']
    except KeyError: pass
    if hdfRawFile!=None:
      fid = h5py.h5f.open(hdfRawFile,flags=h5py.h5f.ACC_RDONLY)
      self.hidRaw = h5py.h5o.open(fid,hdfRawEntry)

  def Show(self):
    data=self.data

    self.oei=OriEccInt(data)

    eGamma=3 #eccentricity gamma correction
    i0=.25    #intensity minimal value
    iScale=1.5  #multiplier of intensity
    iGamma=1  #intensity gamma correction
    #normalize ranges to 0 to 1
    oeiNorm=self.oei.Normalize(eGamma,i0,iScale,iGamma)

    print ' orientation:', oeiNorm[...,0].min(),oeiNorm[...,0].max()
    print 'eccentricity:', oeiNorm[...,1].min(),oeiNorm[...,1].max()
    print '   intensity:', oeiNorm[...,2].min(),oeiNorm[...,2].max()

    fig=plt.figure()
    ax=plt.subplot(221)
    ax.format_coord=self.OnFmtCoord
    himgO=plt.imshow(oeiNorm[...,0],interpolation='nearest',cmap='hsv')
    plt.colorbar()
    plt.title('orientation')
    ax.invert_xaxis()

    ax=plt.subplot(222)
    ax.format_coord=self.OnFmtCoord
    himgE=plt.imshow(oeiNorm[...,1],interpolation='nearest')
    plt.colorbar()
    plt.title('eccentricity')
    ax.invert_xaxis()

    ax=plt.subplot(223)
    ax.format_coord=self.OnFmtCoord
    himgI=plt.imshow(oeiNorm[...,2],interpolation='nearest')
    plt.colorbar()
    plt.title('sum')
    ax.invert_xaxis()

    sldEG = mpl.widgets.Slider(fig.add_axes([0.60, 0.40, 0.3, 0.03], ), 'eGamma', 0.5, 4.0, valinit=eGamma)
    sldI0 = mpl.widgets.Slider(fig.add_axes([0.60, 0.35, 0.3, 0.03], ), 'i0', 0.0, 0.9, valinit=i0)
    sldIS = mpl.widgets.Slider(fig.add_axes([0.60, 0.30, 0.3, 0.03], ), 'iScale', 1., 10., valinit=iScale)
    sldIG = mpl.widgets.Slider(fig.add_axes([0.60, 0.25, 0.3, 0.03], ), 'iGamma', 0.25, 2.0, valinit=iGamma)
    self.sldLst=(sldEG,sldI0,sldIS,sldIG)
    for sld in self.sldLst:
      sld.on_changed(self.OnUpdate)

    #ax=plt.subplot(224)
    fig=plt.figure()
    himgOEI=plt.imshow(mpl.colors.hsv_to_rgb(oeiNorm))#,interpolation='nearest')
    ax=fig.gca()
    ax.format_coord=self.OnFmtCoord
    plt.title('hsv with orientation, eccentricity, sum')
    cid = fig.canvas.mpl_connect('button_press_event', self.OnEvtBtnPress)
    cid = fig.canvas.mpl_connect('motion_notify_event', self.OnEvtMotion)
    self.hImgLst=(himgO,himgE,himgI,himgOEI)
    ax.invert_xaxis()

    cm=mpl.cm.hsv
    n=cm.N/2*2
    theta = np.linspace(0,2*np.pi,n);
    r = np.arange(2)
    X, Y = np.meshgrid(theta, r)

    C = np.ndarray(shape=(len(theta),len(r)))
    C[:n/2,0]=np.linspace(0,1,n/2)
    C[n/2:,0]=C[:n/2,0]
    C[:,1]=C[:,0]

    ax=fig.add_axes([0.90, 0.80, 0.1, 0.1],polar=True)
    ax.yaxis.set_major_locator(plt.NullLocator())
    plt.pcolormesh(X, Y, C.T,cmap=cm)

    #ax.invert_yaxis()
    #ang=0.5*np.unwrap(np.arctan(2*u11/(u20-u02)))*360/(2*np.pi)
    #plt.plot(m[...,1].ravel())
    try:
      hidRaw=self.hidRaw
    except AttributeError:pass
    else:
      fig=plt.figure()
      ax=fig.gca()
      data=h5py.Dataset(hidRaw)[100,...]
      avg=np.average(data); std=np.std(data)
      vmin=np.min(data);vmax=np.max(data)
      vmin=max(vmin,avg-3*std);vmax=min(vmax,avg+3*std)
      #vmax=500
      vmin=.1
      #self.rawImg = ax.imshow(data,interpolation='nearest',cmap=mpl.cm.jet, vmin=vmin, vmax=vmax)
      self.rawImg = ax.imshow(data,interpolation='nearest',cmap=mpl.cm.jet, norm=mpl.colors.LogNorm(vmin=1, vmax=vmax))
      ax.set_axis_bgcolor('black')
      colBar=fig.colorbar(self.rawImg,orientation='vertical')
      ax.invert_xaxis()
      ax.axis('image')
    plt.show()

  def OnUpdate(self,val):
    #print val
    (sldEG,sldIG,sldI0,sldIS)=self.sldLst
    v=map(lambda v:v.val, self.sldLst)
    #print v

    oeiNorm=self.oei.Normalize(*v)
    (himgO,himgE,himgI,himgOEI)=self.hImgLst
    #print self.hImgLst
    himgE.set_array(oeiNorm[...,1])
    himgI.set_array(oeiNorm[...,2])
    himgI.set_clim(v[1],1.)#set minimal value to i0
    himgO.axes.figure.canvas.draw_idle()
    himgOEI.set_array(mpl.colors.hsv_to_rgb(oeiNorm))
    himgOEI.axes.figure.canvas.draw_idle()
    #amp = samp.val
    #freq = sfreq.val
    #l.set_ydata(amp*np.sin(2*np.pi*freq*t))
    #fig.canvas.draw_idle()


  def OnFmtCoord(self,x, y):
    x=int(round(x));  y=int(round(y))
    try:
      ang=self.oei.norm[y,x,0]*180
      ecc=self.oei.e[y,x]
      m00=self.oei.i[y,x]
      return 'x:%d y:%d idx:%d trans:%g ang:%g ecc:%g'%(x,y,y*self.data.shape[1]+x,m00,ang,ecc)
    except IndexError as e:
      return str(e)

  def OnEvtMotion(self, event):
    #print event.button, event.x, event.y, event.inaxes, event.xdata, event.ydata
    pass

  def OnEvtBtnPress(self, event):
    #print event.button, event.x, event.y, event.inaxes, event.xdata, event.ydata
    if event.inaxes:
      x=int(round(event.xdata)); y=int(round(event.ydata))
      m=self.data[y,x,:]
      m00=m[0];m01=m[1];m10=m[2];m11=m[3];m02=m[4];m20=m[5]
      xm = m10 / m00
      ym = m01 / m00
      u11 = (m11 - xm * m01) / m00
      u20 = (m20 - xm * m10) / m00
      u02 = (m02 - ym * m01) / m00
      cov = np.array([[u20, u11], [u11, u02]],dtype=np.float64)
      a=(u20+u02)/2
      b=np.sqrt(4*u11**2+(u20-u02)**2)/2
      l0=a+b
      l1=a-b
      ang=0.5*np.arctan2(2*u11,(u20-u02))/(2*np.pi)*360. #orientation value 0..1
      if ang<0:ang+=180
      exc=np.sqrt(1-l1/l0) #eccentricity :circle=0: http://en.wikipedia.org/wiki/Eccentricity_%28mathematics%29
      print 'x:%d y:%d idx:%d trans:%g ang:%g exc:%g'%(x,y,y*self.data.shape[1]+x,m00,ang,exc)
      print 'xb:%g yb:%g cov:%g %g %g %g'%((xm, ym)+tuple(cov.ravel()))

      try:
        hidRaw=self.hidRaw
      except AttributeError:pass
      else:
        idx=y*self.data.shape[1]+x
        data=h5py.Dataset(hidRaw)[idx]
        img=self.rawImg
        ax=img.axes
        fig=ax.figure
        img.set_array(data)
        #self.rawFig=fig
        ax.set_title('image %d'%idx)
        #self.rawFig.show()
        try:
          for o in self.goMoment:
            o.remove()
        except AttributeError: pass
        try: roi=self.roi
        except AttributeError:pass
        else: xm+=roi[0]; ym+=roi[1] #attributes of entry/data/<dataset>
        self.goMoment=ShowMoment.PlotMoments(ax, xm, ym, cov)
        fig.canvas.draw_idle()

      pass

def RunWithArgs(args):
  if type(args)==dict:
    args=ut.dotdict(args)
    default={
            'numFrm':sys.maxint,
            'matMsk':None,
            'matMskFlt':False,
            'roi':None,
            'dst':None,
            'procmode':ProcMoment.mode[3],
            'multiproc':None,
            'cmpr':False,
            'verbose':3
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
    if args.multiproc==True: args.multiproc=[3]
    _log.setLevel(logging.INFO if args.verbose&1 else logging.WARNING)
  else:
    args=ut.dotdict(args.__dict__)
  for clp in (args.src,args.dst):
    if clp[0] in ('zmq','zmqMsg'):
      clp[1]=ut.ZMQ.lst2dict(clp[1:])
      del clp[2:]

  print args.src,args.dst

  #print '>>>>',args,'<<<<<'
  if args.multiproc!=None:
    MultiProc(ClampIn,ProcMoment,ClampOutMoment,args)
  else:
    SingleProc(ClampIn,ProcMoment,ClampOutMoment,args)

  if args.dst[0]in('zmq','zmqMsg'):
    sock=ut.ZMQ.Socket( args.dst[1],zmq.PUSH,4)
    sock.send_json({'cmd':'stop'})
    sock.close()

if __name__ == '__main__':
  import os,sys,argparse #since python 2.7

  def ParseArgs():
    if len(sys.argv)>1 and sys.argv[1]=='show':
      hdfFile=sys.argv[2]
      hdfEntry=sys.argv[3]
      width=int(sys.argv[4])
      try:
        hdfRawFile=sys.argv[5]
        hdfRawEntry=sys.argv[6]
      except IndexError:
        hdfRawFile=None
        hdfRawEntry=None
      sm=ShowMoment()
      sm.SetHDFSrc(hdfFile,hdfEntry,width,hdfRawFile,hdfRawEntry)
      sm.Show()
      return

    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('--src hdf $IN_DIR/scan_00106-00161.hdf5 /entry/data/pilatus_1 --procmode pyFast',
                '--src hdf $IN_DIR/scan_00106-00161.hdf5 /entry/data/pilatus_1 --procmode c',
                '--src hdf $IN_DIR/scan_00106-00161.hdf5 /entry/data/pilatus_1 --multiproc --cmpr s-zlib4',
                'show $IN_DIR/moments_00106-00161.hdf5 entry/data/moments 66',
                'show $IN_DIR/moment_00106-00161.hdf5 entry/data/moments 66 $IN_DIR/scan_00106-00161.hdf5 /entry/data/pilatus_1')

    epilog='''\n
the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress
   bit2(0x04): process log
   bit3(0x08): gui masks (DO NOT USE IN MULTIPROCESSING!)
   bit4(0x10):
   bit5(0x20):
   bit6(0x40):
   bit6(0x80):
   bit8(0x100): multiprocessing+clamps
   bit9(0x200): multiprocessing+clamps (more verbose)

   default is 1+2=3 (no -v)
   -v is 1+2+8+256=267
   -v0 is turn all off\n\nData Source description:'''+ClampIn.__types__+'''
Data Destination description:
  HDF file from a given object:
    --src hdf FILE ELEM
    --src hdf myfile.hdf5 entry/mygrp/dataset
  ZMQ stream
    --src zmq JSON-config (with addr and optional queueSz, ifType and timeout)
    --src zmq '{"addr":"tcp://localhost:8080","queueSz":4,"ifType":"PULL"}',

to show result:'''+cmd+''' show HDF5FILE HDF5ENTRY WIDTH [RAWHDF5FILE RAWHDF5ENTRY]\n
Examples:
  IN_DIR=/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__,
                                     epilog=epilog)

    parser.add_argument('--src',   nargs='+', required=True, help='data source (see below for examples)')
    parser.add_argument('--matMsk',  metavar='.mat-File',    help='MATLAB Mask file for valid pixels (e.g. pilatus_valid_mask.mat)')
    parser.add_argument('--matMskFlt', action='store_true', help='filter MATLAB Mask cols and rows')
    parser.add_argument('--roi',  type=int, nargs=4,   metavar=('X','Y','W','H'),    help='ROI to calculate moments')
    parser.add_argument('--dst',  nargs='+', metavar=('mode','param'), required=True, help='data destination: hdf or zmq sink')
    parser.add_argument('--procmode', choices=ProcMoment.mode, default=ProcMoment.mode[3], help='processing method to use')
    parser.add_argument('-m', '--multiproc', nargs='?', const=[3], default=None,help='use multiprocessing method topo,[numproc]')
    parser.add_argument('--cmpr',                                          help='compression method for hdf5 output (zlib[lvl],lz4,lzf,cbf)')
    parser.add_argument('-v',         dest='verbose',  type=int,  nargs='?', const=267,default=3,  help='verbose level (see below)')

    args = parser.parse_args()
    if type(args.multiproc)==str:
      args.multiproc=map(int,args.multiproc.split(','))
      assert(len(args.multiproc)<=2)

    ut.LogConfig(args, {__name__     : -1, 'utilities'  : -1, 'Clamp'  : -1,})
    RunWithArgs(args)

  def Test0():
    base='/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne/analysis/data'
    pm=ProcMoment()
    pm.SetMskMat(os.path.join(base,'pilatus_valid_mask.mat'),True,True)

  def Test1():
    import pylab as plt
    def generate_data():
        shape=(400, 300)
        data = np.zeros(shape, dtype=np.float)
        ctr=(150,180)
        u20=20**2 #varianz
        u02=10**2
        u11=-100
        cov = np.array([[u02,u11],[u11,u20]])
        ij = np.random.multivariate_normal(ctr, cov, int(1e4))
        iMax,jMax=data.shape
        for i,j in ij:
          if i>=0 and j>=0 and i<iMax and j<jMax:
            data[int(i), int(j)] += 1
        return data

    while True:
      data=generate_data()
      data=np.round(data*1000)
      pm=ProcMoment();pm.args=ut.dotdict()
      pm.shape=data.shape
      pm.SetProcess('python')
      ut.StopWatch.Start()
      pm.Process(0,data)
      t0=ut.StopWatch.Log('Process python')
      print pm.resArr
      pm.SetProcess('pyFast')
      ut.StopWatch.Start()
      pm.Process(0,data)
      t0=ut.StopWatch.Log('Process pyFast')
      print pm.resArr
      pm.SetProcess('opencv')
      ut.StopWatch.Start()
      pm.Process(0,data)
      t0=ut.StopWatch.Log('Process opencv')
      print pm.resArr
      pm.SetProcess('c')
      ut.StopWatch.Start()
      data32=np.uint32(data)
      pm.Process(0,data32)

      t0=ut.StopWatch.Log('Process c')
      print
      xbar, ybar, cov=ShowMoment.GetIntertialAxis(pm.resArr)
      fig, ax = plt.subplots()
      ax.imshow(data,interpolation='nearest')
      go=ShowMoment.PlotMoments(ax, xbar, ybar, cov)
      ax.axis('image')
      plt.show()


      #procRoiStat.SetProcess('avg')
      #ut.StopWatch.Start()
      #procRoiStat.Process(data)
      #t1=ut.StopWatch.Log('Process')
      #print procRoiStat.resArr[:,procRoiStat.roiLenArr>0]
      #print 'Process with library implementation %f faster than pure Python'%(t0/t1)
      pass

  def Test2():
    'process an existing mat file of some hdf images (NOT CORRECT CENTER)'
    base='/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne'
    clpIn=ClampIn(('hdf',os.path.join(base,'scan_00106-00161.hdf5'),'/entry/data/pilatus_1'))
    pm=ProcMoment();pm.args=ut.dotdict()
    pm.SetMskMat(os.path.join(base,'analysis/data/pilatus_valid_mask.mat'),True,False)
    pm.SetProcess('python')
    pm.SetProcess('c')
    pm.SetProcess('pyFast')


    import pylab as plt #used for the colormaps

    #data=clpIn.GetData(20)

    from scipy import ndimage as ndi
    while True:
      ut.StopWatch.Start()
      clpIn.SetIdx(2482)
      data=clpIn.GetData()
      if data==None:
        break
      data[pm.mskIdx==False]=0

      fig, ax = plt.subplots(3)
      ax[0].plot(data.sum(axis=0))
      ax[1].plot(data.sum(axis=1))

      ut.StopWatch.Log('getImg')
      pm.Process(0,data)
      t0=ut.StopWatch.Log('Process')
      xbar, ybar, cov=ShowMoment.GetIntertialAxis(pm.resArr)
      #fig, ax = plt.subplots()
      ax[2].imshow(data,vmax=100,interpolation='nearest')
      go=ShowMoment.PlotMoments(ax[2], xbar, ybar, cov)
      ax[2].axis('image')
      plt.show()

    print 'done'
    return

#------------------ Main Code ----------------------------------
  ParseArgs()
  #Test0() #show mask
  #Test1() #calculate moments, center and orientation on generated random images
  #Test2() #calculate moments, center and orientation on generated on real images
  #import cProfile
  #cProfile.run('Test2()')

