#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
Process statistics of multiple ROI.
Inputs are:
- a indexable data
- multiple list of indexes to calculate statistics on
- (optional) weights if the indexes
- outVectors: sum, n, avg, min, max, square
The output will be the statistics for each ROI
'''
import logging, time, json, os, sys
import numpy as np
import h5py, zmq
import multiprocessing as mproc
import utilities as ut
from Clamp import Clamp,ClampIn,ClampOut,ClampLst,MultiProc,SingleProc
from libDetXR import LibDetXR
LibDetXR()
#import setproctitle as spt
#import sys
#print sys.argv
#spt.setproctitle(' '.join(sys.argv))

#python -c "import time;import setproctitle as spt ;print spt.getproctitle();spt.setproctitle('myProc');print 'gugus';time.sleep(10)"
_log=logging.getLogger(__name__)

def LogOut(s):
  log=logging.root.handlers[0].stream
  log.write(s)
  log.flush()

class ProcRoiStat():
  mode=('avg','avgstd','sumsq')
  def __init__(self,args):
    self.args=args
    if args.dst==None:
      assert(args.src[0]=='hdf')
      pt,fn=os.path.split(args.src[1])
      assert(fn.startswith('scan_'))
      args.dst=('hdf',os.path.join(pt,'roiStat_'+fn[5:]),'entry/data/roiStat')
    if args.matRoi==None:
      assert(args.src[0]=='hdf')
      pt=os.path.dirname(args.src[1])
      validMsk=os.path.join(pt,'analysis/data/pilatus_valid_mask.mat')
      integMsk=os.path.join(pt,'analysis/data/pilatus_integration_mask.mat')
      args.matRoi=(validMsk,integMsk)
    self.SetRoiMat(args.matRoi,args.verbose&8)
    self.SetProcess(args.procmode)

  def Close(self):
    pass

  def SetRoiMat(self, fn, show=False):
    '''Sets the ROI with a .mat-filex. the shape is (y,x) of the data image.
       The Content of the .mat-file is as it was at the cSAXS beamline.'''
    import scipy.io

    fnValMsk=fn[0]
    fnIntegMsk=fn[1]

    mat=scipy.io.loadmat(fnValMsk)
    shape     =mat['valid_mask'][0,0]['framesize'].reshape(-1)
    mat=scipy.io.loadmat(fnIntegMsk)
    #print mat.keys()
    self.mskNumSeg=mat['no_of_segments'].squeeze()
    ctr=mat['center_xy'].squeeze()
    #print mat['integ_masks'].dtype
    #print mat['integ_masks'][0][0].dtype

    self.mskRadius=mat['integ_masks'][0,0]['radius'].squeeze()
    if self.args.noSect:
      self.mskNumSeg=1
      tmp=mat['integ_masks'][0,0]['indices']
      mskIndices=np.ndarray(shape=(tmp.shape[0]),dtype=np.object)
      for i in range(tmp.shape[0]):
        mskIndices[i]=np.vstack(tmp[i])
      mskNormSum=mat['integ_masks'][0,0]['norm_sum'].sum(axis=1,dtype=np.uint16)
    else:
      mskIndices=mat['integ_masks'][0,0]['indices'].T.reshape(-1)
      mskNormSum=mat['integ_masks'][0,0]['norm_sum'].T.reshape(-1)


    roiLenArr=np.array(map(np.size,mskIndices),dtype=np.uint32)
    assert((roiLenArr==mskNormSum).all())
    #roiLenArr=roiLenArr[roiLenArr>0]
    roiSz=roiLenArr.sum()

    roiIdxArr=np.ndarray(roiSz,dtype=np.uint32)
    ofs=0
    if show:
      roiPic=np.ndarray(shape,dtype=np.uint32)
      roiPic[:]=0
      v=0
    for i,idx in enumerate(mskIndices):
      if idx.size==0:
        continue
      idx=(idx.reshape(-1)-1)
      y,x=np.unravel_index(idx,shape,order='F')
      x=shape[1]-1-x; y=shape[0]-1-y
      try:
        idx=np.ravel_multi_index((y,x),shape)
      except ValueError:
        print x,y
      ofsEnd=ofs+roiLenArr[i]
      roiIdxArr[ofs:ofsEnd]=idx
      ofs=ofsEnd
      if show:
        roiPic.ravel()[idx]=v
        v=np.uint32(1000*np.random.rand())
        #v+=1
    if show:
      import pylab as plt #used for the colormaps
      plt.figure()
      plt.imshow(roiPic,interpolation='nearest')
      plt.show()
    #plt.show(block=False)
    self.SetRoi(roiIdxArr,roiLenArr)

  def GetMskRadius(self):
    try: return self.mskRadius
    except AttributeError: return None

  def GetRoiLen(self):
    try: return self.roiLenArr
    except AttributeError: return None

  def SetRoi(self,roiIdxArr,roiLenArr,wgtArr=None):
    self.roiIdxArr=roiIdxArr
    self.wgtArr=wgtArr
    self.roiLenArr=roiLenArr
    _log.info('numnber of ROI %d Total number of pixels %d'%(self.roiLenArr.size,self.roiIdxArr.size))

  def SetProcess(self,proc='avg',dtype=np.float32,useLib=True):
    '''set the roi process and allocates result memory
       !!!SetRoi must be called before!!!'''
    if proc=='avg':
      w=1
      if self.wgtArr==None:
        if useLib:
          self.Process=self._ProcessAvg_C
        else:
          self.Process=self._ProcessAvg
      else:
        if useLib:
          self.Process=self._ProcessWgtAvg_C
        else:
          self.Process=self._ProcessWgtAvg
    elif proc=='avgstd':
      w=2
      if useLib:
        self.Process=self._ProcessAvgStd_C
      else:
        self.Process=self._ProcessAvgStd
    elif proc=='sumsq':
      w=2;self.Process=self._ProcessSumSq
    else:
      raise 'unknown processing type'
    self.resArr=np.ndarray((w,self.roiLenArr.size),dtype=dtype)

    if self.mskNumSeg>1:
      try:
        if w==1:
          self.resArr=self.resArr.reshape(self.mskNumSeg,-1)
        else:
          self.resArr=self.resArr.reshape(w,self.mskNumSeg,-1)
      except AttributeError:
        pass
    self.resArr[:]=0

  def _ProcessAvg(self,idx,data):
    roiIdxArr=self.roiIdxArr
    #wgtArr=self.wgtArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr.ravel()
    data=data.ravel()
    ofs=0
    for idx,l in enumerate(roiLenArr):
      if l==0:
        continue
      roi=roiIdxArr[ofs:ofs+l]
      resArr[idx]=data[roi].mean()
      ofs+=l
    return resArr

  def _ProcessAvg_C(self,idx,data):
    roiIdxArr=self.roiIdxArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr
    res=LibDetXR.ProcessRoiStatistics(data.ctypes.data, roiIdxArr.ctypes.data, 0, roiIdxArr.size, roiLenArr.ctypes.data, roiLenArr.size, resArr.ctypes.data, 0)
    return resArr

  def _ProcessWgtAvg(self,idx,data):
    roiIdxArr=self.roiIdxArr
    wgtArr=self.wgtArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr

    ofs=0
    for idx,l in enumerate(roiLenArr):
      roi=roiIdxArr[ofs:ofs+l]
      wgt=wgtArr[ofs:ofs+l]
      resArr[0,idx]=(wgt*data.ravel()[roi]).mean()
      ofs+=l
    return resArr

  def _ProcessWgtAvg_C(self,idx,data):
    roiIdxArr=self.roiIdxArr
    wgtArr=self.wgtArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr
    assert(wgtArr.dtype==np.float32)
    res=LibDetXR.ProcessRoiStatistics(data.ctypes.data, roiIdxArr.ctypes.data, wgtArr.ctypes.data, roiIdxArr.size, roiLenArr.ctypes.data, roiLenArr.size, resArr.ctypes.data, 1)
    return resArr

  def _ProcessAvgStd(self,idx,data):
    roiIdxArr=self.roiIdxArr
    wgtArr=self.wgtArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr
    ofs=0
    for idx,l in enumerate(roiLenArr):
      if l:
        roi=roiIdxArr[ofs:ofs+l]
        d=data.ravel()[roi]
        resArr[0].ravel()[idx]=d.mean()
        resArr[1].ravel()[idx]=d.std()
        ofs+=l
    return resArr

  def _ProcessAvgStd_C(self,idx,data):
    roiIdxArr=self.roiIdxArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr
    res=LibDetXR.ProcessRoiStatistics(data.ctypes.data, roiIdxArr.ctypes.data, 0, roiIdxArr.size, roiLenArr.ctypes.data, roiLenArr.size, resArr.ctypes.data, 2)
    return resArr

  def _ProcessSumSq(self,idx,data):
    roiIdxArr=self.roiIdxArr
    #wgtArr=self.wgtArr
    roiLenArr=self.roiLenArr
    resArr=self.resArr
    ofs=0
    for idx,l in enumerate(roiLenArr):
      roi=roiIdxArr[ofs:ofs+l]
      d=data.ravel()[roi]
      resArr[0,idx]=d.sum()
      resArr[1,idx]=(d*d).sum()
      ofs+=l
    return resArr

class ClampOutRoiStat(ClampOut):
  @staticmethod
  def GetCaps(param):
    if param.dst[0]=='hdf':
      return 0
    return 2

  def __init__(self,args,**kwargs):
    ClampOut.__init__(self,**kwargs)
    self.args=args
    #_log.info('initialize ClampOutMoment'+str(args.dst))
    if args.dst[0]=='hdf':
      self.mode=0
      self.InitHDF()#clpIn,proc)
    elif args.dst[0]=='zmq':
      self.mode=1
      self.InitZMQ()
    else:
      _log.error('unsupported Out Clamp')

  def InitHDF(self):#,clpIn,proc):
    args=self.args
    fnHdf5=args.dst[1]
    self.fid=ut.OpenFile(fnHdf5,'w')
    self.idxMax=0

  def InitZMQ(self):
    args=self.args.dst[1]
    if type(args)==str: args=json.loads(args, object_hook=ut.ConvUtf8)
    _log.info('open socket'+str(args))
    self.sock=ut.ZMQ.Socket(args,zmq.PUSH,4)

  def SetData(self,idx,outBuf):
    #_log.info('%d:'%idx+str(outBuf)) #virtual function to overload
    if self.mode==0:
      args=self.args
      try: dsObj=self.dsObj
      except AttributeError:
        cmpr=self.args.cmpr
        info=self.ClpSrcRPC(0, 'GetInfo')
        try:
          num=info['cnt']
          _log.info('number of Images: %d'%num)
        except (KeyError,TypeError):
          num=512
          _log.info('could not determine number of frames')
        res,v=LibDetXR.GetVersion()
        attr=(('creator','Converted data with '+os.path.basename(__file__)),
              ('version', 'LibDetXR  %d.%d.%d.%d'%tuple(v[::-1])),
              ('source',json.dumps(args.src)),
              ('matRoi',json.dumps(args.matRoi)),
              ('procmode',args.procmode))
        #grpName=args.dst[2]
        grpName,dsName=args.dst[2].rsplit('/',1)
        gid=ut.CreateGrpPath(self.fid,grpName)

        mskRadius=self.ProcRPC(0, 'GetMskRadius')
        if mskRadius!=None:
          grp=h5py.Group(gid)
          grp.create_dataset('mskRadius',  data=mskRadius, compression='gzip', compression_opts=4)
        roiLen=self.ProcRPC(0, 'GetRoiLen')
        if roiLen!=None:
          grp=h5py.Group(gid)
          grp.create_dataset('roiLen',  data=roiLen, compression='gzip', compression_opts=4)
        if cmpr:
          self.CmprChunkBuild=ut.GetChunkBuildFunc(args.cmpr)
        dtype=outBuf.dtype
        shape=(num,)+outBuf.shape
        maxshape=(h5py.h5s.UNLIMITED,)+outBuf.shape
        chunk=(1,)+outBuf.shape
        self.dsObj=dsObj=ut.DsObj(ut.CreateDatasetChunked(gid,dsName,dtype,shape,maxshape,chunk,cmpr),shape,chunk)
        ut.AddAttributes(dsObj.ds,attr)

      dsObj.MaxReshape(idx)
      chkId=(idx,)+(0,)*outBuf.ndim
      try: CmprChunkBuild=self.CmprChunkBuild
      except AttributeError:
        dsObj.ds.write_chunk(chkId,outBuf)
        #LogOut('%d bytes\n'%outBuf.nbytes)
      else:
        (cmprBuf,filter)=CmprChunkBuild(outBuf)
        dsObj.ds.write_chunk(chkId,cmprBuf,filters=filter)
        #LogOut('%d bytes filter %d\n'%(cmprBuf.nbytes,filter))
    elif self.mode==1:
      sock=self.sock
      try:
        header=self.header
      except AttributeError:
        header ={'htype':'chunk-1.0','shape':tuple(reversed(outBuf.shape)),'type':str(outBuf.dtype)}
        #try:
        #  tag=clpIn._tag
        #except AttributeError: pass
        #else:
        #  header['tag']=tag
      header['frame']=idx
      sock.send_json(header, zmq.SNDMORE) #@UndefinedVariable
      sock.send(outBuf)
      pass

    if self.args.verbose&2:#print 'SetData',idx,res;sys.stdout.flush()
      LogOut('.')

  def Close(self):
    if self.args.verbose&2:
      LogOut('\n')
    if self.mode==0:
      self.dsObj.FitShape()
      ut.CloseFile(self.fid)
    if self.mode==1:
      self.sock.close()
      _log.info('close socket'+str(self.args.dst))
    #print 'Close PID %d'%os.getpid()


def ShowMatMsk(fn, shape=(1679,1475)):
  '''Show the content of ROI .mat-files. the shape is (y,x) of the data image.
     The Content of the .mat-file is as it was at the cSAXS beamline.'''

# !!!!!! DO THIS ALSO FOR OTHER MASK FILES !!!!!!!!!!
#valMsk=load('/afs/psi.ch/user/z/zamofing_t/Documents/pilatus_valid_mask.mat')
#integMsk=load('/afs/psi.ch/user/z/zamofing_t/Documents/pilatus_integration_mask.mat')
#validMskBefBS=load('/afs/psi.ch/user/z/zamofing_t/Documents/pilatus_valid_mask_before_bs.mat')
#validMskBSrem=load('/afs/psi.ch/user/z/zamofing_t/Documents/valid_mask_bs_removed.mat')

  import scipy.io
  mat=scipy.io.loadmat(fn)
  baseName=os.path.basename(fn)
  import pylab as plt
  plt.figure()

  print mat.keys()
  if 'integ_masks' in mat.keys():
    mskRadius=mat['integ_masks'][0,0]['radius'].reshape(-1)
    mskIndices=mat['integ_masks'][0,0]['indices'].reshape(-1)
    mskNormSum=mat['integ_masks'][0,0]['norm_sum'].reshape(-1)


    roiLenArr=np.array(map(np.size,mskIndices),dtype=np.uint32)
    assert((roiLenArr==mskNormSum).all())
    #roiLenArr=roiLenArr[roiLenArr>0]
    roiSz=roiLenArr.sum()

    roiIdxArr=np.ndarray(roiSz,dtype=np.uint32)
    ofs=0
    roiPic=np.ndarray(shape,dtype=np.uint32)
    roiPic[:]=0
    v=0
    for i,idx in enumerate(mskIndices):
      if idx.size==0:
        continue
      idx=(idx.reshape(-1)-1)
      y,x=np.unravel_index(idx,shape,order='F')
      x=shape[1]-1-x; y=shape[0]-1-y
      try:
        idx=np.ravel_multi_index((y,x),shape)
      except ValueError:
        print x,y
      ofsEnd=ofs+roiLenArr[i]
      roiIdxArr[ofs:ofsEnd]=idx
      ofs=ofsEnd
      roiPic.ravel()[idx]=v
      v=np.uint32(1000*np.random.rand())
        #v+=1
    plt.imshow(roiPic,interpolation='nearest')
  elif baseName=='pilatus_valid_mask.mat':
    mskIndices=mat['valid_mask'][0,0]['indices'].reshape(-1)
    #mskIndices=mat['indices'].reshape(-1)
    shape     =mat['valid_mask'][0,0]['framesize'].reshape(-1)

    roiPic=np.ndarray(shape,dtype=np.bool8)
    roiPic[:]=False
    idx=(mskIndices.reshape(-1)-1)
    y,x=np.unravel_index(idx,shape,order='F')
    x=shape[1]-1-x; y=shape[0]-1-y
    roiPic[y,x]=True
    try:
      idx=np.ravel_multi_index((y,x),shape)
    except ValueError:
      print x,y
    plt.imshow(roiPic,interpolation='nearest')
  elif baseName=='pilatus_valid_mask_before_bs.mat':
    print mat.keys()
    print mat['ans'][0,0].dtype
    mskIndices=mat['ans'][0,0]['indices'].reshape(-1)
    shape     =mat['ans'][0,0]['framesize'].reshape(-1)
    roiPic=np.ndarray(shape,dtype=np.bool8)
    roiPic[:]=False
    idx=(mskIndices.reshape(-1)-1)
    y,x=np.unravel_index(idx,shape,order='F')
    x=shape[1]-1-x; y=shape[0]-1-y
    roiPic[y,x]=True
    try:
      idx=np.ravel_multi_index((y,x),shape)
    except ValueError:
      print x,y
    import pylab as plt #used for the colormaps
    plt.imshow(roiPic,interpolation='nearest')
  elif baseName=='valid_mask_bs_removed.mat':
    print mat.keys()
    print type(mat['ans']),mat['ans'].shape,mat['ans'].dtype
    mskIndices=mat['ans'].reshape(-1)
    roiPic=np.ndarray(shape,dtype=np.bool8)
    roiPic[:]=False
    idx=(mskIndices.reshape(-1)-1)
    y,x=np.unravel_index(idx,shape,order='F')
    x=shape[1]-1-x; y=shape[0]-1-y
    roiPic[y,x]=True
    try:
      idx=np.ravel_multi_index((y,x),shape)
    except ValueError:
      print x,y
    plt.imshow(roiPic,interpolation='nearest')

  plt.title(baseName)
  plt.show(block=False)

def RunWithArgs(args):
  if type(args)==dict:
    args=ut.dotdict(args)
    default={
            'numFrm':sys.maxint,
            'matRoi':None,
            'dst':None,
            'procmode':ProcRoiStat.mode[0],
            'multiproc':None,
            'cmpr':None,
            'noSect':False,
            'verbose':3
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
    if args.multiproc==True: args.multiproc=[3]
    _log.setLevel(logging.INFO if args.verbose&1 else logging.WARNING)
  else:
    args=ut.dotdict(args.__dict__)
  for clp in (args.src,args.dst):
    if clp[0] in ('zmq','zmqMsg'):
      clp[1]=ut.ZMQ.lst2dict(clp[1:])
      del clp[2:]

  if args.multiproc!=None:
    MultiProc(ClampIn,ProcRoiStat,ClampOutRoiStat,args)
  else:
    SingleProc(ClampIn,ProcRoiStat,ClampOutRoiStat,args)

  if args.dst[0]in('zmq','zmqMsg'):
    sock=ut.ZMQ.Socket(args.dst[1],zmq.PUSH,4)
    sock.send_json({'cmd':'stop'})
    sock.close()

if __name__ == '__main__':
  import os,sys,argparse #since python 2.7

  def ParseArgs():
    if len(sys.argv)>1 and sys.argv[1]=='show':
      import pylab as plt
      fn=sys.argv[2]
      szX=int(sys.argv[3])
      szY=int(sys.argv[4])
      ShowMatMsk(fn,(szX,szY))
      plt.show()
      return
    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('--src hdf $IN_DIR/scan_00106-00161.hdf5 /entry/data/pilatus_1 --matRoi $IN_DIR/analysis/data/pilatus_valid_mask.mat $IN_DIR/analysis/data/pilatus_integration_mask.mat --dst hdf /tmp/out.hdf5 entry/data/result',
                '--src hdf $IN_DIR/scan_00106-00161.hdf5 /entry/data/pilatus_1 --matRoi $IN_DIR/analysis/data/pilatus_valid_mask.mat $IN_DIR/analysis/data/pilatus_integration_mask.mat --dst hdf /tmp/out.hdf5 entry/data/result --multiproc',
                'show $IN_DIR/analysis/data/pilatus_valid_mask.mat 1679 1475',
                'show $IN_DIR/analysis/data/pilatus_integration_mask.mat 1679 1475'
               )
    epilog='''\n
the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress
   bit2(0x04):
   bit3(0x08): gui masks
   bit4(0x10):
   bit5(0x20):
   bit6(0x40):
   bit6(0x80):
   bit6(0x100): multiprocessing

   default is 1+2=3 (no -v)
   -v is 1+2+8+256=267
   -v0 is turn all off

Data Source description:'''+ClampIn.__doc__+'''
Data Destination description:
  HDF file from a given object:
    --src hdf FILE ELEM
    --src hdf myfile.hdf5 entry/mygrp/dataset
  ZMQ stream
    --src zmq JSON-config (with addr and optional queueSz, ifType and timeout)
    --src zmq '{"addr":"tcp://localhost:8080","queueSz":4,"ifType":"PULL"}',

to show matMask:'''+cmd+'''show MATFILE WIDTH HEIGHT

Examples:
  IN_DIR=/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__,
                                     epilog=epilog)

    parser.add_argument('--src',   nargs='+', required=True, help='data source (see below for examples)')
    parser.add_argument('--matRoi', nargs=2, metavar=('validMsk','integMsk'),    help='MATLAB Roi file (e.g. pilatus_integration_mask.mat)')
    parser.add_argument('--noSect', action='store_true', help='reduces matlab multi sectors to one sector')
    parser.add_argument('--dst',  nargs='+', metavar=('mode','param'), help='data destination: hdf or zmq sink')
    parser.add_argument('--procmode', choices=ProcRoiStat.mode, default=ProcRoiStat.mode[0], help='processing method to use (currently only avg is implemented fast and default)')
    parser.add_argument('-m', '--multiproc', nargs='?', const=[3], default=None,help='use multiprocessing method topo,[numproc]')
    parser.add_argument('--cmpr',                                          help='compression method (zlib[lvl],lz4,lzf,cbf)')
    parser.add_argument('-v',         dest='verbose',  type=int,  nargs='?', const=267,default=3,  help='verbose level (see below)')

    args = parser.parse_args()
    if type(args.multiproc)==str:
      args.multiproc=map(int,args.multiproc.split(','))
      assert(len(args.multiproc)<=2)
    ut.LogConfig(args, {__name__     : -1, 'utilities'  : -1,})
    RunWithArgs(args)

  def Test0():
    import pylab as plt
    base='/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne/analysis/data'
    ShowMatMsk(os.path.join(base,'pilatus_integration_mask.mat'),(1679,1475))
    ShowMatMsk(os.path.join(base,'pilatus_valid_mask.mat'),(1679,1475))
    ShowMatMsk(os.path.join(base,'pilatus_valid_mask_before_bs.mat'),(1679,1475))
    ShowMatMsk(os.path.join(base,'valid_mask_bs_removed.mat'),(1679,1475))
    plt.show()

  def Test1():
    shape=(700,500)
    #data=np.ndarray(buffer=np.random.bytes(4*np.prod(shape)),shape=shape,dtype=np.uint32)
    data=np.uint32(500*np.random.rand(np.prod(shape)).reshape(shape))
    #data=np.random.random_integers(0,1000,shape)
    roiMaxSz=1006 #maximal pixel counts per roi
    nRoi=5000 #number of rois
    n=np.prod(data.shape) #number of points in the data
    roiLenArr=np.uint32(roiMaxSz*np.random.rand(nRoi))
    roiSz=roiLenArr.sum()
    roiIdxArr=np.uint32(n*np.random.rand(roiSz))

    procRoiStat=ProcRoiStat()
    wgtArr=None
    procRoiStat.SetRoi(roiIdxArr,roiLenArr,wgtArr)
    print 'numnber of ROI,',nRoi,'Total number of pixels',roiSz
    procRoiStat.SetProcess('avg',useLib=False)
    ut.StopWatch.Start()
    procRoiStat.Process(data)
    t0=ut.StopWatch.Log('Process useLib=False')
    print procRoiStat.resArr[:,procRoiStat.roiLenArr>0]
    procRoiStat.SetProcess('avg')
    ut.StopWatch.Start()
    procRoiStat.Process(data)
    t1=ut.StopWatch.Log('Process')
    print procRoiStat.resArr[:,procRoiStat.roiLenArr>0]
    print 'Process with library implementation %f faster than pure Python'%(t0/t1)


  def Test2():
    'process an existing mat file of some hdf images (NOT CORRECT CENTER)'
    clpIn=ClampIn(('hdf','/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne/scan_00106-00161.hdf5','/entry/data/pilatus_1'))
    data=clpIn.GetData(0)
    procRoiStat=ProcRoiStat()
    procRoiStat.SetRoiMat('/scratch/detectorData/cSAXS_2013_10_e14608_georgiadis_3D_for_Marianne/analysis/data/pilatus_integration_mask.mat',data.shape,True)

    import pylab as plt #used for the colormaps
    procRoiStat.SetProcess('avg',useLib=False)
    print 'numnber of ROI,',procRoiStat.roiLenArr.size,'Total number of pixels',procRoiStat.roiIdxArr.size

    while True:
      ut.StopWatch.Start()
      data=clpIn.GetData()
      if data==None:
        break
      ut.StopWatch.Log('getImg')
      procRoiStat._ProcessAvg(data)
      t0=ut.StopWatch.Log('Process')
      print procRoiStat.resArr.ravel()[procRoiStat.roiLenArr>0]
      ut.StopWatch.Start()
      procRoiStat._ProcessAvg_C(data)
      t1=ut.StopWatch.Log('Process_C')
      print procRoiStat.resArr.ravel()[procRoiStat.roiLenArr>0]
      print 'Process_C %f faster than Process'%(t0/t1)
      plt.figure()
      plt.imshow(np.log(1+procRoiStat.resArr),interpolation='nearest',aspect='auto',vmax=5)
      plt.show()
    print 'done'
    return

#------------------ Main Code ----------------------------------
  ParseArgs()
  #Test0()
  #Test1() #C-implementation 16-17 x faster
  #Test2() #C-implementation 41-42 x faster
  #import cProfile
  #cProfile.run('Test2()')
