#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
Processing SasFit data.

The input must be radial integrated scattering images.
This data is:
- converted to intensity, Q values
- written to an output HDF that can be processed by the SasFit program.

It is (will be) also possible to process fast basic structural analysis of the radial integrated data.

The class can process the jobs using clamp classes and Single/MultiProcessing
topology.
'''

import logging, sys, os, json
import numpy as np
import utilities as ut
import h5py, zmq
import scipy.io
from Clamp import ClampIn,ClampOut,MultiProc,SingleProc
from libDetXR import LibDetXR
LibDetXR()
try:
  import matplotlib as mpl
  #mpl.use('WxAgg')
  import pylab as plt
except:
  pass

_log=logging.getLogger(__name__)

def LogOut(s):
  log=logging.root.handlers[0].stream
  log.write(s)
  log.flush()

class MetaSas:
  def __init__(self,matRoi, detDist=7113.5, energy=12.4, pixSz=0.172):
    detDist=7113.5 #detector distance in mm
    energy=12.4  #energy in keV
    pixSz=0.172  #pixel size in mm
    mat=scipy.io.loadmat(matRoi)
    self.mskRadius=mskRadius=mat['integ_masks'][0,0]['radius'].reshape(-1)
    mskNormSum=mat['integ_masks'][0,0]['norm_sum']
    if len(mskNormSum.shape)>1 and mskNormSum.shape[1]>1:
      self.mskNormSum=mskNormSum.sum(axis=1,dtype=np.uint16)
    else:
      self.mskNormSum=mskNormSum.T.reshape(-1)
    self.q=ProcSasFit.Pix2Q(mskRadius,pixSz,detDist,energy)
    self.NormSum=np.sqrt(mskRadius)

class ProcSasFit():
  mode=('sasfit','struct')
  def __init__(self,args):
    self.args=args
    self.meta=MetaSas(args.matRoi)
    if args.procmode=='sasfit':
      self.Process=self.ProcessSasFit
    elif args.procmode=='struct':
        self.Process=self.ProcessStructParam
        param=args.param
        self.prdStart=param['porod'][0]
        self.prdNum  =param['porod'][1]
        self.gnrStart=param['guinier'][0]
        self.gnrNum  =param['guinier'][1]
    else:
      _log.warning('not supported processing mode')

  @staticmethod
  def Pix2Q(pix, pixSz, detDist, energy):
    '''calculated momentum transfer q in inverse Angstroem from pixel numbers
       relative to the beam center'''
    #pix: distance in pixel
    #pixSz: pixel size in mm
    #detDist: detector distance in mm
    #energy: energy in keV
    #import math
    lambda_A = 12.39852 / energy;
    #q_A = 4*math.pi * math.sin( math.atan(pix*pixSz/detDist)/2)/lambda_A;
    #q_A = 4*np.pi * np.sin( np.arctan(pix*pixSz/detDist)/2)/lambda_A;
    q_A = np.float32(4*np.pi * np.sin( np.arctan(pix*pixSz/detDist)/2)/lambda_A);
    return q_A

  def ProcessSasFit(self,idx,data):
    m=self.meta
    idxD=np.flatnonzero(m.mskNormSum)
    I_avg=data[0,idxD]
    maxnum=m.mskNormSum[idxD].max()
    I_avg[I_avg==0]=.5/maxnum
    #I_std=data[1,idxD]
    #I_err=I_std/m.NormSum[idxD];
    I_err=1./m.NormSum[idxD];
    n=len(idxD)
    res=np.ndarray((n,3),np.float32)
    res[:,0]=m.q[idxD]
    res[:,1]=I_avg
    res[:,2]=I_err
    return res

  def ProcessStructParam(self,idx,data):
    m=self.meta
    idxD=np.flatnonzero(m.mskNormSum)
    q=m.q[idxD]
    I_avg=data[0,idxD]
    maxnum=m.mskNormSum.max()
    I_avg[I_avg==0]=.5/maxnum
    #I_std=data[1,idxD]
    #I_err=I_std/m.NormSum[idxD];
    I_err=1./m.NormSum[idxD];

    q=q.copy()
    Iq=I_avg.copy()
    IqErr=I_err.copy()
    assert(q.dtype==np.float32)
    assert(Iq.dtype==np.float32)
    assert(IqErr.dtype==np.float32)
    res=np.ndarray((47),np.float32);res[:]=2
    #if idx!=9:return res
    #def ProcessPorodGuinier(prdNum, prdStart, gnrNum, gnrStart, q, Iq, IqErr, qCnt, res):
    #LibDetXR.ProcessPorodGuinier(30,1,12,1,q.ctypes.data,Iq.ctypes.data,IqErr.ctypes.data,len(Iq),res.ctypes.data)
    LibDetXR.ProcessPorodGuinier( self.prdStart,self.prdNum,self.gnrStart,self.gnrNum,q.ctypes.data,Iq.ctypes.data,IqErr.ctypes.data,len(Iq),res.ctypes.data)
    #if idx==9:
    #  exit()
    return res

  def Close(self):
    pass

class ClampOutSasFit(ClampOut):
  @staticmethod
  def GetCaps(param):
    if param.dst[0]=='hdf':
      return 0
    return 2

  def __init__(self,args,**kwargs):
    ClampOut.__init__(self,**kwargs)
    self.args=args

    if args.dst[0]=='hdf':
      self.InitHDF()
      if args.procmode=='sasfit':
        self.SetData=self._SetDataHDF
      else:
        self.SetData=self._SetDataSmallHDF
    elif args.dst[0]=='zmq':
      self.InitZMQ()
      self.SetData=self._SetDataZMQ
    elif args.dst[0]=='dat':
      self.InitDat()
      self.SetData=self._SetDataDat
    else:
      _log.error('unsupported Out Clamp')

  def InitHDF(self):#,clpIn,proc):
    args=self.args
    fnHdf5=args.dst[1]
    self.fid=ut.OpenFile(fnHdf5,'w')
    self.idxMax=0

  def InitZMQ(self):
    args=self.args.dst[1]
    if type(args)==str: args=json.loads(args, object_hook=ut.ConvUtf8)
    _log.info('open socket'+str(args))
    self.sock=ut.ZMQ.Socket(args,zmq.PUSH,4)

  def InitDat(self):
    pass

  def _SetDataHDF(self,idx,outBuf):
    args=self.args
    try: dsObj=self.dsObj
    except AttributeError:
      cmpr=self.args.cmpr
      info=self.ClpSrcRPC(0, 'GetInfo')
      try:
        num=info['cnt']
        _log.info('number of Images: %d'%num)
      except (KeyError,TypeError):
        num=512
        _log.info('could not determine number of frames')
      res,v=LibDetXR.GetVersion()
      attr=(('creator','Converted data with '+os.path.basename(__file__)),
            ('version', 'LibDetXR  %d.%d.%d.%d'%tuple(v[::-1])),
            ('source',json.dumps(args.src)),
            ('matRoi',args.matRoi),
            ('procmode',args.procmode),
            ('NeXus_version',"?.?.?"))
      #grpName=args.dst[2]
      grpName,dsName=args.dst[2].rsplit('/',1)
      gid=ut.CreateGrpPath(self.fid,grpName)
      ut.AddAttributes(gid,attr)

      if cmpr:
        self.CmprChunkBuild=ut.GetChunkBuildFunc(args.cmpr)
      dtype=outBuf.dtype
      shape=(num,)+outBuf.shape
      maxshape=(h5py.h5s.UNLIMITED,)+outBuf.shape
      chunk=(1,)+outBuf.shape
      self.dsObj=dsObj=ut.DsObj(ut.CreateDatasetChunked(gid,dsName,dtype,shape,maxshape,chunk,cmpr),shape,chunk)
    dsObj.MaxReshape(idx)
    chkId=(idx,)+(0,)*outBuf.ndim
    try: CmprChunkBuild=self.CmprChunkBuild
    except AttributeError:
      dsObj.ds.write_chunk(chkId,outBuf)
      #LogOut('%d bytes\n'%outBuf.nbytes)
    else:
      (cmprBuf,filter)=CmprChunkBuild(outBuf)
      dsObj.ds.write_chunk(chkId,cmprBuf,filters=filter)
      #LogOut('%d bytes filter %d\n'%(cmprBuf.nbytes,filter))
    if self.args.verbose&2:#print 'SetData',idx,res;sys.stdout.flush()
      LogOut('.')

  def _SetDataSmallHDF(self,idx,outBuf):
    args=self.args
    try: resArr=self.resArr
    except AttributeError:
      self.resArr=resArr=np.ndarray(shape=(512,)+outBuf.shape,dtype=outBuf.dtype)
    try: resArr[idx,:]=outBuf
    except IndexError:
      sh=resArr.shape;sh=(sh[0]+512,)+sh[1:]
      resArr.resize(sh,refcheck=False)
      resArr[idx,:]=outBuf
    if idx>self.idxMax:
      self.idxMax=idx
    if self.args.verbose&2:#print 'SetData',idx,res;sys.stdout.flush()
      LogOut('.')

  def _SetDataZMQ(self,idx,outBuf):
    sock=self.sock
    try:
      header=self.header
    except AttributeError:
      header ={'htype':'chunk-1.0','shape':tuple(reversed(outBuf.shape)),'type':str(outBuf.dtype)}
    header['frame']=idx
    sock.send_json(header, zmq.SNDMORE) #@UndefinedVariable
    sock.send(outBuf)
    if self.args.verbose&2:#print 'SetData',idx,res;sys.stdout.flush()
      LogOut('.')

  def _SetDataDat(self,idx,outBuf):
    np.savetxt(self.args.dst[1]%idx, outBuf, fmt='%.6g',delimiter='\t')
    if self.args.verbose&2:#print 'SetData',idx,res;sys.stdout.flush()
      LogOut('.')

  def WriteHDF(self):
    args=self.args
    cmpr=args.cmpr
    fid=self.fid
    outBuf=self.resArr
    grpName,dsName=args.dst[2].rsplit('/',1)
    gid=ut.CreateGrpPath(fid,grpName)
    res,v=LibDetXR.GetVersion()
    attr=[('creator',os.path.basename(__file__)),
          ('version', 'LibDetXR  %d.%d.%d.%d'%tuple(v[::-1])),
          ('source',json.dumps(args.src)),
          ('procmode',args.procmode),
          ]
    if args.matRoi!=None:
      attr.append(('matRoi',args.matRoi))
    if args.procmode=='struct':
      #I0chisq,I0,DI0,RG,DRG,prdChisq,c0,Dc0,c4,Dc4,Qmin,Qmax,Qq0,DQq0,iIq0,DiIq0,intIq0,DintIq0,intIexp,DintIexp,intIc4,DintIc4,intI,DintI,Qexp,DQexp,Qc4,DQc4,Invariant,DInvariant,iIexp,DiIexp,iI,DiI,iIc4,DiIc4,S_V,DS_V,lc,Dlc,Ac,DAc,li,Dli,VP,DVP,RP,DRP,
      attr.append(('contains','porod-guinier-approximation data'))
      attr.append(('header',json.dumps(('guinier I0chisq','guinier I0','guinier d_I0','guinier Rg','guinier d_Rg',
                                       'porpd chisq','porpd c0','porpd d_c0','porpd c4','porpd d_c4,',
                                       'Qmin','Qmax',
                                       'Qq0','DQq0','iIq0','DiIq0','intIq0','DintIq0','intIexp','DintIexp','intIc4','DintIc4','intI','DintI','Qexp','DQexp','Qc4','DQc4',
                                       'Invariant','DInvariant','iIexp','DiIexp','iI','DiI','iIc4','DiIc4',
                                       'S_V','DS_V','lc','Dlc','Ac','DAc','li','Dli','VP','DVP','RP','DRP'))))
    if args.param!=None:
      attr.append(('param',json.dumps(args.param)))



    sh=outBuf.shape;sh=(self.idxMax+1,)+sh[1:]
    outBuf.resize(sh,refcheck=False)

    dtype=outBuf.dtype
    shape=outBuf.shape
    maxshape=outBuf.shape
    chunk=outBuf.shape
    dsObj=ut.DsObj(ut.CreateDatasetChunked(gid,dsName,dtype,shape,maxshape,chunk,cmpr),shape,chunk)
    ut.AddAttributes(dsObj.ds,attr)
    if cmpr:
      CmprChunkBuild=ut.GetChunkBuildFunc(cmpr)
      (cmprBuf,filter)=CmprChunkBuild(outBuf)
      dsObj.ds.write_chunk((0,0),cmprBuf,filters=filter)
      if args.verbose&64:
        LogOut('shape %s, %d bytes filter %d\n'%(str(shape),cmprBuf.nbytes,filter))
    else:
      dsObj.ds.write_chunk((0,0),outBuf)
      if args.verbose&64:
        LogOut('shape %s, %d bytes\n'%(str(shape),outBuf.nbytes))
    ut.CloseFile(fid)
    print maxshape


  def Close(self):
    if self.SetData.im_func==ClampOutSasFit._SetDataHDF.im_func:
      ut.CloseFile(self.fid)
    elif self.SetData.im_func==ClampOutSasFit._SetDataSmallHDF.im_func:
      self.WriteHDF()
    elif self.SetData.im_func==ClampOutSasFit._SetDataZMQ.im_func:
      self.sock.close()
      _log.info('close socket'+str(self.args.dst))

    print 'Close PID %d'%os.getpid()
    #self.fh.close()

def FigSetPos(fig,x,y):
  be=mpl.backends.backend.lower()
  if be=='TkAgg'.lower():
    fig.canvas.manager.window.wm_geometry("+%d+%d"%(x,y))
  elif be=='WXAgg'.lower():
    fig.canvas.manager.window.SetPosition((x,y))

class ShowSasHDF:
  def __init__(self,hdfFile,sz,argv):
    '''shows hdf file guinier-porod approx. using procmode struct'''
    hdfEntry='entry/sasfit'

    fid = h5py.h5f.open(hdfFile,flags=h5py.h5f.ACC_RDONLY)
    hid = h5py.h5o.open(fid,hdfEntry)
    self.dsRst=dsRst=h5py.Dataset(hid)
    w=sz[0]
    h=dsRst.shape[0]/w

    try:
      matRoi=dsRst.attrs['matRoi']
    except: _log.warning('cannot find a matRoi in hdf source')
    else:
      self.meta=m=MetaSas(matRoi)
    try:
      src=json.loads(dsRst.attrs['source'], object_hook=ut.ConvUtf8)
      if type(src)==list:src=ut.ConvUtf8(src)
      param=json.loads(dsRst.attrs['param'], object_hook=ut.ConvUtf8)
      m.gRng=param['guinier'];m.pRng=param['porod']
      idx=np.flatnonzero(m.mskNormSum)
      m.gRng=[idx[m.gRng[0]],idx[m.gRng[0]+m.gRng[1]]]
      m.gRng=slice(m.gRng[0],m.gRng[1])
      m.pRng=[idx[-m.pRng[1]-m.pRng[0]],idx[-1-m.pRng[0]]]
      m.pRng=slice(m.pRng[0],m.pRng[1])
    except: _log.warning('cannot find a procRoiStat hdf source')
    else:
      try:
        fid = h5py.h5f.open(src[1],flags=h5py.h5f.ACC_RDONLY)
        hid = h5py.h5o.open(fid,src[2])
      except BaseException as e: _log.warning('cannot open object '+src[1]+' '+src[2]+'\n'+repr(e))
      else:
        self.dsFit=dsFit=h5py.Dataset(hid)
        try:
          src=json.loads(dsFit.attrs['source'], object_hook=ut.ConvUtf8)
          if type(src)==list:src=ut.ConvUtf8(src)
        except: _log.warning('cannot find a raw data hdf source')
        else:
          try:
            fid = h5py.h5f.open(src[1],flags=h5py.h5f.ACC_RDONLY)
            hid = h5py.h5o.open(fid,src[2])
          except BaseException as e: _log.warning('cannot open object '+src[1]+' '+src[2]+'\n'+repr(e))
          else:
            self.dsDif=dsDif=h5py.Dataset(hid)
      pass

    hdr=json.loads(dsRst.attrs['header'], object_hook=ut.ConvUtf8)
    if type(hdr)==list:hdr=ut.ConvUtf8(hdr)
    self.hdr=hdr
    self.col=col=int(argv[0])
    self.figRst=figRst=plt.figure(0)
    a=dsRst[:(w*h),col].reshape((h,w))
    ax=figRst.gca()
    figRst.img=img=ax.imshow(a,interpolation='nearest')
    ax.invert_xaxis()
    ax.set_title(hdr[col]+' (%d)'%col)
    ax.format_coord=self.OnFmtCoord

    figRst.colBar=figRst.colorbar(img, ax=ax)
    figRst.canvas.set_window_title('raster scan')
    figRst.canvas.mpl_connect('motion_notify_event', self.OnMotion)
    figRst.canvas.mpl_connect('button_press_event',  self.OnBtnPress)
    figRst.canvas.mpl_connect('button_release_event',self.OnBtnRelease)
    FigSetPos(figRst,100,0)

    try: dsFit
    except UnboundLocalError: pass
    else:
      self.figFit=figFit=plt.figure(1)
      figFit.canvas.set_window_title('porod-guinier fit')
      ax=figFit.gca()
      figFit.lMsr=lMsr=ax.plot(m.q,dsFit[0,0,:],'.')
      figFit.lPrd=ax.plot(m.q[m.pRng],dsFit[0,0,m.pRng]/2)
      figFit.lGnr=ax.plot(m.q[m.gRng],dsFit[0,0,m.gRng]/2)

      #Porod approx.:   I(Q)= c0 + c4 * Q^(D-4)
      #Guinier approx.: I(Q)= I0 * exp(Q^2*Rg^2/3)
      #(I0chisq,I0,d_I0,Rg,d_Rg,chisq,c0,d_c0,c4,d_c4,D,d_D)=dsRst[0,:]
      #q=m.q[-600:];figFit.lPrd=ax.plot(q,c0+c4*q**(D-4))
      #q=m.q[ :100];figFit.lGnr=ax.plot(q,I0*np.exp(q**2*Rg**2/3))

      ax.set_xscale('log');ax.set_xlabel('q')
      ax.set_yscale('log');ax.set_ylabel('I(q)')
      #ax.axis('tight')
      ax.set_ylim(np.nanmin(dsFit[0,0,:]),np.nanmax(dsFit[0,0,:])*2)
      ax.set_xlim(m.q[0],m.q[-1])
      figFit.hParam=figFit.text(0.85, 0.8,
                    'click on raster scan image to select\n'+
                    'click outside axis to change the variable\n'+
                    'drag colorbar (left and right mouse) to change scale', ha='right')
      FigSetPos(figFit,100,600)

    try: dsDif
    except UnboundLocalError: pass
    else:
      self.figDif=figDif=plt.figure(2)
      figDif.canvas.set_window_title('diffraction image')
      data=dsDif[0,...]
      ax=figDif.gca()
      avg=np.average(data); std=np.std(data)
      vmin=np.min(data);vmax=np.max(data)
      vmin=max(vmin,avg-3*std);vmax=min(vmax,avg+3*std)
      vmin=.1
      figDif.img=img=ax.imshow(data,interpolation='nearest',cmap=mpl.cm.jet, norm=mpl.colors.LogNorm(vmin=1, vmax=vmax))
      ax.set_axis_bgcolor('black')
      colBar=figDif.colorbar(img,orientation='vertical')
      ax.invert_xaxis()
      ax.axis('image')
      FigSetPos(figDif,800,0)

    pass

  def OnFmtCoord(self,x, y):
    data=self.figRst.img.get_array()
    x=int(round(x));  y=int(round(y))
    try:
      v=data[y,x]
      return 'x:%d y:%d idx:%d val:%g'%(x,y,y*data.shape[1]+x,v)
    except IndexError as e:
      return str(e)

  def OnMotion(self,event):
    fig=event.canvas.figure
    colBar=fig.colBar
    if event.inaxes==colBar.ax:
      #print 'OnMotion',fig,event.inaxes
      pt=colBar.ax.bbox.get_points()[:,1]
      nrm=colBar.norm
      vmin,vmax,p0,p1,pS = (nrm.vmin,nrm.vmax,pt[0],pt[1],event.y)
    try:
      vmin,vmax,p0,p1,pS=fig.colBarPressed
    except AttributeError:
      pass
    else:
      if event.button==1:#move top,bottom,both
        pD = event.y - pS
        vD=(vmax-vmin)/(p1-p0)*(pS-event.y)
        colBar.norm.vmin = vmin+vD
        colBar.norm.vmax = vmax+vD
      elif event.button==3:#scale around point
        scale= np.exp((pS-event.y)/100)
        vS=vmin+(vmax-vmin)/(p1-p0)*(pS-p0)
        #print scale,vS
        colBar.norm.vmin = vS-scale*(vS-vmin)
        colBar.norm.vmax = vS-scale*(vS-vmax)
      fig.img.set_norm(colBar.norm)#force image to redraw
      colBar.patch.figure.canvas.draw()
    #    if event.inaxes==self.ax:

  def OnBtnPress(self,event):
    """on button press we will see if the mouse is over us and store some data"""
    fig=event.canvas.figure
    colBar=fig.colBar
    if event.inaxes==colBar.ax:
      pt=colBar.ax.bbox.get_points()[:,1]
      nrm=colBar.norm
      fig.colBarPressed = (nrm.vmin,nrm.vmax,pt[0],pt[1],event.y)
    elif event.inaxes==fig.img.axes:
      x=int(round(event.xdata)); y=int(round(event.ydata))
      if event.button==1:
        data=self.figRst.img.get_array()
        idx=y*data.shape[1]+x
        try:
          dsFit=self.dsFit
        except AttributeError:pass
        else:
          m=self.meta
          data=dsFit[idx,0,:]
          data[data==0]=.5/m.mskNormSum.max()

          lMsr=self.figFit.lMsr[0]
          lPrd=self.figFit.lPrd[0]
          lGnr=self.figFit.lGnr[0]
          ax=lMsr.axes
          lMsr.set_ydata(data)
          #Porod approx.:   I(Q)= c0 + c4 * Q^(D-4)
          #Guinier approx.: I(Q)= I0 * exp(Q^2*Rg^2/3)
          #(I0chisq,I0,d_I0,Rg,d_Rg,chisq,c0,d_c0,c4,d_c4,D,d_D)=self.dsRst[idx,:]
          ep=self.dsRst[idx,:]
          I0=ep[1];c0=ep[6];c4=ep[8];Rg=ep[3];D=0.
          q=m.q[m.pRng];lPrd.set_ydata(c0+c4*q**(D-4))
          q=m.q[m.gRng];lGnr.set_ydata(I0*np.exp(-q**2*Rg**2/3))
          ax.set_title('image %d'%idx)
          self.figFit.hParam.set_text('Porod c0:%g c4:%g\nGuinier I0:%g Rg:%g\n'%(c0,c4,I0,Rg))
          lMsr.figure.canvas.draw_idle()
        try:
          dsDif=self.dsDif
        except AttributeError:pass
        else:
          data=dsDif[idx,...]
          img=self.figDif.img
          ax=img.axes
          img.set_array(data)
          ax.set_title('image %d'%idx)
          img.figure.canvas.draw_idle()
    else:
      dsRst=self.dsRst
      figRst=self.figRst
      colBar=figRst.colBar
      shape=figRst.img.get_array().shape
      if event.button==2:
        #do some own fancy calc
        a=dsRst[:np.prod(shape),0].reshape(shape)
        figRst.img.axes.set_title('own calculation')
      else:
        delta=1 if event.button==1 else -1
        self.col=col=(self.col+delta)%dsRst.shape[1]
        a=dsRst[:np.prod(shape),col].reshape(shape)
        figRst.img.axes.set_title(self.hdr[col]+' (%d)'%col)

      figRst.img.set_array(a)
      lim=(a.min(),a.max())
      if not np.isfinite(lim).all():
        msk=a[np.isfinite(a)]
        lim=(msk.min(),msk.max())
      if lim[0]==lim[1]:
        lim=(lim[0]-1e-6,lim[0]+1e-6)
      figRst.colBar.set_clim(lim)
      fig.img.set_norm(colBar.norm) #force image to redraw
      figRst.canvas.draw_idle()
      pass

  def OnBtnRelease(self,event):
    """on release we reset the press data"""
    fig=event.canvas.figure
    try: del fig.colBarPressed
    except AttributeError: pass

def ShowSasCSV(fn,sz,argv):
  '''shows commaseparated values file: result of sasfit guinier-porod approx.
     argv indicates which information has to be displayed (values 0..48)'''

  # show /scratch/detectorData/cSAXS_2014_07_e15291_ott_stone/isp1.csv 201 201 6
  # show /scratch/detectorData/cSAXS_2014_07_e15291_ott_stone/isp1.csv 201 201 6  8 22 24

  #content of the CSV File is
  #  0 runNr                         16  error                        32  error
  #  1 filename                      17  linear Guinier approx chisq  33  intersection length li
  #  2 1st data point (ascending)    18  1st data point (descending)  34  error
  #  3 number of points              19  number of points             35  Porod volume VP
  #  4 Q_min                         20  Q_min                        36  error
  #  5 Q_max                         21  Q_max                        37  Porod radius RP
  #  6 Guinier radius RG             22  background c_0               38  error
  #  7 error                         23  error                        39  specific surface S/V
  #  8 forward scattering I(0)       24  Porod constant c_4           40  error
  #  9 error                         25  error                        41  scattering invariant Inv
  # 10 chisq(Guinier)                26  potential law D              42  error
  # 11  linear Guinier approx I(0)   27  error                        43  integrated intensity iI = int(I(Q)*Q)dQ
  # 12  error                        28  chisq(Porod)                 44  error
  # 13  linear Guinier approx RG     29  correlation length lc        45  integral intensity intI = int(I(Q))dQ
  # 14  error                        30  error                        46  error
  # 15  linear Guinier approx RG^2   31  correlation area Ac          47  Q_min
  #                                                                   48  Q_max

  #arr=np.genfromtxt(fn, names=True, delimiter=';')
  #print arr.dtype
  #print arr.names

  f = open(fn, 'r')  # We need to re-open the file
  hdr=f.readline().split(';')
  f.close()
  cols=map(int,argv)
  arr=np.loadtxt(fn,delimiter=';', skiprows=1, usecols=cols)
  w=sz[0]
  h=arr.shape[0]/w
  for idx,col in enumerate(cols):
    plt.figure()
    if len(cols)>1:
      a=arr[:(w*h),idx].reshape((h,w))
    else:
      a=arr[:(w*h)].reshape((h,w))
    hi=plt.imshow(a,interpolation='nearest')
    plt.title(hdr[col])
  #np.savetxt(self.args.dst[1]%idx, outBuf, fmt='%.6g',delimiter='\t')
  pass

def RunWithArgs(args):
  if type(args)==dict:
    args=ut.dotdict(args)
    default={
            'numFrm':sys.maxint,
            'matRoi':None,
            'dst':None,
            'multiproc':None,
            'cmpr':None,
            'param':None,
            'verbose':3
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
    if args.multiproc==True: args.multiproc=[3]
    _log.setLevel(logging.INFO if args.verbose&1 else logging.WARNING)
  else:
    args=ut.dotdict(args.__dict__)
  for clp in (args.src,args.dst):
    if clp[0] in ('zmq','zmqMsg'):
      clp[1]=ut.ZMQ.lst2dict(clp[1:])
      del clp[2:]

  if args.procmode=='struct':
    try:
      d={}
      for v in args.param:
        k,v=v.split(':',1)
        v=map(int,v.split(','))
        d[k]=v
      args.param=d
    except ValueError as e:
      _log.error('wrong parameters:'+str(e))
      return
  if args.multiproc!=None:
    MultiProc(ClampIn,ProcSasFit,ClampOutSasFit,args)
  else:
    SingleProc(ClampIn,ProcSasFit,ClampOutSasFit,args)

if __name__ == '__main__':
  def LogConfig(args):
    logging.basicConfig()
    hdlr=logging.StreamHandler()
    fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
    hdlr.setFormatter(fmt)
    del logging.root.handlers[:]
    logging.root.addHandler(hdlr)
    logging.root.setLevel(logging.DEBUG)

  def ParseArgs():
    if len(sys.argv)>1 and sys.argv[1]=='show':
      LogConfig(None)
      import pylab as plt
      fn=sys.argv[2]
      szX=int(sys.argv[3])
      szY=int(sys.argv[4])
      if h5py.h5f.is_hdf5(fn):
        ShowSasHDF(fn,(szX,szY),sys.argv[5:])
      else:
        ShowSasCSV(fn,(szX,szY),sys.argv[5:])
      plt.show()
      return

    import argparse #since python 2.7
    (h,t)=os.path.split(sys.argv[0]);cmd='\n  '+(t if len(h)>3 else sys.argv[0])+' '
    exampleCmd=('--src hdf $BASE/roistatNoSect_01465-01665.hdf5 entry/data/roiStat --matRoi $BASE/analysis/data/pilatus_integration_mask_nosectors.mat --dst dat $BASE/sasFit/out%.6d.dat --procmode sasfit -v',
                '--src hdf $BASE/roistatNoSect_01465-01665.hdf5 entry/data/roiStat --matRoi $BASE/analysis/data/pilatus_integration_mask_nosectors.mat --dst hdf $BASE/sasFitStruct_01465-01665.hdf5 entry/sasfit --procmode sasfit -v',
                '--src hdf $BASE/roistatNoSect_01465-01665.hdf5 entry/data/roiStat --matRoi $BASE/analysis/data/pilatus_integration_mask_nosectors.mat  --dst hdf $BASE/sasFitStruct_01465-01665.hdf5 entry/sasfit --procmode struct --param porod:0,600 guinier:0,15 -m',
                '--src zmq addr:tcp://localhost:8080 qs:4 if:PULL',
                'show $BASE/sasFitStruct_01465-01665.hdf5 201 201 1',
                'show filename width height [other params]'
                )
    epilog='''
the verbose bits are:
   bit0(0x01): basic
   bit1(0x02): progress
   bit2(0x04):
   bit3(0x08):
   bit4(0x10):
   bit5(0x20):
   bit6(0x40):
   bit6(0x80):
 - bit8(0x100): multiprocessing+clamps
 - bit9(0x200): multiprocessing+clamps (more verbose)
   default is 1+2=3 (no -v)
   -v is 1+4+8+16=511
   -v0 is turn all off

Data Source description:'''+ClampIn.__types__+'''
Data Destination description:
  HDF file from a given object:
    --dst hdf FILE ELEM
    --dst hdf myfile.hdf5 entry/mygrp/dataset
  DAT files: (used with procmode==sasfit)
    --dst dat FILEPATTERN (one % that will be printf the index number)
    --dst dat /tmp/out%.6d.dat
  ZMQ stream
    --dst zmq JSON-config (with addr and optional queueSz, ifType and timeout)
    --dst zmq '{"addr":"tcp://localhost:8080","queueSz":4,"ifType":"PULL"}',

Processing modes: procmode
  sasfit: generates output (normally *.dat) files that can be processed with SasFit (http://sourceforge.net/projects/sasfit/)
          params: unused
  struct: process porpd-guinier-approximation
          params: porod":start,count guinier:start,count
          e.g. 'porod:1,100 guinier:1,15'

To show result use '''+cmd+''' show filename width height [other params]
This can show:
 - hdf5 files (with procmode=struct)
 - or sasfit .csv files outputs

Examples:
  BASE=/scratch/detectorData/cSAXS_2014_07_e15291_ott_stone'''+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__, epilog=epilog)
    parser.add_argument('--src',   nargs='+', required=True, help='data source (see below for examples)')
    parser.add_argument('--matRoi',                          help='MATLAB integration mask file (e.g. pilatus_integration_mask.mat)') #used for radius and number pixels per ROI
    parser.add_argument('--dst',  nargs='+', metavar=('mode','param'), help='data destination: hdf or zmq sink')
    parser.add_argument('--procmode', choices=ProcSasFit.mode, default=ProcSasFit.mode[0], help='processing method to use')
    parser.add_argument('--param', nargs='+', help='processing parameters')
    #for integParam {'guinier':[1 15], 'zimm':[1 15], 'porod':[1 100]}
    parser.add_argument('-m', '--multiproc', nargs='?', const=[3], default=None,help='use multiprocessing method topo,[numproc]')
    parser.add_argument('--cmpr',                                               help='compression method (zlib[lvl],lz4,lzf,cbf)')
    parser.add_argument('-v',         dest='verbose',  type=int,  nargs='?', const=267,default=3,  help='verbose level (see below)')

    args = parser.parse_args()
    if type(args.multiproc)==str:
      args.multiproc=map(int,args.multiproc.split(','))
      assert(len(args.multiproc)<=2)
    LogConfig(args)
    RunWithArgs(args)

  def GetQ():
    (radius,NormSum)=GetRadiusNormSum()
    detDist=7113.5 #detector distance in mm
    energy=12.4  #energy in keV
    pixSz=0.172  #pixel size in mm
    q=ProcSasFit.Pix2Q(radius,pixSz,detDist,energy)
    return q

  def GetRadiusNormSum():
    fn='/scratch/detectorData/cSAXS_2014_07_e15291_ott_stone/analysis/data/pilatus_integration_mask_nosectors.mat'
    mat=scipy.io.loadmat(fn)
    mskRadius=mat['integ_masks'][0,0]['radius'].reshape(-1)
    mskNormSum=mat['integ_masks'][0,0]['norm_sum'].T.reshape(-1)

    return (mskRadius,mskNormSum)

  def TestRadInteg():
    fn='/scratch/detectorData/cSAXS_2014_07_e15291_ott_stone/analysis/radial_integration/e15291_1_01465_00000_00000_integ.mat'
    mat=scipy.io.loadmat(fn)
    #mat.keys()
    #['__version__', 'I_all', '__header__', '__globals__', 'radius', 'I_std', 'filenames_all', 'norm_sum']
    I_all=mat['I_all'] #(1216, 16, 201)
    radius=mat['radius']#(1, 1216)
    I_std=mat['I_std']#(1216, 16, 201)
    norm_sum=mat['norm_sum']#(1216, 16)
    I_all=mat['I_all'][:,:,50]
    I_std=mat['I_std'][:,:,50]

    print I_all[10:15,:].T
    print I_std[10:15,:].T
    print 'compare these values to the file roistat_01465-01665.hdf5 frame 50'
    print '-> it seems radial integ does not calculate the std deviation!!!'

    pass

  def Test0():
    #s.a. procRoiStat SetRoiMat(self, fn, show=False)
    fnOut='/scratch/detectorData/cSAXS_2014_07_e15291_ott_stone/sasFit/out%.6d.dat'
    fn='/scratch/detectorData/cSAXS_2014_07_e15291_ott_stone/roistatNoSect_01465-01665.hdf5'
    path='entry/data/roiStat'
    fid = h5py.h5f.open(fn,flags=h5py.h5f.ACC_RDONLY)
    dsid=h5py.h5d.open(fid,path)
    ds=h5py.Dataset(dsid)

    q=GetQ()
    (radius,NormSum)=GetRadiusNormSum()

    #for i in xrange(0,ds.shape[0]):
    #for i in (50,):
    for i in range(0,50):

      I_avg=ds[i,0,:]
      I_std=ds[i,1,:]

      idx=np.flatnonzero(I_avg)
      n=len(idx)

      I_err=I_std[idx]/np.sqrt(NormSum[idx]);

      res=np.ndarray((n,3),np.float32)
      res[:,0]=q[idx]
      res[:,1]=I_avg[idx]
      res[:,2]=I_err
      np.savetxt(fnOut%i, res, fmt='%.6g',delimiter='\t')
      print i

    plt.figure()
    plt.plot(q)
    plt.figure()
    plt.plot(ds[50,0,:])
    plt.figure()
    plt.plot(ds[50,1,:])
    plt.show()


#------------------ Main Code ----------------------------------
  ParseArgs()
  #TestRadInteg()
  #Test0()

