#!/usr/bin/env python
'''
Test performance:
all source data is first read into ram, then different measurements can be performed:
cmpSpeed -> measure compression speed (single process)
wrHDF    -> write HDF5 file, single or multiprocess with different interprocess communications
imgStat  -> performs image statistics of the input images
'''
import logging
import  sys, os, time
import numpy as np
import ctypes as ct
import h5py
import zlib, re
import multiprocessing as mproc
from libDetXR import LibDetXR
import utilities as ut
from Clamp import ClampIn,ClampOut,ClampLst
try:
  from mpi4py import MPI
except ImportError as e:
  print e
try:
  import matplotlib as mpl
  import matplotlib.pyplot as plt
except ImportError as e:
  print e


LibDetXR()
_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

def ChunkBuilderPipe1(ChunkBuildFunc,arr,pipeChild,verbose):
  #used by MultProc.Pipe1, MultProc.Pipe1b
  if verbose&8: _log.info('running PID %d'%os.getpid())
  (arr3,filter)=ChunkBuildFunc(arr)
  pipeChild.send((arr3,filter))
  if verbose&8: _log.info('end PID %d'%os.getpid())
  return 5

def ChunkBuilderPipe2(ChunkBuildFunc,pipeChild,verbose):
  #used by MultProc.Pipe2
  if verbose&8: _log.info('running PID %d'%os.getpid())
  while True:
    arr=pipeChild.recv()
    if arr==None:
      break
    (arr3,filter)=ChunkBuildFunc(arr)
    pipeChild.send((arr3,filter))
  if verbose&8: _log.info('end PID %d'%os.getpid())
  return 5

def ChunkBuilderPool(ChunkBuildFunc,arr,verbose):
  if verbose&8: _log.info('running PID %d'%os.getpid())
  (arr3,filter)=ChunkBuildFunc(arr)
  if verbose&8: _log.info('end PID %d'%os.getpid())
  return (arr3,filter)

def ShMemAlloc(arrNbytes):
  szuint32=4
  sz=szuint32*2 #filter, compressed array size
  sz+=arrNbytes*2        #input array data, compressed array data
  shMem=mproc.RawArray(ct.c_ubyte, sz)
  #print 'ShMemAlloc',sz
  return shMem

def ShMemSetFiltCmprSz(shMem,filter,cmprSz):
  szuint32=4
  shMemAddr=shMem._wrapper.get_address()
  shMemSize=shMem._wrapper.get_size()
  buf = np.core.multiarray.int_asbuffer(shMemAddr, szuint32*2)
  shArr = np.frombuffer(buf, np.uint32)
  shArr[0]=filter
  shArr[1]=cmprSz
  pass

def ShMemCpyArrIn(shMem,arr):
  szuint32=4
  shMemAddr=shMem._wrapper.get_address()
  shMemSize=shMem._wrapper.get_size()
  shMemAddr+=szuint32*2
  shMemSize=(shMemSize-szuint32*2)/2
  buf = np.core.multiarray.int_asbuffer(shMemAddr, shMemSize)
  shArr = np.frombuffer(buf, arr.dtype)
  order = 'C' if arr.flags.c_contiguous else 'F'
  shArr=shArr.reshape(arr.shape,order=order)
  shArr[:]=arr[:] #mem copy
  pass

def ShMemGetCmprResult(shMem):
  szuint32=4
  shMemAddr=shMem._wrapper.get_address()
  shMemSize=shMem._wrapper.get_size()
  buf = np.core.multiarray.int_asbuffer(shMemAddr, szuint32*2)
  shArr = np.frombuffer(buf, np.uint32)
  filter=shArr[0]
  cmprSz=shArr[1]
  cmprAddr=shMemAddr+szuint32*2+(shMemSize-szuint32*2)/2
  buf = np.core.multiarray.int_asbuffer(cmprAddr, cmprSz)
  shCmprArr = np.frombuffer(buf, np.uint8)
  return (shCmprArr,filter)

def ShMemGetArrInArrOut(shMem,dtype,shape):
  szuint32=4
  shMemAddr=shMem._wrapper.get_address()
  shMemSize=shMem._wrapper.get_size()
  shMemAddr+=szuint32*2
  shMemSize=(shMemSize-szuint32*2)/2
  buf = np.core.multiarray.int_asbuffer(shMemAddr, shMemSize)
  shArrIn = np.frombuffer(buf, dtype)
  shArrIn=shArrIn.reshape(shape,order='C')
  buf = np.core.multiarray.int_asbuffer(shMemAddr+shMemSize, shMemSize)
  shArrOut = np.frombuffer(buf, np.uint8)
  #shArrOut=shArrOut.reshape(shape,order='C')
  return (shArrIn,shArrOut)

def ChunkBuilderSharedMem(ChunkBuildFunc,shMem,dtype,shape,verbose):
  if verbose&8: _log.info('running PID %d'%os.getpid())
  #print dtype,shape
  (arrIn,arrOut)=ShMemGetArrInArrOut(shMem,dtype,shape)
  #print arrIn,arrOut
  (cmprArr,filter)=ChunkBuildFunc(arrIn,arrOut)
  ShMemSetFiltCmprSz(shMem,filter,cmprArr.nbytes)
  if verbose&8: _log.info('PID %d filter %d arrCmprSz %d'%(os.getpid(),filter,cmprArr.nbytes))
  return 0

def ChunkBuilderSharedMem2(ChunkBuildFunc,shMem,dtype,shape,pipe):

  while True:
    v=pipe.recv()
    if v==None:
      break
    (arrIn,arrOut)=ShMemGetArrInArrOut(shMem,dtype,shape)
    (cmprArr,filter)=ChunkBuildFunc(arrIn,arrOut)
    ShMemSetFiltCmprSz(shMem,filter,cmprArr.nbytes)
    pipe.send(v)
  return 0

class MultProc(object):

  @staticmethod
  def Pipe1(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''starts <cpu_count> new processes, on per image,
    then waits results of the processes through a pipe.
    when all <cpu_count> processes are finished, it starts <cpu_count> new compression processes'''
    verbose=args.verbose
    jobs = []
    maxJobs=min(mproc.cpu_count(),num)
    for idx in range(num):
      arr=imgLst[idx%numLst]
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=ChunkBuilderPipe1, args=(ChunkBuildFunc,arr,pipeChild,verbose))
      jobs.append((p,pipeParent))
      p.start()
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
      if idx+1==num or (idx+1)%maxJobs==0: #len(jobs)>=maxJobs:
        if idx>=zyx[0]:
          zyx=(idx+10,)+shape
          dsid.set_extent(zyx)
          fspace.set_extent_simple(zyx, ulim)
        ii=maxJobs-1
        for p,pipeParent in jobs:
          (arr3,filter)=pipeParent.recv()
          p.join()
          dsid.write_chunk((idx-ii, 0, 0),arr3,filters=filter)
          if verbose&2: sys.stdout.write('w');sys.stdout.flush()
          ii-=1
        jobs = []

  @staticmethod
  def Pipe1b(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''starts <cpu_count> new processes, on per image,
    then waits results of the processes through a pipe,
    when the first process is finished, it starts a new compression process'''
    verbose=args.verbose
    jobs = []
    maxJobs=min(mproc.cpu_count(),num)
    #prolog
    for idx in range(maxJobs):
      arr=imgLst[idx%numLst]
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=ChunkBuilderPipe1, args=(ChunkBuildFunc,arr,pipeChild,verbose))
      jobs.append((p,pipeParent,idx))
      p.start()
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #loop
    for idx in range(maxJobs,num):
      (p,pipeParent,procIdx)=jobs.pop(0)
      (arr3,filter)=pipeParent.recv()
      p.join()
      if idx>=zyx[0]:
        zyx=(idx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arr3,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()

      arr=imgLst[idx%numLst]
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=ChunkBuilderPipe1, args=(ChunkBuildFunc,arr,pipeChild,verbose))
      jobs.append((p,pipeParent,idx))
      p.start()
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #epilog
    while jobs:
      (p,pipeParent,procIdx)=jobs.pop(0)
      (arr3,filter)=pipeParent.recv()
      p.join()
      if idx>=zyx[0]:
        zyx=(idx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arr3,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()

  @staticmethod
  def Pipe2(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''starts <cpu_count> new processes
    sends images via pipe to all compress processes,
    then waits results of all processes through a pipe,
    The  compress processes are finished if a None image is sent via the pipe'''
    verbose=args.verbose
    jobs = []
    maxJobs=min(mproc.cpu_count(),num)
    for idx in range(maxJobs):
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=ChunkBuilderPipe2, args=(ChunkBuildFunc,pipeChild,verbose))
      jobs.append((p,pipeParent))
      p.start()
    for idx in range(num):
      arr=imgLst[idx%numLst]
      jobs[idx%maxJobs][1].send(arr)
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
      if idx+1==num or (idx+1)%maxJobs==0: #len(jobs)>=maxJobs:
        if idx>=zyx[0]:
          zyx=(idx+10,)+shape
          dsid.set_extent(zyx)
          fspace.set_extent_simple(zyx, ulim)
        ii=maxJobs-1
        for p,pipeParent in jobs:
          (arr3,filter)=pipeParent.recv()
          dsid.write_chunk((idx-ii, 0, 0),arr3,filters=filter)
          if verbose&2: sys.stdout.write('w');sys.stdout.flush()
          ii-=1
    #terminate all child processes
    for p,pipeParent in jobs:
      pipeParent.send(None)
      p.join(1)
      if verbose&2: sys.stdout.write('t');sys.stdout.flush()

  @staticmethod
  def Pipe2b(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''starts <cpu_count> new processes
    sends an image via pipe to all compress processes,
    then waits results of the first processe through a pipe, and give a new job.
    The  compress processes are finished if a None image is sent via the pipe'''
    verbose=args.verbose
    jobs = []
    maxJobs=min(mproc.cpu_count(),num)
    #prolog
    for idx in range(maxJobs):
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=ChunkBuilderPipe2, args=(ChunkBuildFunc,pipeChild,verbose))
      jobs.append((p,pipeParent))
      p.start()
    #loop
    for idx in range(num+maxJobs):
      waitIdx=idx-maxJobs #index of image chunk to wait
      if waitIdx>=0: #>>> receive the job <<<
        if waitIdx>=zyx[0]:
          zyx=(idx+10,)+shape
          dsid.set_extent(zyx)
          fspace.set_extent_simple(zyx, ulim)
        p,pipeParent=jobs[waitIdx%maxJobs]
        (arr3,filter)=pipeParent.recv()
        dsid.write_chunk((waitIdx, 0, 0),arr3,filters=filter)
        if verbose&2: sys.stdout.write('w');sys.stdout.flush()

      if idx<num: #>>> send the job <<<
        arr=imgLst[idx%numLst]
        jobs[idx%maxJobs][1].send(arr)
        if verbose&2: sys.stdout.write('.');sys.stdout.flush()

    #terminate all child processes
    for p,pipeParent in jobs:
      pipeParent.send(None)
      p.join(1)
      if verbose&2: sys.stdout.write('t');sys.stdout.flush()

  @staticmethod
  def ShMem1(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''uses shared memory to transfer image
    one supervising/wringing process
    <cpu_count> compressing process
    process terminates at every chunk compression
    round-robin'''
    verbose=args.verbose
    jobs = []
    maxJobs=mproc.cpu_count()
    #prolog
    for idx in range(maxJobs):
      arr=imgLst[idx%numLst]
      shMem=ShMemAlloc(arr.nbytes)
      ShMemCpyArrIn(shMem,arr)
      #ChunkBuilderSharedMem(ChunkBuildFunc,shMem,arr.dtype,arr.shape)
      #(arrCmpr,filter)=ShMemGetCmprResult(shMem)
      p = mproc.Process(target=ChunkBuilderSharedMem, args=(ChunkBuildFunc,shMem,arr.dtype,arr.shape,verbose))
      jobs.append((p,shMem,idx))
      p.start()
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #loop
    for idx in range(maxJobs,num):
      (p,shMem,procIdx)=jobs.pop(0)
      p.join()
      (arrCmpr,filter)=ShMemGetCmprResult(shMem)
      if filter==1: arrCmpr=arr
      if idx>=zyx[0]:
        zyx=(idx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arrCmpr,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()

      arr=imgLst[idx%numLst]
      ShMemCpyArrIn(shMem,arr)
      p = mproc.Process(target=ChunkBuilderSharedMem, args=(ChunkBuildFunc,shMem,arr.dtype,arr.shape,verbose))
      jobs.append((p,shMem,idx))
      p.start()
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #epilog
    while jobs:
      (p,shMem,procIdx)=jobs.pop(0)
      p.join()
      (arrCmpr,filter)=ShMemGetCmprResult(shMem)
      if filter==1: arrCmpr=arr
      if idx>=zyx[0]:
        zyx=(idx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arrCmpr,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()
    pass

  @staticmethod
  def ShMem1b(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''shared memory
    one supervising/wringing process
    <cpu_count> compressing process
    process terminates at every chunk compression
    maximal-load: do not guarantee the chunk order'''
    verbose=args.verbose
    jobs = {}
    maxJobs=mproc.cpu_count()
    #prolog
    for idx in range(maxJobs):
      arr=imgLst[idx%numLst]
      shMem=ShMemAlloc(arr.nbytes)
      ShMemCpyArrIn(shMem,arr)
      #ChunkBuilderSharedMem(ChunkBuildFunc,shMem,arr.dtype,arr.shape)
      #(arrCmpr,filter)=ShMemGetCmprResult(shMem)
      p = mproc.Process(target=ChunkBuilderSharedMem, args=(ChunkBuildFunc,shMem,arr.dtype,arr.shape,verbose))
      p.start()
      jobs[p.pid]=(p,shMem,idx)#;print '+%d'%p.pid
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #loop
    for idx in range(maxJobs,num):
      pid=None
      for k,v in jobs.iteritems():#check if a job has terminated. in this case use that pid
        if not v[0].is_alive():
          pid=k
          break
      if pid==None:
        pid,res=os.wait()
      (p,shMem,procIdx)=jobs.pop(pid)#;print '-%d'%p.pid
      p.join()
      (arrCmpr,filter)=ShMemGetCmprResult(shMem)
      if filter==1: arrCmpr=arr
      if idx>=zyx[0]:
        zyx=(idx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arrCmpr,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()

      arr=imgLst[idx%numLst]
      ShMemCpyArrIn(shMem,arr)
      p = mproc.Process(target=ChunkBuilderSharedMem, args=(ChunkBuildFunc,shMem,arr.dtype,arr.shape,verbose))
      p.start()
      jobs[p.pid]=(p,shMem,idx)#;print '+%d'%p.pid
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #epilog
    while jobs:
      try:
        pid,res=os.wait()
      except OSError as e:
        pid=jobs.iteritems().next()[0]
      (p,shMem,procIdx)=jobs.pop(pid)#;print '-%d'%p.pid
      p.join()
      (arrCmpr,filter)=ShMemGetCmprResult(shMem)
      if filter==1: arrCmpr=arr
      if idx>=zyx[0]:
        zyx=(idx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arrCmpr,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()
    pass

  @staticmethod
  def ShMem2(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''shared memory
    one supervising/wringing process
    <cpu_count> compressing process
    process terminates at the very end
    round-robin'''
    verbose=args.verbose
    jobs = []
    maxJobs=mproc.cpu_count()
    #prolog
    for idx in range(maxJobs):
      arr=imgLst[idx%numLst]
      shMem=ShMemAlloc(arr.nbytes)
      ShMemCpyArrIn(shMem,arr)
      pipeParent,pipeChild = mproc.Pipe()
      p = mproc.Process(target=ChunkBuilderSharedMem2, args=(ChunkBuildFunc,shMem,arr.dtype,arr.shape,pipeChild))
      jobs.append((p,shMem,pipeParent))
      p.start()
      arr=imgLst[idx%numLst]
      ShMemCpyArrIn(shMem,arr)
      pipeParent.send(idx);#start
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #loop
    for idx in range(maxJobs,num):
      (p,shMem,pipeParent)=jobs[idx%maxJobs]
      procIdx=pipeParent.recv()
      (arrCmpr,filter)=ShMemGetCmprResult(shMem)
      if filter==1: arrCmpr=arr
      if procIdx>=zyx[0]:
        zyx=(procIdx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arrCmpr,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()
      arr=imgLst[idx%numLst]
      ShMemCpyArrIn(shMem,arr)
      pipeParent.send(idx);#start
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
    #epilog
    while jobs:
      (p,shMem,pipeParent)=jobs.pop(0)
      procIdx=pipeParent.recv()
      (arrCmpr,filter)=ShMemGetCmprResult(shMem)
      if filter==1: arrCmpr=arr
      if procIdx>=zyx[0]:
        zyx=(procIdx+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((procIdx, 0, 0),arrCmpr,filters=filter)
      if verbose&2: sys.stdout.write('w');sys.stdout.flush()
      pipeParent.send(None);#stop
      p.join()
    pass

  @staticmethod
  def Pool1(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''build a pool of <cpu_count> jobs
       after starting <cpu_count> jobs, wait for their result'''
    verbose=args.verbose
    jobs = []
    maxJobs=mproc.cpu_count()
    pool=mproc.Pool(processes=maxJobs)
    for idx in range(num):
      arr=imgLst[idx%numLst]
      p = pool.apply_async(ChunkBuilderPool, (ChunkBuildFunc,arr,verbose))
      jobs.append(p)
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
      if idx+1==num or (idx+1)%maxJobs==0: #len(jobs)>=maxJobs:
        if idx>=zyx[0]:
          zyx=(idx+10,)+shape
          dsid.set_extent(zyx)
          fspace.set_extent_simple(zyx, ulim)
        ii=maxJobs-1
        for p in jobs:
          (arr3,filter)=p.get()
          dsid.write_chunk((idx-ii, 0, 0),arr3,filters=filter)
          if verbose&2: sys.stdout.write('w');sys.stdout.flush()
          ii-=1
        jobs = []

  @staticmethod
  def Pool1b(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    '''build a pool of <cpu_count> jobs
       after starting <cpu_count> jobs, wait for oldest job and start a new one'''
    verbose=args.verbose
    jobs = []
    maxJobs=min(mproc.cpu_count(),num)
    pool=mproc.Pool(processes=maxJobs)

    for idx in range(num+maxJobs):
      if idx>=maxJobs: #>>> receive the job <<<
        p,pIdx=jobs.pop()
        if pIdx>=zyx[0]:
          zyx=(pIdx+10,)+shape
          dsid.set_extent(zyx)
          fspace.set_extent_simple(zyx, ulim)
        (arr3,filter)=p.get()
        dsid.write_chunk((pIdx, 0, 0),arr3,filters=filter)
        if verbose&2: sys.stdout.write('w');sys.stdout.flush()
      if idx<num: #>>> send the job <<<
        arr=imgLst[idx%numLst]
        p = pool.apply_async(ChunkBuilderPool, (ChunkBuildFunc,arr,verbose))
        jobs.append((p,idx))
        if verbose&2: sys.stdout.write('.');sys.stdout.flush()

  @staticmethod
  def MPI(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args):
    maxJobs=min(mproc.cpu_count(),num)
    verbose=args.verbose
    commAll = MPI.COMM_WORLD
    print commAll.Get_rank(),commAll.Get_size(),'Spawn...'+__file__
    comm = MPI.COMM_SELF.Spawn(sys.executable,
                               args=[__file__,'CompressMPI',args.cmpr],
                               maxprocs=maxJobs)

    print 'running master MPI...'
    #prolog
    for idx in range(maxJobs):
      arr=imgLst[idx%numLst]
      id=idx
      #print 'MasterMPI send rank%d imgIdx%d'%(id,idx)
      comm.send(arr, dest=id, tag=1)
#       sys.stdout.write('.');sys.stdout.flush()
    #loop
    for idx in range(maxJobs,num):
      id=((idx-maxJobs)%maxJobs)
      idxRcv=idx-maxJobs
      #print 'MasterMPI rcv rank%d imgIdx%d...'%(id,idxCmpr)
      (arrCmpr,filter) = comm.recv(source=id, tag=1)
      #print 'MasterMPI rcv rank%d imgIdx%d done'%(id,idxRcv)
      if idxRcv>=zyx[0]:
        zyx=(idxRcv+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((idxRcv, 0, 0),arrCmpr,filters=filter)
      sys.stdout.write('w');sys.stdout.flush()
      arr=imgLst[idx%numLst]
      comm.send(arr, dest=(idx%maxJobs), tag=1)
      sys.stdout.write('.');sys.stdout.flush()
    #epilog
    for idx in range(num,num+maxJobs):
      id=((idx-maxJobs)%maxJobs)
      idxRcv=idx
      (arrOut,filter) = comm.recv(source=id, tag=1)
      #print 'MasterMPI rcv rank%d imgIdx%d done'%(id,idxRcv)
      if idxRcv>=zyx[0]:
        zyx=(idxRcv+10,)+shape
        dsid.set_extent(zyx)
        fspace.set_extent_simple(zyx, ulim)
      dsid.write_chunk((idxRcv, 0, 0),arrCmpr,filters=filter)
      sys.stdout.write('w');sys.stdout.flush()
      comm.send(None, dest=id, tag=1)
    pass

def CompressMPI(cmpr):
  comm = MPI.Comm.Get_parent()
  #commAll = MPI.COMM_WORLD #this is a communicator between the processes
  size = comm.Get_size()
  rank = comm.Get_rank()
  ChunkBuildFunc=ut.GetChunkBuildFunc(cmpr)
  #print rank,size,'Child processing...',sys.argv
  #if rank==1:
  while True:
    data = comm.recv(source=0, tag=1)
    #print rank,size,'Child received!'
    if data==None:
      break;
    (arrOut,filter)=ChunkBuildFunc(data)
    comm.send((arrOut,filter), dest=0, tag=1)

  #time.sleep(5)
  #print 'CompressMPI rank%d finished.'%rank

class PerfMeasure:
  def __init__(self,args):
    self.args=args
    ut.StopWatch.Start()
    numImg=self.args.num
    imgLst=self.InitImgList(numImg)
    if not numImg:
      numImg=len(imgLst)
    self.imgLst=imgLst
    ut.StopWatch.Log('read %d source images to memory'%len(imgLst))

  def InitImgList(self,num):
    verbose=self.args.verbose
    clpIn=ClampIn(self.args,verbose=verbose)
    noise=self.args.noise
    imgLst=[]
    if not num:
      num=sys.maxint
    while len(imgLst)<num:
      arr=clpIn.GetData()
      if arr==None:
        break
      #arr/=4
      #arr[arr==65535]=0
      #print 'set values 0xffff to 0'
      if noise:
        if not imgLst: #list empty
          numByte=1*np.prod(np.array(arr.shape))
          arrRnd=np.random.bytes(numByte)
          arrRnd=np.frombuffer(arrRnd,dtype=np.uint8)
          arrRnd.resize(arr.shape)
          noise=256-noise
          print 'adding %.2f%% values 1 noise'%((arrRnd/noise).sum()*100./arrRnd.ravel().shape[0])
        arr+=arrRnd/noise
      if verbose&2: sys.stdout.write('.');sys.stdout.flush()
      imgLst.append(arr)
    if verbose&2: sys.stdout.write('\n');sys.stdout.flush()
    return imgLst

  def PrepareNBitImageLst(self,cmpr):
    #If we are using an NBit Filter, the image list must contain already
    #N-Bit filtered data.
    #to simulate this, the image source images are just using the front bytes
    #and forgets all remaining data.
    if cmpr[0]!='n':
      return
    m=re.match('(\d*)',cmpr[1:])
    prec=int(m.group(1))
    imgLst=self.imgLst
    arr=imgLst[0]
    numElem=np.prod(arr.shape)
    shape=((prec*numElem+7)/8,)
    for idx in range(len(imgLst)):
      arr=imgLst[idx]
      arr=np.ndarray(buffer=arr.data,shape=shape,dtype=np.uint8)
      imgLst[idx]=arr

  def OpenHDF(self):
    args=self.args
    shape=self.imgLst[0].shape #y x
    dtype=self.imgLst[0].dtype
    fnHDF5=args.dst
    cmpr=args.cmpr

    fid=ut.OpenFile(fnHDF5,'w')

    chunk=(1,)+shape #z y x
    maxshape=(h5py.h5s.UNLIMITED,)+shape
    dsid=ut.CreateDatasetChunked(fid,'/entry/data/NewDataset',dtype,(10,)+shape,maxshape,chunk,cmpr)
    self.fid=fid
    self.dsid=dsid

  def WriteHDF(self,multiProcFunc=None):
    args=self.args
    cmpr=args.cmpr
    imgLst=self.imgLst
    numLst=len(imgLst)
    num=args.num
    if num==None:
      num=numLst
    shape=imgLst[0].shape #y x
    dsid=self.dsid

    zyx  =(10,)+shape
    chunk=(1, )+shape
    ulim=(h5py.h5s.UNLIMITED,)*len(zyx)
    fspace = h5py.h5s.create_simple(zyx, ulim)
    mspace = h5py.h5s.create_simple(chunk, ulim)

    if cmpr:
      self.PrepareNBitImageLst(cmpr)
      ChunkBuildFunc=ut.GetChunkBuildFunc(cmpr)
      if multiProcFunc:
        multiProcFunc(imgLst,numLst,num,zyx,dsid,shape,fspace,ulim,ChunkBuildFunc,args)
      else: #single process compressed
        for idx in range(num):
          arr=imgLst[idx%numLst]
          if idx>=zyx[0]:
            zyx=(idx+10,)+shape
            dsid.set_extent(zyx)
            fspace.set_extent_simple(zyx, ulim)
          (arr3,filter)=ChunkBuildFunc(arr)
          dsid.write_chunk((idx, 0, 0),arr3,filters=filter)
          if args.verbose&2:sys.stdout.write('.');sys.stdout.flush()
    else: #single process uncompressed
      for idx in range(num):
        arr=imgLst[idx%numLst]
        if idx>=zyx[0]:
          zyx=(idx+10,)+shape
          dsid.set_extent(zyx)
          fspace.set_extent_simple(zyx, ulim)
        dsid.write_chunk((idx, 0, 0),arr)
        if args.verbose&2:sys.stdout.write('.');sys.stdout.flush()
    if args.verbose&2:sys.stdout.write('\n')
    zyx=(num,)+shape
    dsid.set_extent(zyx)
    return

  def RunWrHDF(self,multiProcFunc=None):
    self.OpenHDF()
    ut.StopWatch.Log('create output hdf5 file')
    self.WriteHDF(multiProcFunc)
    t=ut.StopWatch.Log('write dataset to hdf5 file')
    fid=self.fid
    del self.dsid; del self.fid #release reference to finally close the file!
    ut.CloseFile(fid)
    st=os.stat(self.args.dst)
    szDst=float(st.st_size)
    imgLst=self.imgLst
    numImg=self.args.num
    if not numImg:
      numImg=len(imgLst)
    szRaw=float(numImg*imgLst[0].nbytes)
    if multiProcFunc:
      mpName=multiProcFunc.__name__
    else:
      mpName=''
    ut.StopWatch.Log('hdf5 file closed. %s: Raw: %.2fMB Compr:%.2fMB Ratio: %.2f%%, %.2fx Speed:%.2fMB/s'%(mpName,szRaw/2**20, szDst/2**20,(1.-szDst/szRaw)*100.,szRaw/szDst,szRaw/2**20/t))

  def RunCmpSpeed(self):
    '''measure speed of different compression algo. takes only the first image'''
    numImg=self.args.num
    cmprLst=self.args.cmpr
    imgLst=self.imgLst
    numLst=len(imgLst)
    if not numImg:
      numImg=numLst
    verbose=self.args.verbose
    shuffle=self.args.shuffle

    arrOut=np.ndarray(shape=(imgLst[0].nbytes,),dtype=np.ubyte)

    szRaw=float(imgLst[0].nbytes*numImg)

    if shuffle:
      imgLstS=[] #shuffeled image list
      for idx in range(numLst):
        arr=imgLst[idx]
        arrS=np.ndarray(shape=arr.shape,dtype=arr.dtype)
        imgLstS.append(arrS)
      ut.StopWatch.Start()
      szDst=0.
      for idx in range(numImg):
        arr=imgLst[idx%numLst]; arrS=imgLstS[idx%numLst]
        res=LibDetXR.ByteShuffle(arr.ctypes.data, arrS.ctypes.data, arr.nbytes, arr.itemsize)
        if verbose&2:sys.stdout.write('.');sys.stdout.flush()
      if verbose&2:sys.stdout.write('\r');sys.stdout.flush()
      t=ut.StopWatch.GetTime()
      ut.StopWatch.Log(('ByteShuffle'+' '*12+'Raw:%7.2fMB'+' '*39+'Speed:%8.2fMB/s')%(szRaw/2**20,szRaw/2**20/t))

    if self.args.diff:
      _log.info('differentiate bytes with identical weight')
      #differentiate image (check if it results in better compression)
      for idx in range(numLst):
        arr=imgLst[idx]
        arrByte=np.frombuffer(arr.data,dtype=np.uint8)
        for i in range(arr.itemsize):
          arrByte[i::arr.itemsize]=np.ediff1d(arrByte[i::arr.itemsize],to_begin=arrByte[i])
        if verbose&2:sys.stdout.write('.');sys.stdout.flush()
      if verbose&2:sys.stdout.write('\n');sys.stdout.flush()

      if shuffle:
        for idx in range(numLst):
          arr=imgLstS[idx]
          arrByte=np.frombuffer(arr.data,dtype=np.uint8)
          for i in range(arr.itemsize):
            pitch=arr.nbytes/arr.itemsize
            s=i*pitch;e=s+pitch
            arrByte[s:e]=np.ediff1d(arrByte[s:e],to_begin=arrByte[s])
          if verbose&2:sys.stdout.write('.');sys.stdout.flush()
        if verbose&2:sys.stdout.write('\n');sys.stdout.flush()

    for cmpr in cmprLst:
      if cmpr.startswith('zlib'):
        for lvl in map(int,cmpr[4:]):
          szDst=0.
          for idx in range(numImg):
            arr=imgLst[idx%numLst]
            res=LibDetXR.CompressZLIB(arr.ctypes.data, arr.nbytes, arrOut.ctypes.data, arrOut.nbytes,lvl);szDst+=res
            if res==0: _log.error(cmpr[:4]+'%d compression failed'%lvl);break
            if verbose&2:sys.stdout.write('.');sys.stdout.flush()
          if verbose&2:sys.stdout.write('\r');sys.stdout.flush()
          if res:
            t=ut.StopWatch.GetTime()
            ut.StopWatch.Log('CompressZLIB%d initial  Raw:%7.2fMB Compr:%7.2fMB Ratio: %.2f%%,%6.2fx Speed:%8.2fMB/s'%(lvl,szRaw/2**20, szDst/2**20,(1.-szDst/szRaw)*100.,szRaw/szDst,szRaw/2**20/t))
          szDst=0.
          if shuffle:
            for idx in range(numImg):
              arr=imgLstS[idx%numLst]
              res=LibDetXR.CompressZLIB(arr.ctypes.data, arr.nbytes, arrOut.ctypes.data+16, arrOut.nbytes,lvl);szDst+=res
              if res==0: _log.error(cmpr[:4]+'%d shuffle compression failed'%lvl);break
              if verbose&2:sys.stdout.write('.');sys.stdout.flush()
            if verbose&2:sys.stdout.write('\r');sys.stdout.flush()
            if res:
              t=ut.StopWatch.GetTime()
              ut.StopWatch.Log('CompressZLIB%d shuffled Raw:%7.2fMB Compr:%7.2fMB Ratio: %.2f%%,%6.2fx Speed:%8.2fMB/s'%(lvl,szRaw/2**20, szDst/2**20,(1.-szDst/szRaw)*100.,szRaw/szDst,szRaw/2**20/t))
      else:
        def CompressZRLE(inBuf, inNumBytes, outBuf, outNumBytes):
          return LibDetXR.CompressZRLE(inBuf, inNumBytes, outBuf, outNumBytes,0)

        cmprFuncDict={'lz4':LibDetXR.CompressLZ4,
                      'lzf':LibDetXR.CompressLZF,
                      'zrle':CompressZRLE}
        try:
          cmprFunc=cmprFuncDict[cmpr]
        except KeyError:
          try:
            #cmpModule = __import__("myapp.commands.%s" % command, fromlist=["myapp.commands"])
            mod,func=cmpr.rsplit('.',1)
            cmpModule = __import__(mod,fromlist=[func])
            cmprFunc=getattr(cmpModule,func)
          except (AttributeError,ImportError) as e:
            _log.error('compression not found: '+cmpr+' : '+str(e))
        cmprName=cmprFunc.__name__
        szDst=0.
        for idx in range(numImg):
          arr=imgLst[idx%numLst]
          res=cmprFunc(arr.ctypes.data, arr.nbytes, arrOut.ctypes.data, arrOut.nbytes);szDst+=res
          if res==0: _log.error(cmpr+' compression failed');break
          if verbose&2:sys.stdout.write('.');sys.stdout.flush()
        if verbose&2:sys.stdout.write('\r');sys.stdout.flush()
        if res:
          t=ut.StopWatch.GetTime()
          ut.StopWatch.Log(cmprName+'  initial Raw:%7.2fMB Compr:%7.2fMB Ratio: %.2f%%,%6.2fx Speed:%8.2fMB/s'%(szRaw/2**20, szDst/2**20,(1.-szDst/szRaw)*100.,szRaw/szDst,szRaw/2**20/t))
        if shuffle:
          szDst=0.
          for idx in range(numImg):
            arr=imgLstS[idx%numLst]
            res=cmprFunc(arr.ctypes.data, arr.nbytes, arrOut.ctypes.data, arrOut.nbytes);szDst+=res
            if res==0: _log.error(cmprName+' shuffle compression failed');break
            if verbose&2:sys.stdout.write('.');sys.stdout.flush()
          if verbose&2:sys.stdout.write('\r');sys.stdout.flush()
          if res:
            t=ut.StopWatch.GetTime()
            ut.StopWatch.Log(cmprName+' shuffled Raw:%7.2fMB Compr:%7.2fMB Ratio: %.2f%%,%6.2fx Speed:%8.2fMB/s'%(szRaw/2**20, szDst/2**20,(1.-szDst/szRaw)*100.,szRaw/szDst,szRaw/2**20/t))

  def RunImgStat(self):
    arr=self.imgLst[0]
    #arr/=4
    print 'itemsize %d'%arr.itemsize
    arrByte=np.frombuffer(arr.data,dtype=np.uint8)

    f=plt.figure(1)
    f.canvas.set_window_title('Byte statistics (top=lowest byte, bottom=highest byte)')
    for i in range(arr.itemsize):
      plt.subplot(arr.itemsize,1,i+1)
      #plt.title('Byte statistics (top=highest byte, bottom=lowest byte)')
      hist,bins=np.histogram(arrByte[i::arr.itemsize], range(257))
      plt.plot(hist)
      if self.args.log:
        plt.yscale('log')

    plt.show(block=False)

    plt.figure(2)
    if self.args.log:
      plt.imshow(arr,norm=mpl.colors.LogNorm(1,arr.max()))
    else:
      plt.imshow(arr)
    plt.title('first image')
    plt.show(block=False)

    if 0: #testing diff  images bytes and reverse it

      for i in range(arr.itemsize):
        arrByte[i::arr.itemsize]=np.ediff1d(arrByte[i::arr.itemsize],to_begin=arrByte[i])

      plt.figure(3)
      plt.imshow(arr)
      plt.show(block=False)

      for i in range(arr.itemsize):
        arrByte[i::arr.itemsize]=np.cumsum(arrByte[i::arr.itemsize],dtype=np.uint8)

      plt.figure(4)
      plt.imshow(arr)
      plt.show(block=False)

    #print 'statistics of first image: %s max %d %.2f%% not 0 values'%(str(arr.dtype),arr.max(),(arr>0).sum()*100./arr.ravel().shape[0])

    plt.show(block=True)


    return

if __name__ == '__main__':
  def RunWrHDF(args):
    if args.verbose&1:print 'RunWrHDF',args
    if args.multiproc:
      pm=PerfMeasure(args)
      for multiProcFunc in  args.multiproc:
        try:
          multiProcFunc=getattr(MultProc,multiProcFunc)
        except AttributeError as e:
          raise(AttributeError('unknown multiprocess type '+multiProcFunc))
          continue
        else:
          pm.RunWrHDF(multiProcFunc)
    else:
      PerfMeasure(args).RunWrHDF()

  def RunCmpSpeed(args):
    if args.verbose&1:print 'RunCmpSpeed',args
    PerfMeasure(args).RunCmpSpeed()

  def RunImgStat(args):
    if args.verbose&1:print 'RunImgStat',args
    PerfMeasure(args).RunImgStat()

  def RunWithArgs():
    import argparse #since python 2.7
    epilog='Data Source description:'+ClampIn.__doc__
    epilog+='''
the verbose bits are:
  - bit0( 1): basic
  - bit1( 2): progress .=frame
  - bit2( 4): progress detail
  - bit3( 8): multiproc
  - bit4(16):
  - bit5(32):
  - bit6(64):
  - bit8(0x100): multiprocessing+clamps
  - bit9(0x200): multiprocessing+clamps (more verbose)
  default is 1+2=3 (no -v)
  -v is 1+8+0x100+0x200=777
  -v0 is turn all off

'''
    cmd='\n  '+os.path.basename(sys.argv[0])+' '
    #epilog+='help about sub-commands::'+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=__doc__, epilog=epilog)
    #common arguments for all subcommands
    exampleCmd=('--src raw uint16 2560 2160 /scratch/detectorData/PCO/ ',
                '--src rand uint16 2560 2160 3',
                '--src inc uint16 2560 2160 6',
                '--src inc uint16 2560 2160 6 --cmpr s-zlib6',
                '--src inc uint16 2560 2160 6 --cmpr s-lz4',
                '--src inc uint16 2560 2160 6 --cmpr s-lzf')
    parser.epilog+='Example:'+''.join(map(lambda s:cmd+s+' <sub-cmd args>',exampleCmd))+'\n '
    parser.add_argument('--src',   nargs='+', required=True, help='data source (see below for examples)')
    parser.add_argument('--noise', type=int, help='adding noise values 1. 0=0%% 255=100%%')
    parser.add_argument('--num',   type=int, help='number of images')
    parser.add_argument('-v',      dest='verbose', type=int, nargs='?', const=777, default=3, help='verbose bit mask:  not set:%(default)s, -v:%(const)s or e.g. -v4 ')

    subparsers = parser.add_subparsers(title='sub command')

    prsImgStat = subparsers.add_parser('imgStat', help='image statistics',epilog='',formatter_class=argparse.RawDescriptionHelpFormatter)
    prsImgStat.add_argument('--log',  action='store_true', help='logarithmic statistics')
    prsImgStat.set_defaults(func=RunImgStat)

    prsCmpSpeed = subparsers.add_parser('cmpSpeed', help='compression speed',epilog='',formatter_class=argparse.RawDescriptionHelpFormatter)
    cmprSet='lz4 lzf zlib246'
    prsCmpSpeed.add_argument('--cmpr',  nargs='+', default=cmprSet.split(), help='compressions to test set of "d '+cmprSet+'"')
    prsCmpSpeed.add_argument('--shuffle', action='store_true', help='shuffle the image bytes')
    prsCmpSpeed.add_argument('--diff', action='store_true', help='differentiate the image bytes')
    prsCmpSpeed.set_defaults(func=RunCmpSpeed)
    prsCmpSpeed.epilog+='''
The cmpr indicates which ompressions to test:
zlib<numbers> :use zlib compressions with given levels
               zlib246 will test zlib2 zlib4 and zlib6
lz4           :use lz4 compression
lzf           :use lzf compression
<module>.<cmpFunc>    :use a user compression algo:
               it tries to load module user
               this module must contain a function

s             :shuffle data a  lz4 lzf zlib246


Examples for user Compression:
Build file usercmp.py:
#!/usr/bin/env python
import numpy as np
import zlib

def MyFunc(inBuf, inNumBytes, outBuf, outNumBytes):
  return inNumBytes/2

def zLibPy(inBuf, inNumBytes, outBuf, outNumBytes):
  lvl=4
  _inBuf = np.core.multiarray.int_asbuffer(inBuf, inNumBytes)
  _outBuf=zlib.compress(_inBuf,lvl)
  if len(_outBuf)<outNumBytes:
    #memcopy of _outBuf to outBuf if len(_outBuf)<outNumBytes
    return len(_outBuf)
  else:
    return 0 #compression bigger than origin

Then call with:
  cmpSpeed --cmpr zlib4 lz4 usercmp.MyFunc usercmp.zLibPy --shuffle

'''


    prsWrHDF = subparsers.add_parser('wrHDF', help='write HDF5 file',epilog='',formatter_class=argparse.RawDescriptionHelpFormatter)
    prsWrHDF.add_argument('--cmpr',  help='prefix ''s-'' for byte shuffle, compression zlib[0-9],.lz4,lzf,cbf. e.g. s-zlib5: shuffle zlib level 5 compression.')
    prsWrHDF.add_argument('--multiproc', nargs='+', help='multiprocessing method')
    prsWrHDF.add_argument('--dst',    default='/scratch/detectorData/testPerf.hdf5', help='output hdf file name')
    prsWrHDF.set_defaults(func=RunWrHDF)


    lst=list()
    for mp in dir(MultProc):
      if mp[0]!='_':
        lst.append(mp)

    prsWrHDF.epilog+='possible multiproc arguments are: \n  ('+' '.join(lst)+')\n   read function descriptions MultiProc.<method> for details\n\n'

    exampleCmd=('--cmpr s-zlib6',
                '--cmpr s-lz4 --multiproc',
                '--cmpr s-lzf --dst /tmp/out.hdf5')
    prsWrHDF.epilog+='Example:'+''.join(map(lambda s:cmd+'<cmd args> '+s,exampleCmd))+'\n '

    parser.epilog+='\nHelp about sub-commands:'+''.join(map(lambda s:cmd+s+' -h',subparsers.choices.keys()))+'\n '
    args = parser.parse_args()
    args.func(args)
#------------------ Main Code ----------------------------------
  hdlr=logging.StreamHandler()
  fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
  hdlr.setFormatter(fmt)
  del logging.root.handlers[:]
  logging.root.addHandler(hdlr)
  logging.root.setLevel(logging.DEBUG)

  logLvls={
           __name__     : logging.DEBUG,
          }
  for k,v in logLvls.iteritems():
    logging.getLogger(k).setLevel(v)

  if sys.argv[1]=='CompressMPI':
    CompressMPI(sys.argv[2])
  else:
    RunWithArgs()

