#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
Writes a ZeroMQ stream to the console (Debug Purpose)
It connects to a ZMQ address that sends json- and binary data messages.
The supported json-htype are:
 - chunk-1.0 (for raw or compressed images)
 - pilatus-1.0 (for cbf-file transfer from pilatus detector)
the verbose bits are:
   bit0( 1): basic
   bit1( 2): progress +=frame -=timeout
   bit2( 4): progress jsonstr
   bit3( 8): progress chunk size
   bit3(16): zmq timeout
   default is 1+2=3 (no -v)
   -v is 1+4+8+16=29
   -v0 is turn all off
'''
#usefull links:
#http://stackoverflow.com/questions/7538988/zeromq-how-to-prevent-infinite-wait
import sys, os, zmq
import utilities as ut

import json,re

class ZmqConsoleSink:
  ifDict={'PULL':zmq.PULL,'SUB':zmq.SUB}

  def Run(self,args):
    totSz=0
    self.args=args
    try: pipe=args.pipe
    except AttributeError: pipe=None
    try: numFrm=args.numFrm
    except AttributeError: numFrm=None

    print 'started ZmqConsoleSink %s ...'%str(args.src[1])
    sock=ut.ZMQ.Socket(args.src[1],zmq.PUSH,4)

    frmCnt=0
    while True:
      try:
        #time.sleep(.5)
        if pipe and pipe.poll():
          cmd=pipe.recv()
          if args.verbose&16:
            print 'received '+cmd+' from '+str(pipe)
          if cmd=='stop':
            break
          elif cmd=='frmCnt':
            if args.verbose&16:
              print 'send '+str(frmCnt)+' to '+str(pipe)
            pipe.send(str(frmCnt))
        header = sock.recv_json()
        try: cmd=header['cmd']
        except KeyError as e: pass
        else:
          if cmd=='stop': break
        if sock.getsockopt(zmq.RCVMORE):
          content= sock.recv(copy=False)
        frmCnt+=1
        totSz+=len(content)
        if args.verbose&4:
          print "%d (%d) "%(frmCnt,len(content))+str(header)
        elif args.verbose&2:
          sys.stdout.write('+')
        if numFrm and frmCnt>=numFrm:
          break
      except zmq.error.Again as e:
        if args.verbose&8:
          print 'zmq timeout after %d ms'%sock.RCVTIMEO
        elif args.verbose&2:
          sys.stdout.write('-')
        if pipe==None:break
      sys.stdout.flush()
    if pipe: pipe.send(str(frmCnt-1)) #send the frame numbers to the pipe
    print 'Frame Count:',frmCnt,'Total Image Bytes:',totSz,totSz/1024./1024/1024,'GB'
    return

def RunWithArgs(args):
  #------ prepare Arguments ------
  if type(args)==dict:
    args=ut.dotdict(args)
    default={
            'numFrm':sys.maxint,
            'timeout':100, #100ms timeout the reveive and check for a stop command
            'ifType':'PULL',
            'queueSz':4,
            'verbose':3,
            'pipe':None,
            } #setup default values
    for k, v in default.items():
      if not args.has_key(k):
        args[k]=v
  else:
    args=ut.dotdict(args.__dict__)
  args.src=['zmq',{'addr':args.addr,'qs':args.queueSz,'if':args.ifType}]
  if args.timeout!=None:
    args.src[1]['timeout']=args.timeout
  for k in ('addr','queueSz','ifType','timeout'):
    del args[k]
  wr=ZmqConsoleSink()
  wr.Run(args)
  pass

if __name__ == '__main__':
  import argparse #since python 2.7
  def ParseArgs():
    #default values
    ifType='PULL'
    queueSz=4

    exampleCmd=('--addr tcp://localhost:8080 --numFrames 500 --timeout 1000 -v',)

    cmd='\n  '+os.path.basename(sys.argv[0])+' '
    epilog='Example:'+''.join(map(lambda s:cmd+s,exampleCmd))+'\n '
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=__doc__, epilog=epilog)
    parser.add_argument('--addr',                    required=True, help='host:port of the zmq connection')
    parser.add_argument('--numFrames',  dest='numFrm',  type=int, default=sys.maxint, help='maximal number of frames to receive')
    parser.add_argument('--timeout',                   type=int,      help='receive timeout in ms')
    parser.add_argument('--if',  dest='ifType', choices=ZmqConsoleSink.ifDict,default=ifType, help='interface type (default:'+ifType+')')
    parser.add_argument('--qs',  dest='queueSz', type=int,default=queueSz, help='queue size (default:'+str(queueSz)+')')
    parser.add_argument('-v',          dest='verbose', type=int, nargs='?', const=29, default=3, help='verbose (optional level bitmask)')
    args = parser.parse_args()
    RunWithArgs(args)

#------------------ Main Code ----------------------------------
  ParseArgs()
