#!/usr/bin/env python
#*-----------------------------------------------------------------------*
#|                                                                       |
#|  Copyright (c) 2013 by Paul Scherrer Institute (http://www.psi.ch)    |
#|                                                                       |
#|              Author Thierry Zamofing (thierry.zamofing@psi.ch)        |
#*-----------------------------------------------------------------------*
'''
utilities to create HDF5 objects
'''
import logging, h5py, re, zlib, zmq, json
import numpy as np
from libDetXR import *
import time

_log=logging.getLogger(__name__)
if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG,format='%(levelname)s:%(module)s:%(lineno)d:%(funcName)s:%(message)s ')
  pass

class StopWatch():
  @classmethod
  def Start(cls):
    cls.ts=time.time()
  @classmethod
  def GetTime(cls):
    ts=time.time()
    return (ts-cls.ts)
  @classmethod
  def Log(cls,s=None,restart=True):
    ts=time.time()
    t=(ts-cls.ts)
    print '%.6f'%t,s
    if restart:
      cls.ts=ts
    return t

def OpenFile(fnHDF5,mode='r',bigCache=False):
  '''Opens a HDF5 file. modes are:

   - 'r': read only
   - 'w': write:  If a file with the same name already exists, its contents are discarded and the file is treated as a new empty file.
   - 'a': append: The file is created if it does not exist.'''

  fapl = h5py.h5p.create(h5py.h5p.FILE_ACCESS)
  fcpl = h5py.h5p.create(h5py.h5p.FILE_CREATE)
  if bigCache:
    _log.debug('cache settings (old):'+str(fapl.get_cache()))
    fapl.set_cache(0,2**15-1,2**23,0.75) #set big chunk cache
    _log.debug('cache settings (new):'+str(fapl.get_cache()))

  if mode=='r':
    fid = h5py.h5f.open(fnHDF5, flags=h5py.h5f.ACC_RDONLY)
  elif mode=='w':
    fid = h5py.h5f.create(fnHDF5, flags=h5py.h5f.ACC_TRUNC, fcpl=fcpl, fapl=fapl)
  elif mode=='a':
    try:
      fid = h5py.h5f.open(fnHDF5, flags=h5py.h5f.ACC_RDWR)
    except IOError as e:
      fid = h5py.h5f.create(fnHDF5, flags=h5py.h5f.ACC_TRUNC, fcpl=fcpl, fapl=fapl)
  else:
    return None
  if _log.isEnabledFor(logging.INFO):
    action={'r':'read only','w':'write/trunc','a':'append'}
    _log.info(action[mode]+' file: '+fnHDF5)
  return fid

def CloseFile(fid):
  'Close a HDF5 file.'
  if _log.isEnabledFor(logging.INFO):
    #st=os.stat(fn);sz=float(st.st_size) #can give a non final size sometimes
    #fid.get_filesize() @can give a wrong size (not size when file is closed)
    #therefore the file handle is kept with reopen, the origin file handle is closed and then the size is determined
    fid2=fid.reopen()
    fid.close()
    sz=fid2.get_filesize()
    fn=fid2.name
    fid2.close()
    szf=float(sz)
    for ext in 'BKMG':
      if szf<1024.:
        break
      szf/=1024.
    _log.info('file %.2f%c (%dB): '%(szf,ext,sz)+fn)
  else:
    fid.close()

def CreateGrpPath(gidRoot,grpPath,nxClass=None):
  'create a group path in an opened hdf5 file'
  grp=h5py.Group(gidRoot)
  if type(grpPath)==str:
    grpPath=grpPath.strip('/').split('/')
  for p in grpPath:
    #gid = h5py.h5g.create(gid, p)
    #lcpl = h5py.h5p.create(h5py.h5p.LINK_CREATE)
    #lcpl.set_create_intermediate_group(True)
    #gid = h5py.h5g.create(gid, p, lcpl=lcpl)
    try:
      grp=grp.create_group(p)
    except ValueError:
      grp=grp[p]
    else:
      if nxClass:
        grp.attrs['NX_class'] = nxClass
  return grp.id

class DsObj:
  'helper class to manage hdf5 dataset objects'
  def __init__(self,ds,shape,chunk,ofsIdx=0,varIdx=0):
    ulim=(h5py.h5s.UNLIMITED,)*len(shape)
    self.fspace = h5py.h5s.create_simple(shape, ulim)
    self.mspace = h5py.h5s.create_simple(chunk, ulim)
    self.ds=ds
    self.ulim=ulim
    self.shape=np.array(shape,dtype=np.uint32)
    self.maxIdx=-1
    self.varIdx=varIdx
    self.ofsIdx=ofsIdx

  def MaxReshape(self,idx):
    '''resize the varIdx dimension to 64 additional entries.
    when calling FitShape, the shape will be reduced to idx '''
    shape=self.shape
    varIdx=self.varIdx
    idx+=self.ofsIdx
    if idx>self.maxIdx:
      self.maxIdx=idx
    if idx>=shape[varIdx]:
      shape[varIdx]=idx+64
      st=tuple(shape)
      self.ds.set_extent(st)
      self.fspace.set_extent_simple(st, self.ulim)

  def FitShape(self):
    'resize the varIdx dimension to the final size'
    maxIdx=self.maxIdx+1
    shape=self.shape
    varIdx=self.varIdx
    if maxIdx!=shape[varIdx]:
      ds=self.ds
      shape[varIdx]=maxIdx
      ds.set_extent(tuple(shape))

def CreateDatasetChunked(gidRoot,objPath,dtype,shape,maxshape,chunk,cmpr):
  'create a chunked dataset in an hdf5 file'
  grp=h5py.Group(gidRoot)
  if type(objPath)==str:
    objPath=objPath.strip('/').split('/')
  gid=CreateGrpPath(gidRoot,objPath[:-1])
  dcpl = h5py.h5p.create(h5py.h5p.DATASET_CREATE) #data-set creation property list
  lcpl = h5py.h5p.create(h5py.h5p.LINK_CREATE)   #link creation property list
  dcpl.set_layout(h5py.h5d.CHUNKED)
  dcpl.set_chunk(chunk)
  dcpl.set_fill_value(np.zeros((1,1),dtype))
  #dcpl.set_fill_value(np.array([666],dtype))

  htype = h5py.h5t.py_create(dtype)
  if type(cmpr)==str:
    if cmpr.startswith('n'):
      m=re.match('(\d*)',cmpr[1:])
      if m:
        try:
          prec=int(m.group(1))
        except ValueError:
          raise TypeError('invalid n-bit compression: '+cmpr)
        cmpr=cmpr[1+m.end():]
        if cmpr!='':
          assert(cmpr[0]=='-')
          cmpr=cmpr[1:]
        e=dcpl.set_filter(h5py.h5z.FILTER_NBIT, h5py.h5z.FLAG_OPTIONAL, None)
        htype.set_precision(prec)
      pass
    elif cmpr.startswith('s'):
      if len(cmpr)>1:
        assert(cmpr[1]=='-')
        cmpr=cmpr[2:]
      else:
        cmpr=''
      if dtype.itemsize>1:
        dcpl.set_shuffle()
      else:
        _log.warning('single byte shuffling does nothing: change compression from s-%s to %s'%(cmpr,cmpr))
    m=re.match('zlib(\d*)',cmpr)
    errFmt='cannot set filter %s. check the environment variable HDF5_PLUGIN_PATH.'
    if m:
      try:
        lvl=int(m.group(1))
      except ValueError:
        lvl=4 # set default level=4
      dcpl.set_deflate(lvl) # set gzip level
    elif cmpr=='lz4':
      try:
        e=dcpl.set_filter(LibDetXR.LZ4_FILTER, h5py.h5z.FLAG_OPTIONAL, None)
      except RuntimeError as e:
        e=RuntimeError(errFmt%cmpr)
        raise(e)
    elif cmpr=='lzf':
      try:
        e=dcpl.set_filter(LibDetXR.LZF_FILTER, h5py.h5z.FLAG_OPTIONAL, None)
      except RuntimeError as e:
        e=RuntimeError(errFmt%cmpr)
        raise(e)
    elif cmpr=='cbf':
      e=RuntimeError('cbf compression not yet implemented')
      raise(e)
    else:
      if cmpr not in ('','dummy'):
        raise TypeError('unknown compression: '+cmpr)
  space = h5py.h5s.create_simple(shape,maxshape)
  dsid = h5py.h5d.create(gid, objPath[-1], htype, space, dcpl=dcpl, lcpl=lcpl)#, dapl=dapl)
  return dsid

class ChunkBuildFuncDummy():
  def __call__(self,arr,arr3=None):
    arr3=arr;filter=0
    return (arr3,filter)

class ChunkBuildFuncSimulateN4():
  #returns half of the data. This simulates an n4 compression
  #n-bit filter is not implemented!
  def __call__(self,arr,arr3=None):
    filter=0
    arr3=arr.reshape(arr.nbytes)
    return (arr3[:arr.nbytes/2],filter)

class ChunkBuildFuncLZF():
  def __init__(self,shuffle=True):
    self.shuffle=shuffle

  def __call__(self,arr,arr3=None):
    if self.shuffle and arr.itemsize>1:
      arr2=np.ndarray(shape=arr.shape,dtype=arr.dtype)
      res=LibDetXR.ByteShuffle(arr.ctypes.data, arr2.ctypes.data, arr.nbytes, arr.itemsize)
    else:
      arr2=arr

    if arr3==None:
      arr3=np.ndarray(shape=(arr2.nbytes,),dtype=np.ubyte)
    res=LibDetXR.CompressLZF(arr2.ctypes.data, arr2.nbytes, arr3.ctypes.data, arr3.nbytes)
    if res:
      #arr3=arr3.data[:res];filter=0
      arr3=arr3[:res];filter=0
      #arr3=np.frombuffer(arr3, np.ubyte)
    else:
      arr3=arr2;filter=3 if self.shuffle else 1
    return (arr3,filter)

class ChunkBuildFuncCBF():
  def __init__(self,shuffle=True):
    self.shuffle=shuffle

  def __call__(self,arr,arr3=None):
    if self.shuffle and arr.itemsize>1:
      arr2=np.ndarray(shape=arr.shape,dtype=arr.dtype)
      res=LibDetXR.ByteShuffle(arr.ctypes.data, arr2.ctypes.data, arr.nbytes, arr.itemsize)
    else:
      arr2=arr
    if arr3==None:
      arr3=np.ndarray(shape=(arr2.nbytes,),dtype=np.ubyte)
    res=LibDetXR.CompressCBF(arr2.ctypes.data, arr2.nbytes, arr3.ctypes.data, arr3.nbytes)
    if res:
      #arr3=arr3.data[:res];filter=0
      arr3=arr3[:res];filter=0
      #arr3=np.frombuffer(arr3, np.ubyte)
    else:
      arr3=arr2;filter=3 if self.shuffle else 1
    return (arr3,filter)

class ChunkBuildFuncLZ4():
  def __init__(self,shuffle=True):
    self.shuffle=shuffle

  def __call__(self,arr,arr3=None):
    if self.shuffle and arr.itemsize>1:
      arr2=np.ndarray(shape=arr.shape,dtype=arr.dtype)
      res=LibDetXR.ByteShuffle(arr.ctypes.data, arr2.ctypes.data, arr.nbytes, arr.itemsize)
    else:
      arr2=arr
    if arr3==None:
      arr3=np.ndarray(shape=(arr2.nbytes,),dtype=np.ubyte)
    #arr3[:30]=42
    res=LibDetXR.CompressLZ4(arr2.ctypes.data, arr2.nbytes, arr3.ctypes.data+16, arr3.nbytes-16)
    #arr4=np.ndarray(shape=shape,dtype=dtype)
    #res=LibDetXR.DecompressLZ4(arr3.ctypes.data+16, res, arr4.ctypes.data, arr4.nbytes)
    if res:
      #write the HDF5 Header as defined in document HDF5_LZ4.pdf (HDF5 - LZ4.xlsx)
      arr3.data[  :8 ]=np.uint64(arr2.nbytes).byteswap().data #Total decompressed size
      arr3.data[ 8:12]=np.uint32(arr2.nbytes).byteswap().data #Decompressed bytes per block
      arr3.data[12:16]=np.uint32(res).byteswap().data         #Compressed length of block 0
      arr3=arr3[:res+16];filter=0# reference, cheap copy
      #arr3=np.frombuffer(arr3, np.ubyte)
    else:
      arr3=arr2;filter=3 if self.shuffle else 1
    return (arr3,filter)

class ChunkBuildFuncZLIB():
  def __init__(self,shuffle=True,lvl=4):
    self.lvl=lvl
    self.shuffle=shuffle

  def __call__(self,arr,arr3=None):
    if self.shuffle and arr.itemsize>1:
      arr2=np.ndarray(shape=arr.shape,dtype=arr.dtype)
      res=LibDetXR.ByteShuffle(arr.ctypes.data, arr2.ctypes.data, arr.nbytes, arr.itemsize)
    else:
      arr2=arr
    if arr3==None:
      arr3=np.ndarray(shape=(arr2.nbytes,),dtype=np.ubyte)
    res=LibDetXR.CompressZLIB(arr2.ctypes.data, arr2.nbytes, arr3.ctypes.data, arr3.nbytes, self.lvl)
    if res:
      arr3=arr3[:res];filter=0
    else:
      arr3=arr2;filter=3 if self.shuffle else 1
    return (arr3,filter)

  def Decompr(self,cmpr,arrOut):
    res=LibDetXR.DecompressZLIB(cmpr.ctypes.data, cmpr.nbytes, arrOut.ctypes.data, arrOut.nbytes)
    if self.shuffle and arrOut.itemsize>1:
      arrUnSh=np.ndarray(shape=arrOut.shape,dtype=arrOut.dtype)
      res=LibDetXR.ByteShuffle(arrOut.ctypes.data, arrUnSh.ctypes.data, arrOut.nbytes, arrOut.nbytes/arrOut.itemsize)
      arrOut=arrUnSh
    return arrOut

class ChunkBuildFuncZLIBpy():
  def __init__(self,shuffle=True,lvl=4):
    self.lvl=lvl
    self.shuffle=shuffle

  def __call__(self,arr):
    if self.shuffle and arr.itemsize>1:
      arr2=np.ndarray(shape=arr.shape,dtype=arr.dtype)
      res=LibDetXR.ByteShuffle(arr.ctypes.data, arr2.ctypes.data, arr.nbytes, arr.itemsize)
    else:
      arr2=arr
    arr3=zlib.compress(arr2,self.lvl)
    arr3=np.frombuffer(arr3, np.ubyte)
    filter=0
    return (arr3,filter)
    #dsid.write_chunk(idx,arr3)

def GetChunkBuildFunc(cmpr):
  'returns a function to compress a numpy array to a compressed chunk, which can be written to a hdf5 file'
  if type(cmpr)==str:
    if cmpr=='dummy':
      return ChunkBuildFuncDummy() # returns data as it is.
    elif cmpr[0]=='n':
      m=re.match('(\d*)',cmpr[1:])
      if m:
        #prec=int(m.group(1))
        cmpr=cmpr[1+m.end():]
        if cmpr!='':
          assert(cmpr[0]=='-')
          cmpr=cmpr[1:]
        else:
          return ChunkBuildFuncDummy() # returns data as it is. we assume that the data comes already n-bit compressed
    if cmpr[0]=='s':
      if len(cmpr)>1:
        assert(cmpr[1]=='-')
        cmpr=cmpr[2:]
      else:
        cmpr=''
      shuffle=True
    else:
      shuffle=False
    m=re.match('zlib(\d*)',cmpr)
    if m:
      try: lvl=int(m.group(1))
      except ValueError:
        lvl=None
      #return ChunkBuildFuncZLIBpy(shuffle,lvl)
      return ChunkBuildFuncZLIB(shuffle,lvl)
    elif cmpr=='lz4':
      return ChunkBuildFuncLZ4(shuffle)
    elif cmpr=='lzf':
      return ChunkBuildFuncLZF(shuffle)
    elif cmpr=='cbf':
      return ChunkBuildFuncCBF(shuffle)
  raise TypeError('unknown compression: '+cmpr)

def AddAttributes(id,attrib):
  'add multiple attributes to a hdf5 object'
  try:
    obj=h5py.Group(id)
  except ValueError as e:
    obj=h5py.Dataset(id)

  for k, v in attrib:
    obj.attrs[k] = v

def SetAttribute(id,name,value):
  'write a single attribure to a hdf5 object'
  if type(value)!=np.ndarray:   #value must be a numpy array
    value = np.asarray(value, order='C')
  htype = h5py.h5t.py_create(value.dtype, logical=True)
  space = h5py.h5s.create_simple(value.shape)
  attr = h5py.h5a.create(id, name, htype, space)
  attr.write(value)


def ConvUtf8(s):
  'convert unicoded json object to ASCII encoded'
  #http://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-ones-from-json-in-python
  if isinstance(s, dict):
    return {ConvUtf8(key): ConvUtf8(value) for key, value in s.iteritems()}
  elif isinstance(s, list):
    return [ConvUtf8(element) for element in s]
  elif isinstance(s, unicode):
    return s.encode('utf-8')
  else:
    return s

class dotdict(dict):
  """dot.notation access to dictionary attributes"""
  def __init__(self,arg=None,**kwargs):
    if arg!=None:
      self.__fill__(arg)
    self.__fill__(kwargs)

  def __fill__(self,kw):
      for k,v in kw.iteritems():
        if type(v)==dict:
          self[k]=dotdict(v)
        else:
          self[k]=v
          if type(v)==list:
            for i,w in enumerate(v):
              if type(w)==dict:
                v[i]=dotdict(w)
            pass

  def __dir__(self):
    l=dir(object)
    #l.extend(self.keys())
    l.extend(map(str,self.keys()))
    return l

  def __getattr__(self, attr):
      #return self.get(attr)
      try:
        return self[attr]
      except KeyError as e:
        raise AttributeError("%r instance has no attribute %r" % (self.__class__, attr))

  def __repr__(self):
    return '<' + dict.__repr__(self)[1:-1] + '>'

  def PrettyPrint(self,indent=0):
    for k,v in self.iteritems():
      if type(v)==dotdict:
        print ' '*indent,str(k)+':'
        v.PrettyPrint(indent+2)
      else:
        print ' '*indent+str(k)+'\t'+str(v)

  __setattr__= dict.__setitem__
  __delattr__= dict.__delitem__
  #__getattr__= dict.__getattr__

def ZmqHeaderFix(hdr):
  'if possible modifies zmq header to fit to chunk1.0'
  htype=hdr['htype']
  if type(htype)==list:
    hdr['htype']=htype=''.join(htype)
  if htype.startswith('array'):
    hdr['htype']=htype='chunk-1.0'
    hdr['shape']=tuple(reversed(hdr['shape']))
  return htype

def LogConfig(args,logLvls):
  '''
  all default log level is set to DEBUG, the loglevels of the dict logLvls is set to the described value.
  if this value is -1, the level will be INFO if verbose&1 is set else WARNING
  further if args has an attribute logFile, the logs are feeded to a rotating logfile (4 backups and max 8MB)
  '''
  logging.basicConfig()
  try:
    logFile=args.logFile
  except AttributeError:
    logFile=None
  if logFile!=None:
    from logging.handlers import RotatingFileHandler
    if os.getuid()==0: #if root is the main process, use a special RotatingFileHandler #getresuid
      class RootRotatingFileHandler(RotatingFileHandler):
        '''same as RotatingFileHandler, but only root is allowed to do
           a file rollover. so a non root process will allways write to the same logfile (a child process inherit opened file descriptors).
     A non root process can not do a file rollover because of the file permissions.'''
        #for debugging check functions:
        #RotatingFileHandler.emit(self, record)
        #RotatingFileHandler.doRollover(self)
        def shouldRollover(self, record):
          if os.getuid()!=0:
            return 0 #if not root, do not rollover (would cause OSError permission)
          return RotatingFileHandler.shouldRollover(self, record)
      hdlr=RootRotatingFileHandler(args.logFile, mode='a', maxBytes=8*1024*1024, backupCount=4)
    else:
      hdlr=RotatingFileHandler(args.logFile, mode='a', maxBytes=8*1024*1024, backupCount=4)
  else:
    hdlr=logging.StreamHandler()
  fmt = logging.Formatter('%(levelname)s:%(name)s:%(lineno)d:%(funcName)s:%(message)s')
  #fmt = logging.Formatter('%(levelname)s:%(message)s')
  hdlr.setFormatter(fmt)
  del logging.root.handlers[:]
  logging.root.addHandler(hdlr)
  logging.root.setLevel(logging.DEBUG)
  if args.verbose&1:
    lvl=logging.INFO
  else:
    lvl=logging.WARNING

  for k,v in logLvls.iteritems():
    if v==-1: v=lvl
    logging.getLogger(k).setLevel(v)
  pass

class ZMQ:
  'zmq helper class'
  ifDict={'PUSH':zmq.PUSH,'PULL':zmq.PULL,'PUB':zmq.PUB,'SUB':zmq.SUB}
  gblCtx=None
  @staticmethod
  def GetCtx():
    'returns the one and only zmq-Context'
    if ZMQ.gblCtx==None or ZMQ.pid!=os.getpid():
      ZMQ.gblCtx=zmq.Context()
      ZMQ.pid=os.getpid()
    return ZMQ.gblCtx

  @staticmethod
  def lst2dict(lst):
    d={}
    for v in lst:
      k,v=v.split(':',1)
      try: v=int(v)
      except ValueError: pass
      d[k]=v
    return d

  @staticmethod
  def Socket(param,ifTypeDef,queueSzDef):
    'creates and returns a socket with various parameters'
    #renaming old keys ifType and queueSz
    try: param['if']=param.pop('ifType')
    except KeyError: pass
    try: param['qs']=param.pop('queueSz')
    except KeyError: pass
    ifType=param.get('if',ifTypeDef)
    if type(ifType)==str:
      param['if']=ifType=ZMQ.ifDict[ifType]
    context = ZMQ.GetCtx()
    sock = context.socket(ifType)#default is zmq.PULL
    if ifType==zmq.SUB: #@UndefinedVariable
      sock.setsockopt(zmq.SUBSCRIBE, '') #@UndefinedVariable
    queueSz=param.get('qs',queueSzDef)
    sock.set_hwm(queueSz)#limit to queueSz messages MUST BE DONE ON SENDER AND RECEIVER?
    addr=param['addr']
    if addr.startswith('bind_'):
      sock.bind(addr[5:])
    elif addr.find('*')>=0:
      sock.bind(addr)
    else:
      sock.connect(addr)
    try:
      timeout=param['timeout']
    except KeyError:pass
    else:
      if timeout!=None:
        sock.RCVTIMEO=timeout # set 1000 ms timeout
    sock.LINGER=1000
    return sock

